-- $LastChangedDate: 2008-10-21 19:35:46 -0400 (Tue, 21 Oct 2008) $
-- Game code.
-- See module.txt for documentation and pitfalls.

-- Randomize.
math.randomseed( os.time() )

----------------------------
-- Game class (singleton) --
----------------------------

Game = { }

-- There are different kinds of scores.
Game.TEAM_NUM_NEUTRAL = 0
Game.TEAM_NUM_PLAYER  = 1
Game.TEAM_NUM_ENEMY   = 2
Game.oppositeTeamNums = { [0]=0, [1]=2, [2]=1 }
Game.SCORE_NONE       = false
Game.SCORE_CRASH      = 1
Game.SCORE_LANDING    = 2
Game.score            = Game.SCORE_NONE
Game.actors           = {}  -- list of game entities

-- Init.
function Game:Init()
	if ( not self.init ) then
		self.init = true
		Timer:Register( self, defs.GAME_TICK_FREQ )
	end
end

-- Reset.
function Game:Reset()
	self.score = Game.SCORE_NONE
end

Events:RegisterReset( function() Game:Reset() end )  -- need anon func

-- Get amount of Actors.
function Game:GetActorCnt()
	return table.getn( self.actors )
end

-- Get table of Actors.
function Game:GetActors()
	return self.actors
end

-- Get next Actor with matching teamNum (or any teamNum if nil).
-- Else returns nil.
function Game:GetNextActor( actor, teamNum )
	-- Get index of next actor.
	local idxNext = nil
	for i=1, #self.actors do
		if ( self.actors[i] == actor ) then
			if ( self.actors[i+1] ) then
				idxNext = i+1
			else
				idxNext = 1 -- wrap
			end

			break
		end
	end
	assert( idxNext ~= nil )  -- oops, actor isn't in list?

	-- Match teamNum?
	if ( teamNum ) then
		-- Scan for actor with matching teamNum.
		for i=idxNext, #self.actors do
			if ( self.actors[i]:GetTeamNum() == teamNum ) then
				return self.actors[i]
			end
		end

		-- Wrap.
		for i=1, idxNext-1 do
			if ( self.actors[i]:GetTeamNum() == teamNum ) then
				return self.actors[i]
			end
		end

		-- Not found.
		return nil
	else
		-- Match any teamNum.
		return self.actors[idxNext]
	end
end

-- Return opposite team number of an actor.
function Game:OppositeTeamNum( actor )
	return self.oppositeTeamNums[actor:GetTeamNum()]
end

-- (private) Increment a type of score in binary configuration.
function Game:WriteScoreInt( key, inc )
	-- Read score from ConfBin.
	local readOk,score = sim.ConfBin:ReadInt( key )
	if ( not readOk ) then
		-- This probably means the integer value isn't yet in the conf file.
		score = 0
	end

	-- Increment score.
	score = score + inc

	-- Write updated score to ConfBin.
	local writeOk = sim.ConfBin:WriteInt( key, score )
	if ( not writeOk ) then
		print( "ERROR [Lua] accessing score in bin conf file" )
	end
end

-- Pass score as Game.SCORE_* enum.
-- Returns true if state transition.
function Game:UpdateScore( score )
	-- Did state change?
	if ( self.score ~= score ) then
		-- Remember state.
		self.score = score

		-- Update binary configuration file (~/.palomino.cfg).
		if ( score == self.SCORE_CRASH ) then
			self:WriteScoreInt( "game_score_crashes", 1 )
		elseif ( score == self.SCORE_LANDING ) then
			self:WriteScoreInt( "game_score_landings", 1 )
		end

		-- Return true to indicate the score changed.
		return true
	else
		-- Return false to indicate no change.
		return false
	end
end

-- Update game every tick.
function Game:Tick()
	-- Cut-scene: death roll with rescue helicopter.
	if ( PlayerActor.aircraft:IfCrash() ) then
		-- This updates only on state transition.
		self:UpdateScore( self.SCORE_CRASH )
		
		-- Show rescue helicopter.
		RescueHelicopter:Start()

		-- Rotate slowly (player's chase-plane).
		ChaseView.chaseViewpointPlayer:Rotate( sim.ChaseViewpoint.AXIS_CHASE_PLANE_HORZ,
											   -Keyboard:ChaseViewRotation() * 0.5 )
	end
end

-- Add actor.
function Game:AttachActor( actor )
	-- C++
	sim.GameHud:AttachTarget( actor:GetObject(), actor:GetName() )

	-- Lua
	table.insert( self.actors, actor )
end

-- Remove actor.
function Game:DetachActor( actor )
	-- C++
	sim.GameHud:DetachTarget( actor:GetObject() )

	-- Lua
	TableRemove( self.actors, actor )
	actor:destroy()  -- must explicitly call destructor/finalizer
end

-- Do it.
Game:Init()

----------------------------------------
-- RescueHelicopter class (singleton) --
----------------------------------------

RescueHelicopter = { }

function RescueHelicopter:Start()
	-- Already started?
	if ( self.rescueHelicopter ) then
		return
	end	

	-- Place rescue helicopter above player's burning wreck.
	local pos = PlayerActor.aircraft:GetPosition()
	pos[LAT] = pos[LAT] - sim.World.Meters2Sim(20.0)
	pos[LON] = pos[LON] - sim.World.Meters2Sim(20.0)
	pos[ALT] = pos[ALT] + sim.World.Meters2Sim(20.0)
	self.rescueHelicopter = sim.Aircraft:new( pos,
		 									  sim.Aircraft.type_Sikorsky,
											  sim.ModelCache.LOAD_SHARED )

	-- Play rotor sound.
	sim.Sound:Play( "rotor.wav.gz", sim.Sound.PLAY_LOOP, 0.6 )

	-- AircraftPhysics is N/A to helicopters.
	self.rescueHelicopter:GetPhysics():Enable( false )
	self.rescueHelicopter:SetThrottle( 1.0 )
	self.rescueHelicopter:SetCollidable( false )
	self.rescueHelicopter:Rotate( AXIS_YAW, Deg2Rad(270.0) );
	sim.World:AttachObject( self.rescueHelicopter )

	-- Change view mode.
	sim.SuperView:SetViewMode( sim.SuperView.eViewMode_CHASE )
	ChaseView.chaseViewpoint = ChaseView.chaseViewpointPlayer
	ChaseView:Reset()
	local distance = sim.World.Meters2Sim( 3.5 * ChaseView.distance )
	ChaseView.chaseViewpoint:SetDistance( 0.0, 0.0, distance )
	ChaseView.chaseViewpoint:Rotate( sim.ChaseViewpoint.AXIS_CHASE_PLANE_VERT,
									 Deg2Rad(-25.0) )
end

function RescueHelicopter:Stop()
	if ( self.rescueHelicopter ) then
		sim.World:DetachObject( self.rescueHelicopter )
		self.rescueHelicopter = nil
	end
end

-- For rescue helicopter to disappear after program reset.
Events:RegisterReset( function() RescueHelicopter:Stop() end )  -- need anon func
