/*****************************************************************************
 * $LastChangedDate: 2008-10-21 14:43:19 -0400 (Tue, 21 Oct 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Various small functions: function macros, inline functions, template functions.
 * @remarks	Include this before types.hh.
 *          Writing "inline" in template functions does increase speed (gcc 3.4).
 *          Omitted (mostly) functions for C++ arrays to encourage use of STL containers or Array.
 *          Write double-braces {{ }} in multi-statement macros
 *          to catch mis-expanding them in if/else clauses.
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef DOXYGEN  // doxygen can't parse this correctly
#ifndef BASE_FUNCS_HH
#define BASE_FUNCS_HH 1

namespace base {

////////////////////////////////////////////////////////////////////////////////
//////////////////////////  function macros  ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

// Min()/Max() is strongly recommended and faster if either arg is a computation,
// but requires both args to be the same type.
// Macro MIN()/MAX() lets the compiler automatically convert args of different types,
// but be warned: has the well-known side-effect of compiling each arg twice!
template<typename T> inline T Min( T a, T b ) { return (a < b) ? a : b; }
template<typename T> inline T Max( T a, T b ) { return (a > b) ? a : b; }
#ifndef MIN
#define MIN(A,B) ( ((A) < (B)) ? (A) : (B) )
#endif
#ifndef MAX
#define MAX(A,B) ( ((A) > (B)) ? (A) : (B) )
#endif

// Uses overloaded abs() from <cmath>.
#define ABS(X) std::abs((X))

// DEPRECATED
#ifndef STREQ
#define STREQ(A,B) ( strcmp((A),(B)) == 0 )
#endif
#ifndef STRNE
#define STRNE(A,B) ( strcmp((A),(B)) != 0 )
#endif
#ifndef STREMPTY
#define STREMPTY(S) ( ((S) == NULL) || ((S)[0] == '\0') )
#endif
#ifndef MEMEQ
#define MEMEQ(S1,S2,LEN) ( memcmp((S1),(S2),(LEN)) == 0 )
#endif
#ifndef MENNE
#define MEMNE(S1,S2,LEN) ( memcmp((S1),(S2),(LEN)) != 0 )
#endif
#define ELEMS(ARRAY) (uint( sizeof((ARRAY)) / sizeof((ARRAY)[0]) ))  // DEPRECATED

/// Nullifying a pointer after deleting can avoid memory corruption.
#define DELETE_NULL( OBJ ) delete (OBJ), (OBJ) = NULL

/// Pass address of bit-field struct.
/// CLEAR_BIT_FIELDS( &mBits );
#define CLEAR_BIT_FIELDS( BIT_FIELD_PTR ) ( *(reinterpret_cast<unsigned int*>( (BIT_FIELD_PTR) )) = 0 )

// Prevent copying an object.  Put these in the private section of a class.
// The copy methods are prototypes without definitions, so usage causes a linker error.
#define PREVENT_COPYING( CLASS )                    \
private:                                            \
    CLASS( const CLASS& );                          \
    CLASS& operator=( const CLASS& );

// To mark classes that are copyable.
#define COPYABLE( CLASS )          // means class is copyable by user-defined copying methods
#define COPYABLE_DEFAULT( CLASS )  // means class is copyable by compiler-generated copying methods
#define ALLOW_COPYING( CLASS )     // synonym

// To mark class that shouldn't be instantiated
#define PREVENT_INSTANTIATING( CLASS )  // NOP
#define PREVENT_INSTANCE( CLASS )  // NOP

/// Helper for correctly writing operator<() for STL container.
/// bool operator<( const C& a, const C& b )
/// {
///     RETURN_LT_TRUE_GT_FALSE( a.x, b.x )
///     RETURN_LT_TRUE_GT_FALSE( a.y, b.y )
///     return false;  // equal
/// }
/// "Effective C++" Item 21 says that a comparison function
/// should return false for equal values or it will corrupt a set/map.
#define RETURN_LT_TRUE_GT_FALSE( X, Y )			\
{{												\
	if ( (X) < (Y) ) return true;				\
	if ( (Y) < (X) ) return false;				\
}}

// Put this in destructors of globals which may call other globals
// to avoid pitfall of undefined order of destruction.
#define RETURN_IF_EXITING  {{ if ( base::gGlobals.mExiting ) return; }}

// In case some cleanup function should be called instead.
void EXIT( int exitCode );

// Arcane indirect trick to give a variable a unique name.
// Subtlety: If written on the same line, UNIQUE_NAME() expands the same name.
// That's a feature not a bug.  BEGIN_ONCE depends on it.
#define UNIQUE_NAME__(NAME,LINE) NAME##LINE
#define UNIQUE_NAME_(NAME,LINE)  UNIQUE_NAME__(NAME,LINE)
#define UNIQUE_NAME(NAME)        UNIQUE_NAME_(NAME,__LINE__)

/// @def BEGIN_ONCE
// BEGIN_ONCE {  <-- write brace  // thread-safe
// ...
// } END_ONCE
// Subtlety: This must be one-line because of UNIQUE_NAME().
// Needs to be locked since BEGIN_ONCE is often placed in threadable code.
#define BEGIN_ONCE							\
{											\
	PERSISTENT FastLock sLockOnce_;			\
	FastLock::Auto autolock(&sLockOnce_);	\
	PERSISTENT int UNIQUE_NAME(once);		\
	++UNIQUE_NAME(once);					\
	if ( UNIQUE_NAME(once) == 1 )

#define END_ONCE \
}

/// @def likely
/// Give branch hint to compiler.
#if __GNUC__ >= 3
#define likely(x)	__builtin_expect((x),1)
#define unlikely(x)	__builtin_expect((x),0)
#else
#define likely(x) (x)
#define unlikely(x) (x)
#endif
#define EX(x) likely((x))	///< abbrev for expected
#define UX(x) unlikely((x))	///< abbrev for unexpected

/// @def UNUSED
/// Tell compiler a var is unused.
#if GCC_VERSION >= 30100  // if gcc-3.1 or later
#define UNUSED    __attribute__ ((unused))
#define DEBUG_VAR __attribute__ ((unused))
#define NOINLINE  __attribute__ ((noinline))
#else
#define UNUSED
#define DEBUG_VAR
#define NOINLINE
#endif

////////////////////////////////////////////////////////////////////////////////
////////////////////////////  inline functions  ////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/// Clear memory
INLINE void
CLEAR( void* p, uint n )
{
	memset( p, 0, n );
}

/// Force a value negative.  Opposite of absolute.
INLINE int         NEG( int         x ) { return (x < 0) ? x : -x; }
INLINE long int    NEG( long int    x ) { return (x < 0) ? x : -x; }
INLINE float       NEG( float       x ) { return (x < 0) ? x : -x; }
INLINE double      NEG( double      x ) { return (x < 0) ? x : -x; }
INLINE long double NEG( long double x ) { return (x < 0) ? x : -x; }

INLINE bool
IfPow2( int n )
{
	// Trick: result is 0 if n is a power of 2.
//  return (n & (n-1)) == 0;  // bug: 0 isn't a power-of-2
    return (n != 0) and ((n & (n-1)) == 0);
}

INLINE bool IfEven( int n ) { return (n & 1) == 0; }
INLINE bool IfOdd( int n )  { return (n & 1) == 1; }

// +1 or -1 if true or false
INLINE int
PLUS_MINUS_1( bool plus )
{
	return plus ? 1 : -1;
}

////////////////////////////////////////////////////////////////////////////////
/////////////////////////  template functions  /////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Swap.
 *****************************************************************************/
template<typename T>
inline void
Swap( T& a, T& b )
{
	const T a2 = a;
	a = b;
	b = a2;
}

/*****************************************************************************
 * @return True if two numbers are both positive or both negative.
 *****************************************************************************/
template<typename T>
bool
IfSameSign( T a, T b )
{
	return ((a  < 0) and (b  < 0))
        or ((a >= 0) and (b >= 0));  // 0 is positive
}

} // namespace base

#endif // BASE_FUNCS_HH
#endif // DOXYGEN
