/*****************************************************************************
 * $LastChangedDate: 2008-04-11 10:14:17 -0400 (Fri, 11 Apr 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Globals class (the fewer the better).
 *//*
 * LEGAL:	COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

// NOTE: "global" is a namespace, "gGlobals" is a var (instance of Globals class).

#ifndef BASE_GLOBALS_HH
#define BASE_GLOBALS_HH 1

namespace base {

////////////////////////////////////////////////////////////////////////////////
/// @brief Globals class collects globals as members to determine their order of construction.
///
/// NOTE: "global" is a namespace, "gGlobals" is a var (instance of Globals class).
///
/// Design philosophy is to minimize globals but make them explicit
/// by putting them in a "global" namespace and using the qualifier "global::".
/// The Singleton pattern or otherwise hiding globals is detrimental to thread-safety.
///
/// Globals construction/destruction order:
/// Some globals may depend on other globals being constructed first.
/// C++ doesn't define the order which separate global variables are constructed.
/// But the order of constructing class members is defined as the order
/// in which they're declared (destruction order is defined as the reverse).
///
/// Globals class vs. conventional singletons:
/// Singleton disadvantages: tedious to write, obscure the fact they're globals.
/// Singleton advantages: are initialized after load-time, solve run-time dependencies.
/// Globals class advantages: searchable, the order of initializing class members is defined.
/// World::GetEdge();          // obscure global
/// global::gWorld.GetEdge();  // obvious global
///
class Globals
{
PREVENT_COPYING(Globals)  // plural to differentiate from "global" namespace
public:
	Globals( void )
	:	mExiting(false),
		mTesting(false)
	{
		// NOP
	}

	// The order of construction/destruction of class members
	// is determined by the order in which they're declared.
	// ctor initialization list has no effect on order.

	// mExiting: Destructors that might destroy a static object should
	// check mExiting to avoid calling methods of another destroyed static object.

	bool				mExiting;
    bool				mTesting;
	Streams				mStreams;
#if COMPILE_THREADS
	Thread				mThread;
#endif
#if DEBUG
	bool				mAssertAbort;  // call abort() if ASSERT() fails
#endif
};

// The dreaded Globals var.
BASE_GLOBAL Globals gGlobals;

} // namespace base

#endif // BASE_GLOBALS_HH
