/*****************************************************************************
 * $LastChangedDate: 2008-09-09 00:17:30 -0400 (Tue, 09 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Operating-system specific header file.
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef BASE_OS_HH
#define BASE_OS_HH 1

// base_os.hh is included early so it should be independent of other base*.hh.

// Try to define the OS_* defines which may have been pre-defined by make files.
#if ! defined(OS_LINUX) && (defined(linux) || defined(__linux__))
#	define OS_LINUX 1
#elif ! defined(OS_FREEBSD) && defined(__FreeBSD__)
#	define OS_FREEBSD 1
#elif ! defined(OS_APPLE) && defined(__APPLE__)
#	define OS_APPLE 1
#elif ! defined(OS_WINDOWS) && defined(_MSC_VER)
#	define OS_WINDOWS 1  // native Windows
#elif ! defined(OS_CYGWIN) && defined(__CYGWIN__)
#	define OS_CYGWIN 1
#endif

// --- OS-specific ---
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <stdlib.h>
#include <stdio.h>
#ifndef OS_WINDOWS
#include <unistd.h>
#include <sys/times.h>
#endif
#include <time.h>
#include <fcntl.h>
#include <errno.h>

// --- ANSI/POSIX/universal header files (base_cc.hh includes C++ standard headers) ---
#include <limits.h>
#include <float.h>
//#include <math.h>  // base_cc.hh includes C++ <cmath>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#ifndef OS_WINDOWS
#include <dirent.h>
#endif

#ifdef OS_WINDOWS
#pragma warning(disable:4005)  // suppress macro redefinition
#pragma warning(disable:4355)  // "this" in base init list (ok if just pointer stored)
#endif

// To overcome VC++ warning 4800 (it has caught severe bugs).
#define BOOL_CAST( EXPR ) ((EXPR) != 0)

//------------------------------------------------------------------------------

namespace base {

// BASE_LITTLE_ENDIAN or BASE_BIG_ENDIAN abstracts compiler differences.
#if BYTE_ORDER == LITTLE_ENDIAN  // gcc
#   define BASE_LITTLE_ENDIAN 1
#elif BYTE_ORDER == BIG_ENDIAN  // gcc
#   define BASE_BIG_ENDIAN 1
#else
#   error "Need to define BASE_LITTLE/BIG_ENDIAN in base_os.hh"
#endif

// Testing GCC_VERSION is tricky.
//
// Test if gcc is version 4.1 or higher:
// GCC_VERSION >= 40100 // right: 2 digits for minor/micro fields!
// GCC_VERSION >= 41000 // wrong
//
// Test if gcc is below gcc 3.0 requires two conditions:
// GCC_VERSION > 0 && GCC_VERSION < 30000  // right
//
#if defined(__GNUC__)
#	if defined(__GNUC_PATCHLEVEL__)
#		define GCC_VERSION (__GNUC__*10000 + __GNUC_MINOR__*100 + __GNUC_PATCHLEVEL__)
#	else
#		define GCC_VERSION (__GNUC__*10000 + __GNUC_MINOR__*100)
#	endif
#else
#	define GCC_VERSION 0
#endif

// For checking a specific gcc minor version (inappropriate for testing at-least-this-version).
#if defined(__GNUC__) && (GCC_VERSION >= 40000 && GCC_VERSION < 40100)
#define GCC_40  1
#endif

#if defined(__GNUC__) && (GCC_VERSION >= 40100 && GCC_VERSION < 40200)
#define GCC_41  1
#endif

#if defined(__GNUC__) && GCC_VERSION < 30400
#error "gcc 3.4 or later is required"
#endif

#if defined(__GNUC__) && GCC_VERSION < 40000
// Workaround: gcc 3.4 has problems with friends that are nested classes
// and friend class eponyms.
#define private public
#define protected public
#endif

// Filesystem specific.
#define EXEC_SUFFIX_STR				""
#define DIRECTORY_SEPARATOR_CHAR	'/'  // Windows supports UNIX forward slash
#define DIRECTORY_SEPARATOR_STR		"/"
const string DIRECTORY_SEPARATOR_STRING("/");
#if OS_WINDOWS
#define DOSFS  1
#define GETCWD				_getcwd
#else
#define GETCWD				getcwd
#endif

// Kludges for VC++.
#if OS_WINDOWS
#define and &&
#define or ||
#define not !
#undef uint
#define uint unsigned int
#endif

} // namespace base

#endif // BASE_OS_HH
