/*****************************************************************************
 * $LastChangedDate: 2008-09-14 08:59:31 -0400 (Sun, 14 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief   Template class for sharing objects by pointer.
 * @remarks Documented at shared_ptr_doc.hh.
 *//*
 * LEGAL:	COPYRIGHT (C) JIM E. BROOKS 2007
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef BASE_SHARED_PTR_HH
#define BASE_SHARED_PTR_HH 1

//..............................................................................
#if COMPILE_OSG

#include <osg/Referenced>
#include <osg/ref_ptr>
#define RefPtr   osg::ref_ptr
#define RefPtr0  osg::ref_ptr

#endif
//..............................................................................

namespace base {

////////////////////////////////////////////////////////////////////////////////
// Shared
//
// Clients can just derive using "public Shared" since mRefCnt is private.
// "private Shared" is ideal but every client would have to declare SharedPtr a friend.
//
class Shared
{
public:
	uint GetRefCnt( void ) const { CHECK_TYPESIG(this,TYPESIG_SHARED);
                                   return mRefCnt; }
	// LIMITED USE: A special case is where a class has a Shared as an
	// actual member but has methods that return it as a SharedPtr.
	// The class implicitly hold a reference to it, so ctor should increment ref-cnt.
	void IncRefCnt( void ) { ++mRefCnt; }

protected:
	// Shared initially has zero references.
	// Copy ctor makes a copy also with zero references.
	Shared( void )              : mRefCnt(0) { SET_TYPESIG(this,TYPESIG_SHARED); }
	Shared( const Shared& src ) : mRefCnt(0) { SET_TYPESIG(this,TYPESIG_SHARED); }
    ~Shared() { INVALIDATE_TYPESIG(this,TYPESIG_SHARED); }

	// operator=() is explicitly defined as NOP (compiler would copy ref-cnts).
	// Reassigning the value of a Shared should not affect ref-cnts.
    Shared& operator=( const Shared& src ) { return *this; }

private:
    mutable int mRefCnt;

public:
	DECL_TYPESIG(TYPESIG_SHARED)

template<typename SHARED> friend class SharedPtr;
template<typename SHARED> friend class SharedPtrNull;
template<typename SHARED> friend class SharedPtrLocked;
template<typename SHARED> friend class SharedPtrNullLocked;
};

// A faux-pas for the sake of finer-granularity is to put the lock inside the Shared class.
// That was tried but it lead to having to acquire two locks in the copy methods which
// caused potential deadlock and code would've become convoluted to avoid deadlock.
#if COMPILE_THREADS
namespace restrict { BASE_GLOBAL FastLock gLockShared; }
#endif

// SharedPtr methods contain the macro TEST_SHARED_PTR_YIELD_CPU()
// to help better test lock contention.
#ifndef TEST_SHARED_PTR_YIELD_CPU
#define TEST_SHARED_PTR_YIELD_CPU()
#endif

// Abbreviation.
#define shptr SharedPtr
#define shptr0 SharedPtrNull

// Explicit long names used by tests.
#define SharedPtrNoLock     SharedPtr
#define SharedPtrNullNoLock SharedPtrNull

// SharedPtr<FastLock>
#define SharedPtr  SharedPtrLocked
#define THREADABLE  Threadable
#define IF_NOT_NULL(PTR)
#define SHARED_PTR_NULL  0
#define SHARED_PTR_LOCK()   THREAD_CODE( restrict::gLockShared.Lock() )
#define SHARED_PTR_UNLOCK() THREAD_CODE( restrict::gLockShared.Unlock() )
#include "base/shared_ptr2.hh"
#undef SharedPtr
#undef THREADABLE
#undef IF_NOT_NULL
#undef SHARED_PTR_NULL
#undef SHARED_PTR_LOCK
#undef SHARED_PTR_UNLOCK

// SharedPtrNull<FastLock>
#define SharedPtr  SharedPtrNullLocked
#define THREADABLE  Threadable
#define IF_NOT_NULL(PTR)  if ( EX( (PTR) != (NULL) ) )
#define SHARED_PTR_NULL  1
#define SHARED_PTR_LOCK()   THREAD_CODE( restrict::gLockShared.Lock() )
#define SHARED_PTR_UNLOCK() THREAD_CODE( restrict::gLockShared.Unlock() )
#include "base/shared_ptr2.hh"
#undef SharedPtr
#undef THREADABLE
#undef IF_NOT_NULL
#undef SHARED_PTR_NULL
#undef SHARED_PTR_LOCK
#undef SHARED_PTR_UNLOCK

// SharedPtr<NoLock>
#define SharedPtr  SharedPtr
#define THREADABLE  NonThreadable
#define IF_NOT_NULL(PTR)
#define SHARED_PTR_NULL  0
#define SHARED_PTR_LOCK()
#define SHARED_PTR_UNLOCK()
#include "base/shared_ptr2.hh"
#undef SharedPtr
#undef THREADABLE
#undef IF_NOT_NULL
#undef SHARED_PTR_NULL
#undef SHARED_PTR_LOCK
#undef SHARED_PTR_UNLOCK

// SharedPtrNull<NoLock>
#define SharedPtr  SharedPtrNull
#define THREADABLE  NonThreadable
#define IF_NOT_NULL(PTR)  if ( EX( (PTR) != (NULL) ) )
#define SHARED_PTR_NULL  1
#define SHARED_PTR_LOCK()
#define SHARED_PTR_UNLOCK()
#include "base/shared_ptr2.hh"
#undef SharedPtr
#undef THREADABLE
#undef IF_NOT_NULL
#undef SHARED_PTR_NULL
#undef SHARED_PTR_LOCK
#undef SHARED_PTR_UNLOCK

} // namespace base

#endif // BASE_SHARED_PTR_HH
