/******************************************************************************
 * $LastChangedDate: 2008-08-10 18:21:18 -0400 (Sun, 10 Aug 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Streams.
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 ******************************************************************************/

#define BASE_STREAM_CC 1
#include "base/module.hh"
#include "base/stream.hh"
#include "base/file.hh"
#undef cout
#undef cerr

namespace base {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////  functions  //////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Open a stream to a file.
 * @param	fs
 *			File stream.
 * @param	bits
 *			Open for reading and/or writing.
 * @param	filename
 * @return	False if stream error.
 *****************************************************************************/
bool
StreamOpen( fstream& fs, ios::openmode bits, const string& filename )
{
	// Binary else trouble.
	bits |= ios::binary;

	// Try opening a stream to a file.
	// For the open-for-writing case, first try to open the file without the truncate bit.
	// But some iostream implementations might fail if the file doesn't exist.
	fs.open( filename.c_str(), bits );
	if ( (not fs.good())				// if error
     and (bits & ios::out)				// if output stream
     and (not (bits & ios::trunc)) )	// if truncate not already set
	{
		// This might be the file-doesn't-exist case depending
		// on the iostream implementation.  Retry with truncate bit.
		fs.clear();
		fs.open( filename.c_str(), bits | ios::trunc );
	}

	return fs.good();
}

/*****************************************************************************
 * Read all chars to EOL.
 * @return False if stream error.
 *****************************************************************************/
bool
StreamReadLine( istream& strm, string& line )
{
	line.erase();
	char c;
	while ( strm.good() and (c = strm.get()) != '\n' )
	{
		line += c;
	}
	return strm.good();
}

/*****************************************************************************
 * Jump to the next line.
 * @return False if stream error.
 *****************************************************************************/
bool
StreamToNextLine( istream& strm )
{
	while ( strm.good() && (strm.get() != '\n') ) { }
	return strm.good();
}

/*****************************************************************************
 * Peek input stream.
 * @return True if word is in input stream.
 *****************************************************************************/
bool
StreamPeek( istream& strm, const string& word )
{
	bool result = false;

	// Save read position.
	std::ios::pos_type pos = strm.tellg();

	try
	{
		// Read char-by-char into a buf
		string buf;
		for ( uint i = 0; i < word.size(); ++i )
		{
			const char c = char(strm.get());
			if ( not strm.good() )
				break;

			switch ( c )
			{
				case CHAR_LF: case CHAR_CR: case ' ': case '\t':
				{
					--i;  // don't append whitespace to buf[]
					continue;
				}
				break;

				default:
				{
					buf += char(c);
				}
				break;
			}

			if ( buf[i] != word[i] )  // mismatch?
				break;
		}
		result = (buf == word) && strm.good();
	}
	catch (...)
	{
		result = false;
	}

	// Restore read position.
	strm.seekg( pos );

	return result;
}

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////  Streams  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Streams::Streams( void )
:	mOut(&cout),
	mError(&cerr),
	mLog(&cerr),
	mDebug(&cerr),
	mNopStreambuf(),            // -- ORDER-DEPENDENT --
	mNopStream(&mNopStreambuf)  // -- ORDER-DEPENDENT --
{
	Open();
}

Streams::~Streams()
{
RETURN_IF_EXITING;

	Close();
}

/*****************************************************************************
 * This is called very-very early in startup before modules/conf is available.
 *****************************************************************************/
void  // CLASS_METHOD
Streams::Init( int argc, char** argv )
{
BEGIN_ONCE {

	// Enable/disable the log files and CLOG stream.
	bool verbose = true;
	for ( int i = 1; i < argc; ++i )
	{
		if ( STREQ(argv[i],"-v" ) )
			verbose = true;
		else if ( STREQ(argv[i],"-q" ) )
			verbose = false;
	}

	// Ok (gGlobals.mStreams was constructed even before main()).
	gGlobals.mStreams.SetVerbose( verbose );  // affects CLOG

} END_ONCE
}

/*****************************************************************************
 * Open/close streams for current thread.
 * Thread::Run() calls these.
 *****************************************************************************/
void
Streams::Open( void )
{
THREAD_AUTO_SLOW_LOCK;

	mOut   = &cout;
	mDebug = &cerr;

#if COMPILE_THREADS
	// Let only main thread open log file.
	// Thereafter, threads reuse FILE*.
	BEGIN_ONCE
	{
		mErrorLogFile = fopen( mErrorLogFname.c_str(), "w" );
	} END_ONCE
#endif

	// Error stream forks to stderr and a log file.
	if ( mErrorLogFile != NULL )
	{
		mError = new ostream( new StreambufFile2( stderr, false,
                                                  mErrorLogFile, false ) );
	}
	else
	{
		mError = &cerr;  // couldn't open log file
	}

#if 0 // too soon to access BASE_CONF
	// Disable the CLOG stream?
	SetVerbose( BASE_CONF.mVerbose );
#endif
}

void
Streams::Close( void )
{
	// Just leave log files and streams open/allocated since program is exiting.
}

/*****************************************************************************
 * Set logging.
 *****************************************************************************/
void
Streams::SetErrorLogFname( const string& fname )
{
	mErrorLogFname = fname;
}

void
Streams::SetVerbose( const bool verbose )
{
	if ( verbose )
		mLog = &cerr;
	else
		mLog = &mNopStream;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////  StreambufFile2  //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 * @param	file0, file1
 *          Second is optional.
 * @param	close0, close1
 *			If destructor should close file.  Never deletes FILE* object.
 *          Second is optional.
 *****************************************************************************/
StreambufFile2::StreambufFile2( SafePtr<FILE> file0, bool close0,
                                SafePtr<FILE> file1, bool close1 )
:	mFile0(file0),
	mFile1(file1),
	mClose0(close0),
	mClose1(close1)
{
}

StreambufFile2::~StreambufFile2()
{
	// First file is always present, second is optional.
	if ( mClose0 )                   fclose( mFile0.PTR() );
	if ( mClose1 && mFile1 != NULL ) fclose( mFile1.PTR() );
}

/*****************************************************************************
 * Either is called by iostream to write out buffer.  xsputn() is optional/optimal.
 *****************************************************************************/
int
StreambufFile2::overflow( int c )
{
                          fputc( c, mFile0.PTR() );
	if ( mFile1 != NULL ) fputc( c, mFile1.PTR() );
	return ~EOF;
}

std::streamsize
StreambufFile2::xsputn( const char* buf, std::streamsize n )
{
                          fwrite( buf, n, 1, mFile0.PTR() );
	if ( mFile1 != NULL ) fwrite( buf, n, 1, mFile1.PTR() );
	return n;
}

} // namespace base
