/*****************************************************************************
 * $LastChangedDate: 2008-05-03 17:36:09 -0400 (Sat, 03 May 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Types for time.
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef BASE_TIME_HH
#define BASE_TIME_HH 1

namespace base {

////////////////////////////////////////////////////////////////////////////////
/// @brief Time units.
///
/// @verbatim
/// Examples:
/// Milliseconds(Second(secs))        // convert seconds to milliseconds
///
/// Avoid explicit casting to int.  If that's necessary, cast to typedef Int.
/// Instead of casting to int to do arithmetic, cast to Milliseconds to use
/// the operator() arithmetic methods.  operator Int() can automatically
/// produce an integer.
///
/// int droppedFrames = int64(mMillisecFrame) / int64(mMillisecTimer) - 1;  // BAD
/// int droppedFrames = mMillisecFrame / mMillisecTimer - Milliseconds(1);  // better
/// Milliseconds t = Max( int(millisec1), int(millisec2) );                 // BAD
/// Milliseconds t = Max<Milliseconds::Int>( millisec1, millisec2 );        // better
///
/// int64 was chosen to avoid overflow.  uint64 wasn't chosen
/// since there's a need to compute negative time units.
/// @endverbatim
///
struct Seconds
{
    typedef int64 Int;
    Seconds( void )     : mSeconds(0) { }
    Seconds( Int secs ) : mSeconds(secs) { }
    operator Int() const { return mSeconds; }
    friend inline Seconds operator+( const Seconds& a, const Seconds& b ) { return a.mSeconds + b.mSeconds; }
    friend inline Seconds operator-( const Seconds& a, const Seconds& b ) { return a.mSeconds - b.mSeconds; }
private:
    Int mSeconds;
};

/// @brief Distinct type.  See Seconds.
///
struct Milliseconds
{
    typedef int64 Int;
    Milliseconds( void )          : mMilliseconds(0) { }
    Milliseconds( Int millisecs ) : mMilliseconds(millisecs) { }
    Milliseconds( Seconds secs )  : mMilliseconds(Int(secs)*Int(THOUSAND)) { }
    operator Int() const { return mMilliseconds; }
    friend inline Milliseconds operator+( const Milliseconds& a, const Milliseconds& b ) { return a.mMilliseconds + b.mMilliseconds; }
    friend inline Milliseconds operator-( const Milliseconds& a, const Milliseconds& b ) { return a.mMilliseconds - b.mMilliseconds; }
    friend inline Milliseconds operator*( const Milliseconds& a, const Milliseconds& b ) { return a.mMilliseconds * b.mMilliseconds; }
    friend inline Milliseconds operator/( const Milliseconds& a, const Milliseconds& b ) { return a.mMilliseconds / b.mMilliseconds; }
    friend inline         bool operator<(  const Milliseconds& a, const Milliseconds& b ) { return a.mMilliseconds <  b.mMilliseconds; }
    friend inline         bool operator<=( const Milliseconds& a, const Milliseconds& b ) { return a.mMilliseconds <= b.mMilliseconds; }
    friend inline         bool operator>(  const Milliseconds& a, const Milliseconds& b ) { return a.mMilliseconds >  b.mMilliseconds; }
    friend inline         bool operator>=( const Milliseconds& a, const Milliseconds& b ) { return a.mMilliseconds >= b.mMilliseconds; }
	Milliseconds& operator+=( const Milliseconds& src ) { mMilliseconds += src.mMilliseconds; return *this; }
	Milliseconds& operator-=( const Milliseconds& src ) { mMilliseconds -= src.mMilliseconds; return *this; }
private:
    Int mMilliseconds;
};

/// @brief Distinct type.  See Seconds.
///
struct Microseconds
{
    typedef int64 Int;
    Microseconds( void )                   : mMicroseconds(0) { }
    Microseconds( Int usecs )              : mMicroseconds(usecs) { }
    Microseconds( Milliseconds millisecs ) : mMicroseconds(Int(millisecs)*Int(THOUSAND)) { }
    Microseconds( Seconds secs )           : mMicroseconds(Int(secs)*Int(MILLION)) { }
    operator Int() const { return mMicroseconds; }
private:
    Int mMicroseconds;
};

} // namespace base

#endif // BASE_TIME_HH
