/*****************************************************************************
 * $LastChangedDate: 2008-10-21 14:43:19 -0400 (Tue, 21 Oct 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Collision-detection class.
 *//*
 * LEGAL:	COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define COLLISION_COLLISION_CC 1
#include "base/module.hh"
#include "base/conf.hh"
using namespace base;
#include "object/module.hh"
#include "object/dyna.hh"
using namespace object;
#include "world/module.hh"
#include "world/events.hh"
using namespace world;
#include "collision/module.hh"
#include "collision/collision.hh"
#include "collision/defs.hh"

#include <osgUtil/LineSegmentIntersector>
#include <osgUtil/IntersectionVisitor>

namespace collision {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  Collision  //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

Collision* Collision::msInstance;

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Collision::Collision( void )
:	mCollidables()
{
	// Unlisten() isn't necessary for singletons.
    WORLD_EVENTS.mDetachObject.Listen( new DetachObjectListener( *this ) );
}

Collision::~Collision()
{
    // NOP
}

/*****************************************************************************
 * Set an Object as collidable or indestructible.
 *****************************************************************************/
void
Collision::SetCollidable( shptr<Object> object, const bool collidable )
{
CHECK_TYPESIG(object,TYPESIG_OBJECT);

	if ( collidable )
		mCollidables.insert( object );
	else
		mCollidables.erase( object );
}

/*****************************************************************************
 * @return (true,object) or (false,NULL).
 *
 * Compute intersections of 2 rays:
 * - ray of "down vector" for contact force
 * - ray of the movement of Dyna
 *****************************************************************************/
pair<bool,shptr0<Object> >
Collision::IfCollision( shptr<Dyna> dyna )
{
CHECK_TYPESIG(dyna,TYPESIG_DYNA);

	// Is this Dyna indestructible?
	// A special-case is to ignore zombie Dynas with a minimal graph.
	// SUBTLE: IfCollidable() will return false for "zombie" Objects
	// (which one-node "minimal graph" that will cause OSG intersection-test to fail).
	if ( not dyna->IfCollidable() )
		return make_pair<bool,shptr0<Object> >( false, NULL );

	// Group of 1 or 2 LineIntersectors.
	RefPtr<osgUtil::IntersectorGroup> intersectorGroup = new osgUtil::IntersectorGroup;

	// Compute ray of "down vector" for contact force.
	const WorldVertex worldUpVector  = gWorld.ComputeUpVector( *dyna );  // normalized
	const fp          dynaHeightHalf = dyna->GetHeight() * 0.5f;
	const WorldVertex dynaCenter = dyna->GetPosition();
	const WorldVertex dynaTop(    dynaCenter[XX] + worldUpVector[XX] * dynaHeightHalf,
                                  dynaCenter[YY] + worldUpVector[YY] * dynaHeightHalf,
                                  dynaCenter[ZZ] + worldUpVector[ZZ] * dynaHeightHalf );
	const WorldVertex dynaBottom( dynaCenter[XX] - worldUpVector[XX] * dynaHeightHalf,
                                  dynaCenter[YY] - worldUpVector[YY] * dynaHeightHalf,
                                  dynaCenter[ZZ] - worldUpVector[ZZ] * dynaHeightHalf );
	intersectorGroup->addIntersector( new osgUtil::LineSegmentIntersector( dynaTop,
                                                                           dynaBottom ) );

	// Compute ray of movement of Dyna (skip if Dyna hasn't moved far enough).
	const pair<WorldVertex,WorldVertex> step = dyna->GetTranslationStep();
	const WorldVertex& translationBegin      = step.first;
	const WorldVertex& translationEnd        = step.second;
	const fp           translationDistance   = Distance( translationBegin, translationEnd );
	if ( translationDistance > Geo::Meters2Sim( collision::defs::MIN_TRANSLATION_DISTANCE_METERS ) )
	{
		intersectorGroup->addIntersector( new osgUtil::LineSegmentIntersector( translationBegin,
	                                                                           translationEnd ) );
	}

	// Attach Intersectors to IntersectionVisitor.
	osgUtil::IntersectionVisitor visitor( intersectorGroup.get() );

	// For each collidable Object.
	for ( Collidables::const_iterator iter = mCollidables.begin();
	      iter != mCollidables.end();
	      ++iter )
	{
		// Get collidable.
		shptr<Object> collidable = *iter;
		CHECK_TYPESIG(collidable,TYPESIG_OBJECT);

		// Don't crash into itself!
		if ( EX( dyna.PTR() != collidable.PTR() ) )
		{
			// Detect collision using by visiting this Object's nodes.
			collidable->GetRootNode()->accept( visitor );

			// Collision?
			if ( intersectorGroup->containsIntersections() )
			{
				return make_pair<bool,shptr0<Object> >( true, collidable.PTR() );
			}
		}

		// Reset intersection.
		visitor.reset();
		intersectorGroup->reset();
	}

	// No collision.
	return make_pair<bool,shptr0<Object> >( false, NULL );
}

/*****************************************************************************
 * Listen for the detaching of an Object.
 *****************************************************************************/
void
Collision::DetachObjectListener::operator()( shptr<Object> object )
{
	mCollision.mCollidables.erase( object );
}

} // namespace collision
