/*****************************************************************************
 * $LastChangedDate: 2008-10-14 22:37:47 -0400 (Tue, 14 Oct 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Higher-level Joystick class.
 * @verbatim
 *
 * Overview:
 * ---------
 * Every tick, joystick events are dequeued from InputQueue.
 * control::Joystick handles axis movements by manipulating
 * the current Craft (defined by Control).
 *
 * Continuing of axis events:
 * --------------------------
 * The handler will continue an axis event until another one supercedes it.
 * For example, holding the joystick all the way right for a while
 * is supposed to continually do a roll, yet one axis event will be enqueued
 * at the axis extreme then enqueuing will stop.
 *
 * @endverbatim
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define CONTROL_JOYSTICK_CC 1
#include "base/module.hh"
#include "base/conf.hh"
using namespace base;
#include "input/module.hh"
#include "input/joystick.hh"
#include "input/conf.hh"
using namespace input;
#include "control/module.hh"
#include "control/joystick.hh"
#include "control/defs.hh"
#include "control/conf.hh"

namespace control {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  Joystick  ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

Joystick*   Joystick::msInstance;  // there can be only one

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Joystick::Joystick( void )
:	mEnableInitial(false),  // if EnableInitial() was called (regardless if it succeeded or not)
	mCalibration(),
	mAxisMap(),
	mEvent(),
	mAutofire(false),
	mAxisValues(0)  // fill TinyArray
{
ASSERT_RUN_ONCE;
}

Joystick::~Joystick()
{
    // NOP
}

/*****************************************************************************
 * @return True if joystick is enabled.
 *****************************************************************************/
bool
Joystick::IfEnabled( void )
{
	// Defer to lower-level Joystick.
	return input::Joystick::GetInstance().IfEnabled();
}

/*****************************************************************************
 * Enable joystick.
 * @return True if joystick is enabled.
 *****************************************************************************/
bool
Joystick::Enable( bool enable )
{
	// The first attempt to enable does the full enablement
	// (installing a timer functor and print status should be done once).
	// Subsequent attempts affect the low-level Joystick object.

	if ( enable and not mEnableInitial )
	{
		// Remember initial attempt to enable.
		mEnableInitial = true;

		EnableInitial();  // enable=true
	}
	else
	{
		EnableSubsequent( enable );  // enable/disable
	}

	return input::Joystick::GetInstance().IfEnabled();
}

void
Joystick::EnableInitial( void )
{
	// Try to open the joystick driver.
	input::Joystick::GetInstance().Enable();
	if ( input::Joystick::GetInstance().IfEnabled() )
	{
		// Set poll freq.
		input::Joystick::GetInstance().SetPollFreq( Milliseconds(CONTROL_CONF.mJoystickTickFreq) );

		// Confirm to user that joystick is available/enabled.
		input::Joystick::GetInstance().PrintStatus();

    	// Register a timer functor to poll for joystick events to control the current Craft.
        Timer::GetInstance().Register( new TimerFunctor( *this ),
                                       Milliseconds(CONTROL_CONF.mJoystickTickFreq) );
	}
}

void
Joystick::EnableSubsequent( bool enable )
{
	// Defer to low-level Joystick object.
	input::Joystick::GetInstance().Enable( enable );
}

/*****************************************************************************
 * Set/get calibration.
 *****************************************************************************/
void
Joystick::SetCalibration( const Calibration& calibration )
{
	mCalibration = calibration;
}

Joystick::Calibration
Joystick::GetCalibration( void )
{
	return mCalibration;
}

/*****************************************************************************
 * Set/get axis map.
 *****************************************************************************/
void
Joystick::MapAxis( const int axisLogical, const int axisReal )
{
ASSERT( mAxisMap.mAxises.size() == uint(MAX_AXIS) );

	if ( (uint(axisLogical) < mAxisMap.mAxises.size())
     and (uint(axisReal) < mAxisMap.mAxises.size()) )
	{
		mAxisMap.mAxises[axisLogical] = axisReal;
	}
	else
	{
		CERROR << "WARNING: bad axis map " << axisLogical << ' ' << axisReal << endl;
	}
}

/*****************************************************************************
 * Timer-tick handler.
 *****************************************************************************/
void
Joystick::TimerFunctor::operator()( UNUSED const Milliseconds millisecElapsed )
{
	BUSY_BEGIN {			// prevent being interrupted by another timer-tick
		mJoystick.Poll();	// a loop around polling here caused problems
	} BUSY_END
}

/*****************************************************************************
 * Poll joystick axises and buttons.
 *****************************************************************************/
void
Joystick::Poll( void )
{
	// Is joystick enabled?
	if ( input::Joystick::GetInstance().IfEnabled() )
	{
		// Try to dequeue an event.  Queue could be empty.
		input::InputQueue::GetInstance().DequeueJoyEvent( mEvent /*OUT*/ );

		// These two methods are called always.
		// Axis events may be repeated since movement
		// must continue until moved back to the deadzone.
		// Button events may be repeated in case of autofire.
		ProcessAxisEvent( mEvent );
		ProcessButtonEvent( mEvent );
	}
}

/*****************************************************************************
 * Process a joystick axis event.
 *
 * Continuing axis movements:
 * Imagine the joystick moved to top-left corner then moved up and down.
 * The leftward movement remains in effect while the new events are occurring
 * for up/down movements.  Therefore, persistent variables are used to
 * continue axis movement until the axis is moved to its deadzone.
 *
 * Remapping axises:
 * Remapping an axis is done effectively by calling AxisMap::If*()
 * (remapping isn't done by changing the Event struct).
 *****************************************************************************/
void
Joystick::ProcessAxisEvent( const input::Joystick::Event& event )
{
	PERSISTENT fp sRollFraction  = 0.0f;
	PERSISTENT fp sPitchFraction = 0.0f;
	PERSISTENT fp sYawFraction   = 0.0f;

	control::Control& control = control::Control::GetInstance();
	shptr0<Craft>     craft   = control.GetCurrentCraft();

    // For GetAxisValue().
	if ( uint(event.mAxis) < mAxisValues.size() )  // could be NONE=-1 or superfluous axis
	    mAxisValues[event.mAxis] = event.mAxisVal;

	// If axis val is between {-max..-min} or {+min..+max}
	#define IF_IN_AXIS_RANGE( AXIS_RANGE )													\
		((event.mAxisVal < 0)																\
         and (event.mAxisVal >= AXIS_RANGE.mMax) and (event.mAxisVal <= AXIS_RANGE.mMin))	\
     or ((event.mAxisVal >= 0)																\
         and (event.mAxisVal >= AXIS_RANGE.mMin) and (event.mAxisVal <= AXIS_RANGE.mMax))

	// -------------------------------------------------------------------------
	// Roll:
	if ( mAxisMap.IfRoll( event ) )
	{
		if ( IF_IN_AXIS_RANGE( mCalibration.mRollLeft ) )
		{
			sRollFraction = ComputeFraction( event, AXIS_ROLL, mCalibration.mRollLeft, control.GetRollLeftDir() );
		}
		else if ( IF_IN_AXIS_RANGE( mCalibration.mRollRight ) )
		{
			sRollFraction = ComputeFraction( event, AXIS_ROLL, mCalibration.mRollRight, control.GetRollRightDir() );
		}
		else
		{
			sRollFraction = 0.0f;  // moved to deadzone
		}
	}
	// -------------------------------------------------------------------------
	// Pitch:
	if ( mAxisMap.IfPitch( event ) )
	{
		if ( IF_IN_AXIS_RANGE( mCalibration.mPitchDown ) )
		{
			sPitchFraction = ComputeFraction( event, AXIS_PITCH, mCalibration.mPitchDown, control.GetPitchDownDir() );
		}
		else if ( IF_IN_AXIS_RANGE( mCalibration.mPitchUp ) )
		{
			sPitchFraction = ComputeFraction( event, AXIS_PITCH, mCalibration.mPitchUp, control.GetPitchUpDir() );
		}
		else
		{
			sPitchFraction = 0.0f;  // moved to deadzone
		}
	}
	// -------------------------------------------------------------------------
	// Yaw:
	if ( mAxisMap.IfYaw( event ) )
	{
		if ( IF_IN_AXIS_RANGE( mCalibration.mYawLeft ) )
		{
			sYawFraction = ComputeFraction( event, AXIS_YAW, mCalibration.mYawLeft, control.GetYawLeftDir() );
		}
		else if ( IF_IN_AXIS_RANGE( mCalibration.mYawRight ) )
		{
			sYawFraction = ComputeFraction( event, AXIS_YAW, mCalibration.mYawRight, control.GetYawRightDir() );
		}
		else
		{
			sYawFraction = 0.0f;  // moved to deadzone
		}
	}
	// -------------------------------------------------------------------------
	// Throttle:
	// Different since throttle has no center deadzone.
	if ( mAxisMap.IfThrottle(event) )
	{
		// Some ThrottleRange possibilities:
		// ThrottleRange 0 32767
        // ThrottleRange 32767 0  # min > max (reverse throttle's direction)
		fp d, fraction;
		if ( mCalibration.mThrottle.mMin < mCalibration.mThrottle.mMax )
		{
			d = mCalibration.mThrottle.mMax - mCalibration.mThrottle.mMin;
			fraction = (event.mAxisVal - mCalibration.mThrottle.mMin) / d;
		}
		else  // min < max (backwards)
		{
			d = mCalibration.mThrottle.mMin - mCalibration.mThrottle.mMax;
			fraction = 1.0f - (event.mAxisVal - mCalibration.mThrottle.mMax) / d;
		}

		// Clamp fraction (axis value could be outside calibration range).
		fraction = Range<fp>( fraction, 0.0f, 1.0f );

		// Set Craft's throttle.
		craft->SetThrottle( fraction );
	}

	// Although one axis event happens at a time,
	// previous axis movements remain in effect
	// until the axis is moved to its deadzone.
	// Rotate the current craft (for stability, ignore small rotations.
    if ( ABS(sRollFraction)  > control::defs::JOYSTICK_MIN_FRACTION ) craft->PhysicalRotate( AXIS_ROLL,  sRollFraction,  CONTROL_CONF.mJoystickTickFreq );
    if ( ABS(sPitchFraction) > control::defs::JOYSTICK_MIN_FRACTION ) craft->PhysicalRotate( AXIS_PITCH, sPitchFraction, CONTROL_CONF.mJoystickTickFreq );
    if ( ABS(sYawFraction)   > control::defs::JOYSTICK_MIN_FRACTION ) craft->PhysicalRotate( AXIS_YAW,   sYawFraction,   CONTROL_CONF.mJoystickTickFreq );
}

/*****************************************************************************
 * Process a joystick button event.
 *****************************************************************************/
void
Joystick::ProcessButtonEvent( const input::Joystick::Event& event )
{
	// This is updated only if buttons changed or autofire.
	PERSISTENT input::Joystick::Event sPrevEvent;

	// Ignore superfluous buttons.
	if ( event.mButton < MAX_BUTTON )
	{
		// Ignore if no buttons changed and autofire disabled.
		if ( mAutofire or not sPrevEvent.IfEqualButtons(event) )
		{
			sPrevEvent = event;

			// Broadcast joystick button to listeners.
			mEventButton.Broadcast( event );
		}
	}
}

/*****************************************************************************
 * @return Fraction of axis movement {0.0,..,+-1.0}.  May return negative.
 * @param	event
 * @param	axis
 *			AXIS_ROLL/PITCH/YAW
 * @param	axisRange
 *			Calibration.AxisRange
 * @param	dir
 *			+-1 which can reverse direction of axis.
 *****************************************************************************/
fp
Joystick::ComputeFraction( const input::Joystick::Event& event, const uint axis, const Calibration::AxisRange& axisRange, const int dir )
{
ASSERT_AXIS( uint(axis) < uint(MAX_AXIS) );
ASSERT( dir == 1 or dir == -1 );
ASSERT( (axisRange.mMax | axisRange.mMin) != 0 );

	// Compute the axis movement as a fraction {0.0,..,1.0}.
	// This fraction is the level of control input that
	// partially determines the degree to rotate the current Craft.
	fp fraction;
	if ( ABS(event.mAxisVal) < ABS(axisRange.mMax) )
		fraction = ABS( fp(event.mAxisVal) / fp(axisRange.mMax) );
	else
		fraction = 1.0f;  // impose ceiling
	fraction *= dir;  // axis direction
	ASSERT( ABS(fraction) < 1.05f );  // ceiling should be 1.0
	return fraction;
}

/*****************************************************************************
 * Enable/disable autofire.
 *****************************************************************************/
void
Joystick::EnableAutofire( bool f )
{
	mAutofire = f;
}

/*****************************************************************************
 * @return Axis value.
 * For clients (not used internally).
 *****************************************************************************/
int
Joystick::GetAxisValue( uint axis )
{
	// mAxisValues[] is updated by axis events.
    if ( EX( axis < mAxisValues.size() ) )
	{
        return mAxisValues[axis];
	}
    else
	{
		ASSERT(0);
        return 0;  // oops
	}
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////  Joystick::AxisMap  ///////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor.
 *****************************************************************************/
Joystick::AxisMap::AxisMap( void )
{
	// Identity map.
	mAxises.resize( Joystick::MAX_AXIS );
	for ( uint i = 0; i < mAxises.size(); ++i )
		mAxises[i] = i;
}

/*****************************************************************************
 * @return True if joystick event struct matches the aircraft control.
 *****************************************************************************/
bool
Joystick::AxisMap::IfRoll( const input::Joystick::Event& event )
{
	return event.mAxis == mAxises[ROLL];
}

bool
Joystick::AxisMap::IfPitch( const input::Joystick::Event& event )
{
	return event.mAxis == mAxises[PITCH];
}

bool
Joystick::AxisMap::IfYaw( const input::Joystick::Event& event )
{
	return event.mAxis == mAxises[YAW];
}

bool
Joystick::AxisMap::IfThrottle( const input::Joystick::Event& event )
{
	return event.mAxis == mAxises[THROTTLE];
}

} // namespace control
