/*****************************************************************************
 * $LastChangedDate: 2008-09-13 22:52:35 -0400 (Sat, 13 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	ParticleSystem and Particle classes.
 * @remarks ParticleSystem is generalized.
 *			ParticleSystemFX is a derivative meant to be contained by the FX singleton.
 *//*
 * LEGAL:	COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define FX_PARTSYS_CC 1
#include "base/module.hh"
#include "base/conf.hh"
using namespace base;
#include "object/module.hh"
using namespace object;
#include "world/module.hh"
using namespace world;
#include "fx/module.hh"
#include "fx/partsys.hh"
#include "fx/defs.hh"

namespace fx {

#if DEBUG
INTERN int sParticleCnt = 0;  // in all particle-systems
#endif

// Detach particle while iterating thru particle container.
// STL idiom for associative (not sequential) containers.
#define DETACH_PARTICLE_INC_ITER(ITER)						\
{{															\
	Particles::iterator iter2 = ITER;						\
	++iter2;												\
	DetachParticle( ITER );									\
	ITER = iter2;											\
}}

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  Particle  ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 * Particle is based on Sprite.
 * Passing Sprite::Args struct is for speed/terseness.
 *****************************************************************************/
Particle::Particle( const Sprite::Args& spriteArgs,
                    const Milliseconds lifetime )
:	Sprite(spriteArgs),
	mTimeOfDeath(GET_TIMER().GetElapsedTime() + lifetime)
{
ASSERT( lifetime > Milliseconds(0) );

	SET_TYPESIG(this,TYPESIG_PARTICLE);

#if DEBUG
	// Limit amount of particles in all particle-systems.
	++sParticleCnt;
	ASSERT( uint(sParticleCnt) < fx::defs::MAX_PARTICLES_ALL_SYSTEMS );
	ASSERT( sParticleCnt > 0 );
#endif

#if DEBUG_PARTSYS
	CDEBUG << "new Particle lifetime=" << uint(lifetime) << " total=" << sParticleCnt << endl;
#endif
}

Particle::~Particle()
{
#if DEBUG
	--sParticleCnt;
	ASSERT( sParticleCnt >= 0 );  // catch double-deletion
#endif

#if DEBUG_PARTSYS
	CDEBUG << "delete Particle total=" << sParticleCnt << endl;
#endif

	INVALIDATE_TYPESIG(this,TYPESIG_PARTICLE);
}

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////  ParticleSystem  ////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 * @param	particleStateBits
 *			For state-sorting.
 * @param	timerEnabled
 *			Faster speed if this can be disabled.
 *****************************************************************************/
ParticleSystem::ParticleSystem( const StateBits particleStateBits,
                                const bool timerEnabled )
:	mParticleStateBits(particleStateBits),
	mParticles(),
	mTimerEnabled(timerEnabled),
	mTimerFunctor(new ParticleSystem::TimerFunctor(*this))
{
	SET_TYPESIG(this,TYPESIG_PARTICLE_SYSTEM);

	if ( mTimerEnabled )
		Timer::GetInstance().Register( mTimerFunctor.PTR(), fx::defs::PARTICLE_SYSTEM_TICK_FREQ );
}

ParticleSystem::~ParticleSystem()
{
#if DEBUG_PARTSYS
	CDEBUG << "~ParticleSystem()" << endl;
#endif

	// Detach all remaining Particles out of World.
	for ( Particles::iterator iter = mParticles.begin();
          iter != mParticles.end();
          ++iter )
	{
		CHECK_TYPESIG(*iter,TYPESIG_PARTICLE);		// *iter = shptr<Particle>
		gWorld.Detach( (*iter).PTR() );				// detach from World
	}

	// Must unregister!  Else functor would access a destroyed object.
	if ( mTimerEnabled )
		Timer::GetInstance().Unregister( mTimerFunctor.PTR() );

	INVALIDATE_TYPESIG(this,TYPESIG_PARTICLE_SYSTEM);
}

/*****************************************************************************
 * Attach particle (defaults).
 *****************************************************************************/
void
ParticleSystem::AttachParticle( shptr<Particle> particle )
{
	AttachParticle( particle, mParticleStateBits );
}

/*****************************************************************************
 * Attach particle with specified graphical state for state-sorting.
 * @param	particle
 * @param	stateBits
 * @param	stateSet
 *			Optional.
 *****************************************************************************/
void
ParticleSystem::AttachParticle( shptr<Particle> particle,
                                const StateBits stateBits,
                                RefPtr<osg::StateSet> stateSet )
{
CHECK_TYPESIG(this,TYPESIG_PARTICLE_SYSTEM);
CHECK_TYPESIG(particle,TYPESIG_PARTICLE);
#if DEBUG_PARTSYS
	CDEBUG << "ParticleSystem::AttachParticle()" << endl;
#endif

	// Add Particle to ParticleSystem container.
	mParticles.push_back( particle );  // so that the oldest are first (queue)

	// Attach Particle to scene-graph.
	gWorld.Attach( particle.PTR(), stateBits, stateSet );

ASSERT( mParticles.size() <= fx::defs::MAX_PARTICLES_ONE_SYSTEM );
}

/*****************************************************************************
 * Remove oldest particles.
 *****************************************************************************/
void
ParticleSystem::DetachOldestParticles( const uint cnt )
{
CHECK_TYPESIG(this,TYPESIG_PARTICLE_SYSTEM);
ASSERT( cnt < fx::defs::MAX_PARTICLES_ONE_SYSTEM );

	// The oldest particle is the head of the STL list.
	// Quickest way is to use one loop.
	// list::erase(first,last) and distance() might cause two loops.
	Particles::iterator iter = mParticles.begin();
	for ( uint i = 0; i < cnt; ++i )
	{
		if ( EX( iter != mParticles.end() ) )
		{
			DETACH_PARTICLE_INC_ITER( iter );
		#if DEBUG_PARTSYS
			CDEBUG << "detaching oldest particle" << endl;
		#endif
		}
		else
		{
			// Oops, amount to detach > size of container.
			ASSERT(false);
			break;
		}
	}
}

/*****************************************************************************
 * (private) Detach particle.
 * NOTE: Invalidates iter!
 *****************************************************************************/
void
ParticleSystem::DetachParticle( Particles::iterator iter )
{
CHECK_TYPESIG(this,TYPESIG_PARTICLE_SYSTEM);

	shptr<Particle> particle = *iter;   // keep holding a reference to it
	CHECK_TYPESIG(particle,TYPESIG_PARTICLE);
	gWorld.Detach( particle.PTR() );    // detach from World
	mParticles.erase( iter );			// erase from container
}

/*****************************************************************************
 * Tick.
 *****************************************************************************/
void
ParticleSystem::Tick( const Milliseconds millisecElapsed )
{
CHECK_TYPESIG(this,TYPESIG_PARTICLE_SYSTEM);
#if DEBUG_PARTSYS
	CDEBUG << "ParticleSystem::Tick() mParticles.size() " << mParticles.size() << endl;
#endif

	// Age and animate the particles.
	for ( Particles::iterator iter = mParticles.begin();
          iter != mParticles.end();
          /* NOP */ )
	{
		CHECK_TYPESIG(*iter,TYPESIG_PARTICLE);  // *iter = shptr<Particle>
		if ( millisecElapsed > (*iter)->mTimeOfDeath )  // past lifetime?
		{
			// Bury this particle.
			DETACH_PARTICLE_INC_ITER( iter );
		}
		else
		{
			// Animate this living particle.
			(*iter)->Tick( millisecElapsed );
			++iter;
		}
	}
}

// Reflect.
void
ParticleSystem::TimerFunctor::operator()( const Milliseconds millisecElapsed )
{
	mParticleSystem.Tick( millisecElapsed );
}

} // namespace fx
