/*****************************************************************************
 * $LastChangedDate: 2008-09-13 22:52:35 -0400 (Sat, 13 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	ParticleSystem and Particle classes.
 *//*
 * LEGAL:	COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef FX_PARTSYS_HH
#define FX_PARTSYS_HH 1

#if DEBUG && ! defined(DEBUG_PARTSYS)
//#define DEBUG_PARTSYS 1  // enables logging
#endif

#include "base/timer.hh"
#include "fx/sprite.hh"

namespace fx {

class ParticleSystem;

////////////////////////////////////////////////////////////////////////////////
/// @brief Particle.
///
/// Particle contains a Sprite.
/// ParticleSystem gives a pulse to Particles but doesn't animate them directly.
/// Particle::Tick() is meant to define a particle's animation.
///
class Particle : public Sprite
{
PREVENT_COPYING(Particle)
friend class ParticleSystem;
public:
	CLASS_CONST Milliseconds::Int UNLIMITED_LIFETIME = 100 * MILLION;

					Particle( const Sprite::Args& spriteArgs,
			                  const Milliseconds lifetime );
	virtual			~Particle();
	virtual void	Tick( const Milliseconds millisecElapsed ) = 0;

protected:
	Milliseconds	mTimeOfDeath;	///< when to destroy the particle
public:
	DECL_TYPESIG(TYPESIG_PARTICLE)
};

////////////////////////////////////////////////////////////////////////////////
/// @brief Particle system.
///
/// Features of particle system:
/// - ParticleSystem can be derived and Tick() extended
/// - pulsed by timer-tick
/// - adds particles
/// - ages particles
/// - deletes particles past their lifetime
/// - calls Particle::Animate() for optional animation of particles
///
/// ParticleSystem itself doesn't age.
/// Client should eventually destroy ParticleSystem.
///
/// To create a new particle system:
/// - Make a derivative of ParticleSystem that redefines Tick()
///   to add more particles every tick.
/// - Make a derivative of Particle which is passed to
///   DerivedParticleSystem::AddParticle().
///
class ParticleSystem : public Shared
{
PREVENT_COPYING(ParticleSystem)
protected:
	typedef list<shptr<Particle> > Particles;	///< must be a list because of how expired Particles are removed

protected:
	/// For state-sorting.
	CLASS_CONST StateBits DEFAULT_STATE_BITS = \
	STATE_BIT_BLEND | STATE_BIT_TRANSPARENT | STATE_BIT_NO_DEPTH_TEST | STATE_BIT_NO_LIGHTING | STATE_BIT_NO_FOG;

public:
	enum { TIMER_DISABLED = false, TIMER_ENABLED = true };
						ParticleSystem( const StateBits particleStateBits = DEFAULT_STATE_BITS,
                                        const bool timerEnabled = TIMER_ENABLED );
	virtual				~ParticleSystem();

protected:
	virtual void		Tick( const Milliseconds millisecElapsed );
	uint				GetParticleCnt( void ) { return mParticles.size(); }
	void				AttachParticle( shptr<Particle> particle );
	void				AttachParticle( shptr<Particle> particle,
                                        const StateBits stateBits,
                                        RefPtr<osg::StateSet> stateSet = NULL );
	void				DetachOldestParticles( const uint cnt );

private:
	void				DetachParticle( Particles::iterator iter );

private:
	////////////////////////////////////////////////////////////////////////////
	/// @brief Pulsed by timer-tick to animate particle system.
	///
	class TimerFunctor : public Timer::Functor
	{
	public:
		TimerFunctor( ParticleSystem& partsys ) : mParticleSystem(partsys) { }
		void operator()( const Milliseconds millisecElapsed );
	private:
		ParticleSystem&	mParticleSystem;
	};

protected:
	const StateBits			mParticleStateBits;		///< for state-sorting
	Particles				mParticles;				///< container of Particles
	bool					mTimerEnabled;			///< timer-tick can be disabled
	shptr<TimerFunctor>		mTimerFunctor;			///< pulses animation of particle system
public:
	DECL_TYPESIG(TYPESIG_PARTICLE_SYSTEM)
};

} // namespace fx

#endif // FX_PARTSYS_HH
