/*****************************************************************************
 * $LastChangedDate: 2008-10-20 10:36:48 -0400 (Mon, 20 Oct 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Game-related HUD widgets.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define GAME_HUD_CC 1
#include "base/module.hh"
#include "base/timer.hh"
using namespace base;
#include "object/module.hh"
using namespace object;
#include "hud/module.hh"
using namespace hud;
#include "game/module.hh"
#include "game/hud.hh"
#include "game/defs.hh"

namespace game {

INTERN const int  TARGET_BOX_WIDTH			= 60;  // uint causes sign bug!
INTERN const int  TARGET_BOX_HEIGHT			= 60;
INTERN const fp   TARGET_BOX_FONT_SIZE		= 20.0f;
INTERN const int  TARGET_BOX_TEXT_OFFSET_X	= -TARGET_BOX_WIDTH / 2 - 2;   // offset to avoid
INTERN const int  TARGET_BOX_TEXT_OFFSET_Y	= -TARGET_BOX_WIDTH / 2 - 16;  // obscuring 3D object
INTERN const RGBA TARGET_BOX_COLOR_DESELECTED( 0.8f, 0.8f, 0.1f, 1.0f );  // yellow
INTERN const RGBA TARGET_BOX_COLOR_SELECTED(   0xff, 0x99, 0x00, 0xff );  // orange
INTERN const RGBA TARGET_BOX_COLOR_RADAR_LOCK( 1.0f, 0.4f, 0.4f, 1.0f );  // red

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  game::Hud  //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

Hud* Hud::msInstance;

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Hud::Hud( void )
:	mTickFunctor(new TickFunctor),
	mDummyTarget(Object::MakeDummyObject()),
	mSelectedTarget(mDummyTarget),  // -- ORDER-DEPENDENT --
	mTargetIndicatorMap()
{
ASSERT_RUN_ONCE;

	Timer::GetInstance().Register( mTickFunctor, game::defs::GAME_TICK_FREQ );
}

Hud::~Hud()
{
	Timer::GetInstance().Unregister( mTickFunctor );
}

/*****************************************************************************
 * Updates the target boxes.
 *****************************************************************************/
void
Hud::Tick( const Milliseconds millisecElapsed )
{
	// Update HUD with positions of targets.
	for ( TargetIndicatorMap::iterator iter = mTargetIndicatorMap.begin();
          iter != mTargetIndicatorMap.end();
          ++iter )
	{
		Object&  object  = *iter->first;           CHECK_TYPESIG(&object,TYPESIG_OBJECT);
		HudBox&  hudBox  = *iter->second.mHudBox;  CHECK_TYPESIG(&hudBox,TYPESIG_HUD_BOX);
		HudText& hudText = *iter->second.mHudText; CHECK_TYPESIG(&hudText,TYPESIG_HUD_TEXT);
		hudBox.Set3D( object.GetPosition(), TARGET_BOX_WIDTH, TARGET_BOX_HEIGHT );
		hudText.SetPosition3D( object.GetPosition() );
	}
}

// Reflect.
void
Hud::TickFunctor::operator()( const Milliseconds millisecElapsed )
{
	GET_GAME_HUD().Tick( millisecElapsed );
}

/*****************************************************************************
 * Attach target.
 * Represent this in the HUD as a "target indicator box".
 *****************************************************************************/
void
Hud::AttachTarget( shptr<Object> target, const string& text )
{
CHECK_TYPESIG(target,TYPESIG_OBJECT);

	// Does this have a target-indicator box in HUD?
	if ( mTargetIndicatorMap.find( target ) == mTargetIndicatorMap.end() )
	{
		// Compose TargetIndicator with a HudBox and HudText.
		shptr<HudBox> hudBox = new HudBox;
		hudBox->SetColor( TARGET_BOX_COLOR_DESELECTED );

		HudText::Args hudTextArgs;
		hudTextArgs.mText       = text;
		hudTextArgs.mColor      = TARGET_BOX_COLOR_DESELECTED;
		hudTextArgs.mFontSize   = TARGET_BOX_FONT_SIZE;
		hudTextArgs.mOffset[XX] = TARGET_BOX_TEXT_OFFSET_X;  // offset to avoid
		hudTextArgs.mOffset[YY] = TARGET_BOX_TEXT_OFFSET_Y;  // obscuring 3D object
		shptr<HudText> hudText = new HudText( hudTextArgs );

		// Map TargetIndicator.
		mTargetIndicatorMap.insert( make_pair<>( target, TargetIndicator(hudBox,hudText) ) );

		// Attach TargetIndicator to underlying HUD.
		hud::Hud::GetInstance().Attach( hudBox );
		hud::Hud::GetInstance().Attach( hudText->GetNode() );
	}
}

/*****************************************************************************
 * Detach target.
 *****************************************************************************/
void
Hud::DetachTarget( shptr<Object> target )
{
CHECK_TYPESIG(target,TYPESIG_OBJECT);

	// Is this the selected target?
	if ( mSelectedTarget == target )
		mSelectedTarget = mDummyTarget;

	// Remove from map.
	mTargetIndicatorMap.erase( target );
}

/*****************************************************************************
 * Select a target.  Represented in HUD as a "target indicator box".
 *****************************************************************************/
void
Hud::SelectTarget( shptr<Object> target, const bool radarLock )
{
CHECK_TYPESIG(target,TYPESIG_OBJECT);

#define SELECT_TARGET( TARGET, COLOR, RADAR_LOCK )												\
{{																								\
	TargetIndicatorMap::iterator iter = mTargetIndicatorMap.find( TARGET );						\
	if ( EX( iter != mTargetIndicatorMap.end() ) )												\
	{																							\
		HudBox& hudBox = *iter->second.mHudBox; CHECK_TYPESIG(&hudBox,TYPESIG_HUD_BOX);			\
		HudText& hudText = *iter->second.mHudText; CHECK_TYPESIG(&hudText,TYPESIG_HUD_TEXT);	\
		hudBox.SetColor( COLOR );																\
		hudBox.SetDrawX( RADAR_LOCK );															\
		hudText.SetColor( COLOR );																\
	}																							\
}}

	// Deselect previous target.
	if ( mSelectedTarget != mDummyTarget )
	{
		SELECT_TARGET( mSelectedTarget, TARGET_BOX_COLOR_DESELECTED, false );
	}

	// Select next target.
	mSelectedTarget = target;
	SELECT_TARGET( target,
                   radarLock ? TARGET_BOX_COLOR_RADAR_LOCK : TARGET_BOX_COLOR_SELECTED,
                   radarLock );
}

} // namespace game
