/*****************************************************************************
 * $LastChangedDate: 2008-10-21 13:59:37 -0400 (Tue, 21 Oct 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Graph class.
 *//*
 * LEGAL:	COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define GRAPH_GRAPH_CC 1
#include "base/module.hh"
#include "base/conf.hh"
using namespace base;
#include "types/module.hh"
#include "types/rgba.hh"
#include "types/rgba_colors.hh"
using namespace types;
#include "object/module.hh"
using namespace object;
#include "world/module.hh"
#include "world/conf.hh"
using namespace world;
#include "graph/module.hh"
#include "graph/graph.hh"

#include <osgUtil/Optimizer>
#include <osgDB/WriteFile>

namespace graph {

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  functions  ////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Insert a group node between two others.
 * All must be group nodes to continue the chain of nodes.
 * Before insertion: parent --> oldChild
 * After insertion:  parent --> newChild --> oldChild
 *****************************************************************************/
void InsertGroupNode( RefPtr<GroupNode> parent,
                      RefPtr<GroupNode> newChild,
                      RefPtr<Node>      oldChild )
{
ASSERT( parent->containsNode( oldChild.get() ) );
ASSERT( not parent->containsNode( newChild.get() ) );

	newChild->addChild(  oldChild.get() );  // hold reference to old child
	parent->removeChild( oldChild.get() );  // parent becomes grandparent
	parent->addChild(    newChild.get() );  // parent has new child
}

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  Graph  //////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *
 * SUBTLE: Another transform node can be passed to the ctor as the root node.
 * Having two transform nodes is useful, since it allows setting
 * the correct orientation of a 3D model, which then is transformed
 * by the usual transform node.
 *****************************************************************************/
Graph::Graph( RefPtr<Node> rootNode )
:	mSwitchNode(new osg::Switch),
	mTransformNode(new osg::MatrixTransform),
	mStateSet(new osg::StateSet)
{
	SET_TYPESIG(this,TYPESIG_GRAPH);

	// Prevent side-effects from OSG's optimizer such as FLATTEN_STATIC_TRANSFORMS.
	rootNode->setDataVariance( osg::Object::DYNAMIC );
	mSwitchNode->setDataVariance( osg::Object::DYNAMIC );
	mTransformNode->setDataVariance( osg::Object::DYNAMIC );
	mStateSet->setDataVariance( osg::Object::DYNAMIC );

	// Attach stateset to top node.
	mSwitchNode->setStateSet( mStateSet.get() );

	// Splice a transform node.
	mTransformNode->addChild( rootNode.get() );

	// Splice a Switch node.
	// The switch node is used to enable/disable rendering graph.
	mSwitchNode->addChild( mTransformNode.get() );

	// Solves problem with lighting in 3D models that are scaled.
	// http://www.3drealtimesimulation.com/osg/osg_faq_1.htm#f19
	mStateSet->setMode( GL_RESCALE_NORMAL, osg::StateAttribute::ON );
}

Graph::~Graph()
{
	INVALIDATE_TYPESIG(this,TYPESIG_GRAPH);
}

/*****************************************************************************
 * Make a minimal-but-usable Graph for use with an Object.
 *****************************************************************************/
shptr<Graph>
Graph::MakeMinimalGraph( void )  // CLASS_METHOD
{
	// This must create a new Graph as some Objects build the Graph with more nodes.
	return new Graph( new Node );
}

/*****************************************************************************
 * Attach a node to Graph.
 *****************************************************************************/
void
Graph::AttachNode( RefPtr<Node> node )
{
CHECK_TYPESIG(this,TYPESIG_GRAPH);
ASSERT( node.get() != mTransformNode.get() );  // prevent cyclic graph!

    // Attach as a child of transform node.
    mTransformNode->addChild( node.get() );
}

/*****************************************************************************
 * Insert/splice a group node into the Graph.
 * Must be a group to continue the chain of nodes.
 *****************************************************************************/
void
Graph::InsertGroupNode( RefPtr<GroupNode> group )
{
CHECK_TYPESIG(this,TYPESIG_GRAPH);

	ASSERT( mTransformNode->getNumParents() > 0 );
	graph::InsertGroupNode( mTransformNode->getParent(0),  // parent
                            group,                         // new child
                            mTransformNode.get() );        // old child
}

/*****************************************************************************
 * Enable/disable rendering a Graph.
 *
 * NOTE: This isn't effective for enabling models per-view in OSG.
 * The reason is that OSG generates a render list from the scene graph,
 * then renders every camera.  So trying to alter nodes is too late in the pipeline.
 * Rather, use an "cull mask" for the osg::Camera with a "node mask" for the model.
 * See View::ExcludeFromMainView().
 *****************************************************************************/
void
Graph::Enable( const bool enable )
{
CHECK_TYPESIG(this,TYPESIG_GRAPH);

	GraphEnablement::Enable( *mSwitchNode, enable );  // private base class
}

bool
Graph::IfEnabled( void ) const
{
CHECK_TYPESIG(this,TYPESIG_GRAPH);
ASSERT( mSwitchNode->getNumChildren() == 1 );

	return GraphEnablement::IfEnabled( *mSwitchNode );  // private base class
}

/*****************************************************************************
 * LIMITED-USE.
 * Rather, use World::Attach(object,stateBits) for state-sorting.
 *****************************************************************************/
void
Graph::EnableState( const bool enable, const osg::StateAttribute::GLMode glState )
{
CHECK_TYPESIG(this,TYPESIG_GRAPH);

	GetStateSet()->setMode( glState,
                            enable ? osg::StateAttribute::ON
                                   : osg::StateAttribute::OFF | osg::StateAttribute::OVERRIDE );
}

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////  SceneGraph  ////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
SceneGraph::SceneGraph( void )
:	mSuperRootGroup(new osg::ClearNode),  // ClearNode is a Group
	mRootGroup(new osg::Group),
	mStateSet(new osg::StateSet)
{
	// Arrange top nodes of scene graph.
	mSuperRootGroup->addChild( mRootGroup.get() );

	// Stateset of entire scene graph.
	mSuperRootGroup->setStateSet( mStateSet.get() );

	// Set the default clear color.
	SetClearColor( RGBA( RGB_SKY_BLUE, 1.0f ) );
}

SceneGraph::~SceneGraph()
{
	// NOP
}

/*****************************************************************************
 * Directly attach node to scene graph.
 * LIMITED-USE:
 * Use World::Attach() instead.
 *****************************************************************************/
void
SceneGraph::AttachNode( RefPtr<Node> node )
{
	mRootGroup->addChild( node.get() );
}

/*****************************************************************************
 * Insert/splice a group node between the SceneGraph's super-root and root nodes.
 * The new child must be a group to continue the chain of nodes.
 *****************************************************************************/
void
SceneGraph::InsertGroupNode( RefPtr<GroupNode> group )
{
	// Insert between the usual root group and and its parent.
	//                         parent     --> new child --> old child
	// mSuperRootGroup ..-->.. mRootGroup --> group     --> mRootGroup's old child
	ASSERT( mRootGroup->getNumParents() > 0 );
	graph::InsertGroupNode( mRootGroup->getParent(0),  // parent
                            group,                     // new child
                            mRootGroup.get() );        // old child
}

/*****************************************************************************
 * Run OSG optimizer.
 * REMOVE_REDUNDANT_NODES breaks this app's ability to find subgraphs in 3D models.
 *****************************************************************************/
void
SceneGraph::OptimizeNode( RefPtr<Node> node )
{
	if ( WORLD_CONF.mEnableOsgOptimizer )
	{
	    osgUtil::Optimizer optimizer;
		const uint optimizerOptions = osgUtil::Optimizer::FLATTEN_STATIC_TRANSFORMS
	                              //| osgUtil::Optimizer::REMOVE_REDUNDANT_NODES
	                                | osgUtil::Optimizer::COMBINE_ADJACENT_LODS
	                                | osgUtil::Optimizer::SHARE_DUPLICATE_STATE
	                                | osgUtil::Optimizer::MERGE_GEOMETRY
	                                | osgUtil::Optimizer::SPATIALIZE_GROUPS
	                                | osgUtil::Optimizer::TRISTRIP_GEOMETRY
	                                | osgUtil::Optimizer::OPTIMIZE_TEXTURE_SETTINGS
	                                | osgUtil::Optimizer::MERGE_GEODES;
	    optimizer.optimize( node.get(), optimizerOptions );
	}
}

/*****************************************************************************
 * Optimize entire scene-graph.
 *****************************************************************************/
void
SceneGraph::OptimizeSceneGraph( void )
{
	OptimizeNode( GetSuperRootNode() );
}

/*****************************************************************************
 * Set clear/background color.
 *****************************************************************************/
void
SceneGraph::SetClearColor( const RGBA color )
{
	// mSuperRootGroup doubles as a osg::ClearNode (group despite its name).
	mSuperRootGroup->setClearColor( RGBA2VEC4F(color) );
}

/*****************************************************************************
 * Dump scene graph to file.
 *****************************************************************************/
void
SceneGraph::DumpToFile( const string& filename )
{
	COUT << "Dumping scene graph to file " << filename << "...";
	if ( osgDB::writeNodeFile( *GetSuperRootNode(), filename ) )
		COUT << "ok." << endl;
	else
		CERROR << "ERROR" << endl;
}

} // namespace graph
