/*****************************************************************************
 * $LastChangedDate: 2008-09-13 12:35:14 -0400 (Sat, 13 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Graph class.
 *//*
 * LEGAL:	COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef GRAPH_GRAPH_HH
#define GRAPH_GRAPH_HH 1

// GET_SCENE_GRAPH() is defined at view/window.hh

#include <osg/Switch>
#include "types/rgba.hh"
using namespace types;
#include "object/rotation_list.hh"
using namespace object;
#include "graph/graph_enablement.hh"

namespace graph {

// Functions:
void InsertGroupNode( RefPtr<GroupNode> parent,
                      RefPtr<GroupNode> newChild,
                      RefPtr<Node>      oldChild );

////////////////////////////////////////////////////////////////////////////////
/// @brief Graph class (per-object graph)
///
/// This can be thought of as ObjectGraph.
/// Graph defines the set of nodes of one individual Object.
///
/// Graph arrangement (OSG):
/// osg::Switch  -->  osg::Transform  -->  rootNode
///
/// The Switch and Transform nodes belong to Graph.
/// "root node" (for lack of a better name) is the root node
/// passed to the Graph ctor or AttachNode().
///
/// Graph and Object have their own separate Transform nodes.
/// Graph's Transform is used for orienting 3D models transparently
/// to an Object (Objects are attached to a Graph).
///
/// SUBTLE: Another transform node can be passed to the ctor as the root node.
/// Having two transform nodes is useful, since it allows setting
/// the correct orientation of a 3D model, which then is transformed
/// by the usual transform node.
///
class Graph : public Shared, private GraphEnablement
{
// Graph (ctor, dtor):
public:                             Graph( RefPtr<Node> rootNode );
public:                             ~Graph();

// Graph enablement:
public:  void                       Enable( const bool enable = true );  // enable rendering of Graph
public:	 bool                       IfEnabled( void ) const;

// Making graphs:
public:  CLASS_METHOD shptr<Graph>  MakeMinimalGraph( void );  // make a minimal-but-usable Graph

// Nodes:
// Use AttachNode() [instead of addChild() with GetRootNode()/GetTransformNode()].
// GetRootNode() must return the actual root node (this cannot be faked, if World::Attach()
// were to attach its child then the root node would be outside the scene graph).
// There is no separate Group node (it is really the Transform node),
// but accomodates clients that only need a Group node.
public:  void                       AttachNode( RefPtr<Node> node );
public:  void						InsertGroupNode( RefPtr<GroupNode> group );
public:  RefPtr<Node>               GetRootNode( void )      { CHECK_TYPESIG(this,TYPESIG_GRAPH); return mSwitchNode.get(); }
public:  RefPtr<TransformNode>      GetTransformNode( void ) { CHECK_TYPESIG(this,TYPESIG_GRAPH); return mTransformNode.get(); }
public:  RefPtr<osg::Group>         GetGroupNode( void )     { CHECK_TYPESIG(this,TYPESIG_GRAPH); return GetTransformNode().get(); }
public:  RefPtr<osg::StateSet>		GetStateSet( void )      { CHECK_TYPESIG(this,TYPESIG_GRAPH); return mStateSet; }

// LIMITED-USE (for state-sorting, use World::Attach(object,stateBits) instead).
private: void						EnableState( const bool enable, osg::StateAttribute::GLMode glState );  // OSG/OpenGL-specific

// Graph (data members):
private: RefPtr<osg::Switch>    mSwitchNode;		///< its child is what client passed as "root node"
private: RefPtr<TransformNode>	mTransformNode;		///< outer-level transform node
private: RefPtr<osg::StateSet>	mStateSet;
public:  DECL_TYPESIG(TYPESIG_GRAPH)  // final member
};

////////////////////////////////////////////////////////////////////////////////
/// @brief SceneGraph class (graph containing everything).
///
/// SuperRootGroup --> [inserted groups] --> RootGroup --> [3D objects]
///
/// This defines the top of the scene graph.
/// The limited-use "super root" is what is passed to OSG setSceneData().
/// The regular root group is where nodes are usually attached.
///
/// The separation between the two root nodes allows placing special nodes,
/// such as lighting, over the regular root node (the "working root node").
/// Hence there is a "super root" to contain those special nodes.
///
/// SceneGraph is a low-level class with limited-use.
/// The app should attach nodes of 3D models using the higher-level World class
/// which provides state-sorting.
///
class SceneGraph : public Shared
{
// Graph (ctor, dtor):
public:							SceneGraph( void );
public:							~SceneGraph();

// Graph (public interface):
// Subtle: GetSuperRootNode() doesn't return a group to encourage using GetRootGroup().
public:  RefPtr<Node>			GetSuperRootNode( void ) { return mSuperRootGroup.get(); }
public:  RefPtr<GroupNode>		GetRootGroup( void ) { return mRootGroup; }
public:  RefPtr<osg::StateSet>	GetRootStateSet( void ) { return mStateSet; }
public:	 void					AttachNode( RefPtr<Node> node );             // LIMITED-USE: use World::Attach() instead
public:  void					InsertGroupNode( RefPtr<GroupNode> group );  // LIMITED-USE
public:  void					OptimizeSceneGraph( void );                  // entire scene-graph
public:  void					OptimizeNode( RefPtr<Node> node );           // sub-graph
public:	 void					SetClearColor( const RGBA color );
public:  void					DumpToFile( const string& filename );

// Graph (data members):
private: RefPtr<osg::ClearNode>	mSuperRootGroup;	///< the absolute/private root
private: RefPtr<osg::Group>		mRootGroup;			///< the usual/public root
private: RefPtr<osg::StateSet>	mStateSet;			///< stateset of entire scene graph
};

} // namespace graph

#endif // GRAPH_GRAPH_HH
