/*****************************************************************************
 * $LastChangedDate: 2008-09-20 10:29:31 -0400 (Sat, 20 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Module configuration.
 * @remarks Most of the joystick configuration (axis3,calibration,autofire)
 *          is done by the higher-level "control" module.
 *//*
 * LEGAL:	COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define INPUT_CONF_CC 1
#include "base/module.hh"
#include "base/conf.hh"
using namespace base;
#include "input/module.hh"
#include "input/defs.hh"
#include "input/conf.hh"

namespace input {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  Conf  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Conf::Conf( void )
:	mValid(false),
	// cmd-line args 
	mJoystickArg(false),
	// .conf file
	mInputQueueTickFreq(defs::DEFAULT_INPUT_QUEUE_TICK_FREQ),
	mKeyboardInputQueueOverflow(0),
	mJoystickInputQueueOverflow(0)
{
}

/*****************************************************************************
 * Initialize/read configuration.
 *****************************************************************************/
void
Conf::Init( int argc, char** argv, const string& confTxtFilename )
{

BEGIN_ONCE {

	mValid = true;  // assign this first as *_CONF may expand to GetConf() which asserts it

	ParseCmdlineArgs( argc, argv );   // cmdline args parsed prior to reading .conf file
	ReadConfTxt( confTxtFilename );   // read from textual .conf file
	ParseCmdlineArgs2( argc, argv );  // cmdline args that override .conf file

} END_ONCE

}

/*****************************************************************************
 * Parse cmdline args.
 *****************************************************************************/
void
Conf::ParseCmdlineArgs( int argc, char** argv )
{
	// Parse args.
	for ( int i = 1; i < argc; ++i )
	{
		// control module processes all of the joystick args,
		// but input module just needs to know if any -j was passed.
		if ( (argv[i][0] == '-')
         and (argv[i][1] == 'j') )  // -j
		{
			mJoystickArg = true;
		}
	}
}

void
Conf::ParseCmdlineArgs2( int argc, char** argv )
{
    // NOP
}

/*****************************************************************************
 * Read from .conf file.
 *****************************************************************************/
void
Conf::ReadConfTxt( const string& confTxtFilename )
{
	try
	{
		ModuleConfTxt confTxt;
		confTxt.Read( string(base::GetDataDir()) + string(confTxtFilename) );
	}
	catch (...)
	{
		CERROR << "ERROR: cannot read " << confTxtFilename << endl;
		EXIT( 1 );
	}
}

void
Conf::ModuleConfTxt::ProcessField( string& field, ifstream& conf )
{
	bool error = false;

	// -----------------------  Timer  -----------------------------------------
	if ( field == "InputQueueTickFreq" )
	{
		int i; conf >> i;
		INPUT_CONF.mInputQueueTickFreq = Milliseconds(i);
		if ( i < 1 || i >= 1000 ) error = true;
	}
	else if ( field == "KeyboardInputQueueOverflow" )
	{
		conf >> INPUT_CONF.mKeyboardInputQueueOverflow;
	}
	else if ( field == "JoystickInputQueueOverflow" )
	{
		conf >> INPUT_CONF.mJoystickInputQueueOverflow;
	}

    // Errors?
	if ( error )
	{
		CERROR << "ERROR[conf]: parameter for " << field << " is invalid." << endl;
		EXIT( 1 );
	}
}

/*****************************************************************************
 * Print Conf configuration.
 *****************************************************************************/
ostream&
operator<<( ostream& strm, const Conf& conf )
{
	std::ios::fmtflags savedFlags = strm.flags();
	strm << "conf [input]:" << endl
         << "Key/joy overflow   =" <<        INPUT_CONF.mKeyboardInputQueueOverflow
                                   << ',' << INPUT_CONF.mJoystickInputQueueOverflow; //endl
	strm.flags(savedFlags);
	return strm;
}

} // namespace input
