/*****************************************************************************
 * $LastChangedDate: 2008-09-20 10:29:31 -0400 (Sat, 20 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Input event queue.
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define INPUT_QUEUE_CC 1
#include "base/module.hh"
using namespace base;
#include "input/module.hh"
#include "input/defs.hh"
#include "input/input_queue.hh"
#include "input/conf.hh"

namespace input {

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  InputQueue  ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

InputQueue* InputQueue::msInstance;  // there can be only one

/*****************************************************************************
 * ctor initially disables input queue.
 * When the app is ready, it should call InputQueue::Enable().
 *****************************************************************************/
InputQueue::InputQueue( void )
:	mEnabled(false),
    mWasStarted(false),
	mPollTickFreq(Milliseconds(INPUT_CONF.mInputQueueTickFreq))
{
ASSERT_RUN_ONCE;

	// Correct in case .conf wasn't read.
	if ( mPollTickFreq == Milliseconds(0) )
	{
		ASSERT(false);
		mPollTickFreq = Milliseconds(defs::DEFAULT_INPUT_QUEUE_TICK_FREQ);
	}
}

InputQueue::~InputQueue()
{
	// NOP
}

/*****************************************************************************
 * Start the input queue (initially disabled).
 * This is one-shot.  Tolerates being called more than once.
 * This should be called after the app is ready to respond to input events.
 *****************************************************************************/
void
InputQueue::Start( const bool enableJoystick )
{
    // This is one-shot.  Tolerates being called more than once.
	if ( not mWasStarted )
	{
		mWasStarted = true;							 	// start InputQueue
		mEnabled = true;                                // enabling implied

        Keyboard& keyboard = Keyboard::GetInstance();
        Joystick& joystick = Joystick::GetInstance();

	    keyboard.Enable();      						// enable Keyboard

		if ( enableJoystick )
		{
			joystick.Enable();                          // enable Joystick

		#if COMPILE_GLUT_JOYSTICK
			// If the current joystick driver failed, try GLUT.
			if ( not joystick.IfEnabled() )
			{
				CERROR << "Linux joystick driver unavailable, trying GLUT joystick." << endl;
			    joystick.SetJoystickDriver( Joystick::eJoystickDriver_GLUT );
				joystick.Enable();
				if ( not joystick.IfEnabled() )
				{
					CERROR << "GLUT joystick driver unavailable." << endl;
				}
			}
		#endif  // COMPILE_GLUT_JOYSTICK
		}

       	// Register timer-tick to pulse the input queue.
       	Timer::GetInstance().Register( new TimerFunctor(*this), mPollTickFreq );
	}
}

/*****************************************************************************
 * Enable/disable the input queue.
 * Can be used as part of pausing the simulation.
 *****************************************************************************/
void
InputQueue::Enable( const bool enable )
{
    // App should've called Start() prior but in case not.
    if ( enable and not mWasStarted )
	{
        Start( INPUT_CONF.mJoystickArg );  // if -j arg
		mWasStarted = true;
	}

    mEnabled = enable;
}

bool
InputQueue::IfEnabled( void )
{
    return mEnabled;
}

/*****************************************************************************
 * @return Amount of events in keyboard queue.
 *****************************************************************************/
uint
InputQueue::KeyQueueCnt( void )
{
	return mKeyQueue.size();
}

/*****************************************************************************
 * @return Amount of events in joystick queue.
 *****************************************************************************/
uint
InputQueue::JoyQueueCnt( void )
{
	return mJoyQueue.size();
}

/*****************************************************************************
 * Dequeue the next keyboard event.
 * @return False if no more keyboard events.
 *****************************************************************************/
bool
InputQueue::DequeueKeyEvent( Keyboard::Event& event/*OUT*/ )
{
	if ( mKeyQueue.size() )
	{
		event = mKeyQueue.front();
		mKeyQueue.pop_front();
		return true;
	}
	else
	{
		return false;
	}
}

/*****************************************************************************
 * Dequeue the next joystick event.
 * @return False if no more joystick events.
 *****************************************************************************/
bool
InputQueue::DequeueJoyEvent( Joystick::Event& event/*OUT*/ )
{
	if ( mJoyQueue.size() )
	{
		event = mJoyQueue.front();
		mJoyQueue.pop_front();
		return true;
	}
	else
	{
		return false;
	}
}

/*****************************************************************************
 * Flush all queues.
 *****************************************************************************/
void
InputQueue::Flush( void )
{
	mKeyQueue.clear();
	mJoyQueue.clear();
}

/*****************************************************************************
 * Pulsed by a timer tick to poll input devices.
 * Ticks feed the queue.
 *
 * Joy/key asymmetry:
 * Timer-tick is used to poll the joystick driver to enqueue events into InputQueue.
 * But Keyboard doesn't use timer-tick.
 * Instead, GLUT or OSG calls keyboard handlers which enqueue into InputQueue.
 *****************************************************************************/
void
InputQueue::Tick( void )
{
    // mEnabled=false can pause the input queue (and the simulation).
	if ( mEnabled )
	{
    	VentOverflowingInputQueue();

		Joystick::GetInstance().PollAndEnqueueEvents();
	}
}

void
InputQueue::TimerFunctor::operator()( const Milliseconds millisecElapsed )
{
    mInputQueue.Tick();  // reflect to InputQueue
}

/*****************************************************************************
 * Enqueue a keyboard event.
 *****************************************************************************/
void
InputQueue::EnqueueKeyEvent( const Keyboard::Event& event )
{
	if ( mEnabled ) mKeyQueue.push_back( event );
}

/*****************************************************************************
 * Enqueue a joystick event.
 *****************************************************************************/
void
InputQueue::EnqueueJoyEvent( const Joystick::Event& event )
{
	if ( mEnabled ) mJoyQueue.push_back( event );
}

/*****************************************************************************
 * Begin dropping events if an event queue is growing too quickly.
 * The joystick is more prone to causing this problem than the keyboard.
 *****************************************************************************/
void
InputQueue::VentOverflowingInputQueue( void )
{
	// 0 disables the vent.

    // Keyboard.
	if ( INPUT_CONF.mKeyboardInputQueueOverflow > 0 )
	{
		while ( KeyQueueCnt() > INPUT_CONF.mKeyboardInputQueueOverflow )
		{
			// Drop oldest keyboard event.
			Keyboard::Event event;
			DequeueKeyEvent( event /*OUT*/ );
		}
	}

    // Joystick.
	if ( INPUT_CONF.mJoystickInputQueueOverflow > 0 )
	{
		while ( JoyQueueCnt() > INPUT_CONF.mJoystickInputQueueOverflow )
		{
			// Drop oldest joystick event.
			Joystick::Event event;
			DequeueJoyEvent( event /*OUT*/ );
		}
	}
}

} // namespace input
