/*****************************************************************************
 * $LastChangedDate: 2008-02-25 03:32:37 -0500 (Mon, 25 Feb 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Input event queue.
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef INPUT_QUEUE_HH
#define INPUT_QUEUE_HH 1

#include "base/timer.hh"

namespace input {

////////////////////////////////////////////////////////////////////////////////
/// @brief Input event queue class.
/// @verbatim
///
/// Rationale:
/// ----------
/// Input events should be put into a queue and processed synchronously at
/// an appropriate time, rather than asynchronously.  The reason is that an
/// input event shouldn't be processed in the middle of drawing a view,
/// or worse, while processing an inappropriate view (such as an enemy view
/// which would affect the enemy rather than the player).
///
/// Integration:
/// ------------
/// InputQueue is initially disabled to give time to the app to become ready
/// to respond to input events.  When the app becomes ready, it should call Start().
///
/// How to use this class:
/// ----------------------
/// - InputQueue is a Singleton.
/// - When the app is ready, call InputQueue::Start().
/// - Input queue can be paused by Enable(false).
/// - Client can call InputQueue::*Events() to query
///   the amount of events in a single queue.
/// - Keyboard and joystick events can be dequeued out of separate queues
///   by calling DequeueKeyEvent() and DequeueJoyEvent().
///   The idea is for a client to poll input prior to drawing a frame.
/// - If events overflow, InputQueue can be configured to drop (vent) events.
///
/// Queues and queue sizes:
/// -----------------------
/// Each input device has its own queue.  This helps in the situation
/// where the application is able to keep pace with one input device but
/// another is being enqueued too quickly.  The application can process
/// events in one queue while dropping events in another queue.
///
/// Implementation:
/// ---------------
/// Enqueueing of keyboard and joystick events are similar but asymmetrical.
/// Both Keyboard and Joystick classes use InputQueue to queue events.
/// Keyboard events are pushed out by GLUT or OSG to callbacks (async I/O),
/// but joystick events must be dequeued by a timer-tick handler (polling I/O).
/// Polling isn't as bad in this case because the Linux joystick driver
/// internally queues joystick events (InputQueue forms a queue-over-a-queue).
///
/// Other sources of input and InputQueue:
/// --------------------------------------
/// Other sources of input that aren't enqueued in InputQueue (mouse events)
/// should disable themselves if InputQueue is disabled, as a disabled InputQueue
/// is a kind of input control flag.
///
/// @endverbatim
///
class InputQueue
{
PREVENT_COPYING(InputQueue)
private:
	InputQueue( void );
	~InputQueue();

private:
    /// @brief Pulses InputQueue.
	///
    class TimerFunctor : public base::Timer::Functor
    {
    public:
                TimerFunctor( InputQueue& inputQueue ) : mInputQueue(inputQueue) { }
        void    operator()( const Milliseconds millisecElapsed );
    private:
        InputQueue& mInputQueue;
    };

// Interface for application:
// Initially, call Start().  Thereafter, Enable(bool) can be called (to pause input).
public:
    CLASS_METHOD InputQueue& GetInstance( void )  // Singleton
	{
		if ( UX( msInstance == NULL ) )
			msInstance = new InputQueue;
		return *msInstance;
	}

	void			Start( const bool enableJoystick );  // one-shot
    void			Enable( const bool enable = true );
	bool			IfEnabled( void );
	Milliseconds	GetPollTickFreq( void ) { return mPollTickFreq; }
	uint			KeyQueueCnt( void );  // amount of events in queue
	uint			JoyQueueCnt( void );
	bool			DequeueKeyEvent( Keyboard::Event& event/*OUT*/ );
	bool			DequeueJoyEvent( Joystick::Event& event/*OUT*/ );
	void			Flush( void );  // flush all queues

// Interface for Keyboard/Joystick classes:
public:
	void			EnqueueKeyEvent( const Keyboard::Event& event );
	void			EnqueueJoyEvent( const Joystick::Event& event );

private:
	void			Tick( void );  // pulses InputQueue
    void			VentOverflowingInputQueue( void );

private:
	// queue<> unsuitable since it lacks erase()/clear() for flushing.
    CLASS_VAR InputQueue*   msInstance;
	bool					mEnabled;       ///< false if queue is disabled (paused)
    bool                    mWasStarted;    ///< true if Start() was ever called
	Milliseconds			mPollTickFreq;	///< frequency of polling input queue
	deque<Keyboard::Event>	mKeyQueue;
	deque<Joystick::Event>	mJoyQueue;
};

} // namespace input

#endif // INPUT_QUEUE_HH
