/*****************************************************************************
 * $LastChangedDate: 2008-09-20 10:29:31 -0400 (Sat, 20 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Base low-level Joystick class (system-neutral).
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define INPUT_JOYSTICK_CC 1
#include "base/module.hh"
using namespace base;
#include "input/module.hh"
#include "input/joystick.hh"

namespace input {

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  Joystick  /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

Joystick*   Joystick::msInstance;     // will be a derivative
int         Joystick::msInstanceCnt;  // only one Joystick should exist at a time

/*****************************************************************************
 * ctor/dtor.
 * Joystick is initially disabled to give chance for client/module
 * to become ready to respond to joystick events.
 *****************************************************************************/
Joystick::Joystick( void )
:	mEnabled(false),
	mPollFreq(DEFAULT_JOY_POLL_TICK_FREQ)
{
ASSERT( MAX_AXIS_VALUE == 32767 );
//ASSERT_RUN_ONCE;  // Joystick-derivatives may be switched at run-time
// Prevent more than one instance of any class of Joystick.
ASSERT_ONE_INSTANCE_CTOR( msInstanceCnt );
}

Joystick::~Joystick()
{
// Prevent more than one instance of any class of Joystick.
ASSERT_ONE_INSTANCE_DTOR( msInstanceCnt );
}

/*****************************************************************************
 * Singleton method.
 * This instantiates a derivative.
 *****************************************************************************/
Joystick&
Joystick::GetInstance( void )
{
    if ( UX( msInstance == NULL ) )
    {
        // Construct a Joystick-derivative.
#if OS_WINDOWS
        msInstance = new JoystickWindows;
#elif OS_LINUX || OS_FREEBSD
        msInstance = new JoystickLinux;
#elif COMPILE_GLUT_JOYSTICK
        msInstance = new JoystickGlut;
#else
        msInstance = new JoystickNop;  // default
#endif
    }

    ASSERT( msInstance != NULL );

    return *msInstance;
}


/*****************************************************************************
 * Switch joystick driver.
 * This is used if multiple drivers are available and an attempt is made
 * to select one over another.
 *****************************************************************************/
void
Joystick::SetJoystickDriver( eJoystickDriver joystickDriver )
{
    delete msInstance;
    msInstance = NULL;

    switch ( joystickDriver )
    {
#if OS_LINUX || OS_FREEBSD
        case eJoystickDriver_LINUX:
            msInstance = new JoystickLinux;
        break;
#endif

#if COMPILE_GLUT_JOYSTICK
        case eJoystickDriver_GLUT:
            msInstance = new JoystickGlut;
        break;
#endif

#if OS_WINDOWS
        case eJoystickDriver_WINDOWS:
            msInstance = new JoystickWindows;
        break;
#endif

        default:
            msInstance = new JoystickNop;
        break;
    }

    ASSERT( msInstance != NULL );
}

/*****************************************************************************
 * Enable/disable joystick.
 * @return True if joystick device is enabled and present, else false.
 *****************************************************************************/
bool
Joystick::Enable( bool enable )
{
	// Template Method using Open() and Close() derived methods.

	// NOP unless change in enablement.
	if ( enable && not mEnabled )  // if to open and was closed
	{
		// Open joystick driver.
		mEnabled = Open();  // might return false despite arg
	}
	else if ( not enable && mEnabled )  // if to close and was open
	{
		// Close joystick driver.
		Close();
		mEnabled = false;  // after close
	}

	return mEnabled;
}

/*****************************************************************************
 * Driver-specific Joystick-derivative code may or may not have a use for this.
 *****************************************************************************/
void
Joystick::SetPollFreq( const Milliseconds pollFreq )
{
	mPollFreq = pollFreq;
}

/*****************************************************************************
 * Poll all joystick events (up to a limit) from OS-specific joystick driver
 * and enqueue them into InputQueue.
 *****************************************************************************/
void
Joystick::PollAndEnqueueEvents( const uint limit )
{
BUSY_BEGIN {

	// Is joystick enabled?
	if ( mEnabled )
	{
		// Dequeue events from OS-specific joystick driver.
		for ( uint i = 0; i < limit; ++i )
		{
			Event event;
			if ( PollEvent( event ) )
			{
			ASSERT( event.IfValid() );

				// Enqueue event into higher-level InputQueue.
				InputQueue::GetInstance().EnqueueJoyEvent( event );
			}
			else
			{
				// No more events.
				break;
			}
		}
	}

} BUSY_END
}

} // namespace input
