/*****************************************************************************
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Joystick functions (Windows).
 * @remarks	- Typically, the range of a joystick axis on Windows is 0..65535.
 *			  To be similar to the Linux joystick code, 32768 is subtracted.
 *          - Link with winmm.lib.
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *			THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE (GPL).
 *****************************************************************************/

#if OS_WINDOWS

#define INPUT_JOYSTICK_WINDOWS_CC 1
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>
#include <Mmsystem.h>
#include <io.h>
#include "base/module.hh"
#include "base/conf.hh"
using namespace base;
#include "input/module.hh"
#include "input/joystick.hh"
#include "input/joystick_windows.hh"

namespace input {

INTERN const int MAX_AXIS_WINDOWS       = 6;        // X,Y,Z,R,U,V
INTERN const int MAX_BUTTONS_WINDOWS    = 4;
INTERN const int BUTTON_MASK            = 0xf;		// 4 bits:buttons
INTERN const int JOY_THRESHOLD_W32      = 500;      // affects sensitivity of joystick movement
INTERN bool		 sJoystickOpen          = false;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////  JoystickWindows  /////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
JoystickWindows::JoystickWindows( void )
{
    // NOP
}

JoystickWindows::~JoystickWindows()
{
	// NOP
}


/*****************************************************************************
 * Open joystick device.
 *****************************************************************************/
bool
JoystickWindows::Open( void )
{
	// Test the joystick once then remember the result.
	PERSISTENT int sOnce = 0;
	if ( sOnce++ )
    {
        return sJoystickOpen;
    }
    else
    {
    	JOYINFOEX joyinfo;
        joyinfo.dwSize  = sizeof(joyinfo);
        joyinfo.dwFlags = JOY_RETURNALL;
    	UINT wNumDevs, wDeviceID; 
    	BOOL bDev1Attached, bDev2Attached; 
    	if ( (wNumDevs = joyGetNumDevs()) == 0 )
            return (sJoystickOpen = false);
    	bDev1Attached = (joyGetPosEx(JOYSTICKID1,&joyinfo) != JOYERR_UNPLUGGED);
    	bDev2Attached = (wNumDevs == 2) and (joyGetPosEx(JOYSTICKID2,&joyinfo) != JOYERR_UNPLUGGED);
    	if ( bDev1Attached or bDev2Attached )  // decide which joystick to use
    	{
    		wDeviceID = bDev1Attached ? JOYSTICKID1 : JOYSTICKID2;
    		return (sJoystickOpen = true);
    	}
    	else 
    	{
    		return (sJoystickOpen = false);
    	}
    }
}

/*****************************************************************************
 * Close joystick device.
 *****************************************************************************/
void
JoystickWindows::Close( void )
{
	// NOP
}

/*****************************************************************************
 * Poll the next joystick event (non-blocking).
 * @return False if no more events.
 *****************************************************************************/
bool
JoystickWindows::PollEvent( Event& event /*OUT*/ )
{
	// Forget previous event.
   	event.Clear();

	// Is joystick enabled?
	if ( mEnabled and sJoystickOpen )
    {
		// Poll all axises.
		// For better responsiveness, during this tick, poll all axises.
		// This has to be done since the Windows joystick driver isn't a queue
		// (this code functions as if it were to directly read the PC I/O ports).
		// Note: This is done behind-the-back of Joystick::PollAndEnqueueEvents()
		// as it thinks one event is enqueued at a time (ok).
        for ( uint i = 0; i < MAX_AXIS_WINDOWS; ++i )
        {
            if ( PollAxisEvent( event, i ) )
                InputQueue::GetInstance().EnqueueJoyEvent( event );
        }

        // Poll button.
        return PollButtonEvent( event );
    }
    else
    {
	   return false;  // disabled
    }
}

/*****************************************************************************
 * Poll for one axis.
 *****************************************************************************/
bool
JoystickWindows::PollAxisEvent( Event& event /*OUT*/, const int axisNum )
{
ASSERT( MAX_AXIS_VALUE == 32767 );

    // Clear event struct.
    event.Clear();

	// Poll the joystick driver.
    int dwFlags = 0;
    switch ( axisNum )
    {
        case 0: dwFlags = JOY_RETURNX; break;
        case 1: dwFlags = JOY_RETURNY; break;
        case 2: dwFlags = JOY_RETURNZ; break;
        case 3: dwFlags = JOY_RETURNR; break;
        case 4: dwFlags = JOY_RETURNU; break;
        case 5: dwFlags = JOY_RETURNV; break;
        default:
        {
            event.Clear();  // ignore superfluous axis
            return false;
        }
        break;
    }
	JOYINFOEX joyinfo;  // extended joystick info
    base::CLEAR( &joyinfo, sizeof(joyinfo) );
    joyinfo.dwSize  = sizeof(joyinfo);
    joyinfo.dwFlags = dwFlags;
	if ( joyGetPosEx( 0, &joyinfo ) != JOYERR_NOERROR )
    {
        event.Clear();
        return false;
    }
    else
    {
    	// Fill Event struct.
    	// Typically, the range of a joystick axis on Windows is {0..65535}
    	// but remap it as {-32767..+32767}.
        event.mAxis = axisNum;
        switch ( axisNum )
        {
            case 0: event.mAxisVal = joyinfo.dwXpos - MAX_AXIS_VALUE; break;
            case 1: event.mAxisVal = joyinfo.dwYpos - MAX_AXIS_VALUE; break;
            case 2: event.mAxisVal = joyinfo.dwZpos - MAX_AXIS_VALUE; break;
            case 3: event.mAxisVal = joyinfo.dwRpos - MAX_AXIS_VALUE; break;
            case 4: event.mAxisVal = joyinfo.dwUpos - MAX_AXIS_VALUE; break;
            case 5: event.mAxisVal = joyinfo.dwVpos - MAX_AXIS_VALUE; break;
        }

        // Adjust axis value (32768 will appear on Windows but is 1 too much).
        if ( event.mAxisVal < -0x7fff )
            event.mAxisVal = -0x7fff;
        else if ( event.mAxisVal > 0x7fff )
            event.mAxisVal = 0x7fff;

        // Discard event if axis hasn't changed.
        PERSISTENT int sAxisVals[MAX_AXIS_WINDOWS];
        if ( event.mAxisVal != sAxisVals[axisNum] )
        {
            ASSERT( axisNum < MAX_AXIS_WINDOWS );
            sAxisVals[axisNum] = event.mAxisVal;
            return true;
        }
        else
        {
            event.Clear();  // axis hasn't moved
            return false;
        }
    }
}

/*****************************************************************************
 * Poll for any buttons.
 *****************************************************************************/
bool
JoystickWindows::PollButtonEvent( Event& event /*OUT*/ )
{
    // Clear event struct.
    event.Clear();

	// Poll the joystick driver.
	JOYINFO joyinfo;
	if ( joyGetPos( 0, &joyinfo ) != JOYERR_NOERROR )
    {
	   return false;  // no event or error
    }
    else
    {
    	// Button state.
    	int             buttonDelta;
    	PERSISTENT int	sButtonPrev = 0;
    	buttonDelta = joyinfo.wButtons ^ sButtonPrev;
    	buttonDelta &= BUTTON_MASK;  // ignore other bits
    	if ( buttonDelta )
    	{
    		// Buttons are indicated by bits.
    		// Correlate bit to its decimal number.
    		int button;
    		for ( button = 0; button < MAX_BUTTONS_WINDOWS; ++button )
    		{
    			if ( buttonDelta & (1<<button) )
    				break;
    		}
            ASSERT( button < MAX_BUTTONS_WINDOWS );
    		event.mButton    = button;  // zero-based
    		event.mButtonVal = BOOL_CAST( joyinfo.wButtons & (1<<button) );

    		// Since only one button is reported at a time,
    		// button_prev must record only the reported button,
    		// in order to possibly report changes to other buttons.
    		sButtonPrev &= (~(1<<button)); // merge reported bit, keep other bits
    		sButtonPrev |= (joyinfo.wButtons & (1<<button));
            return true;
    	}
    	else
    	{
    		sButtonPrev = joyinfo.wButtons;
            return false;  // no event
    	}
    }
}

/*****************************************************************************
 * Print status of joystick.
 *****************************************************************************/
void
JoystickWindows::PrintStatus( void )
{
	if ( not BASE_CONF.mQuiet && IfEnabled() )
	{
		CLOG << "Enabled Windows joystick." << endl;
	}
}

} // namespace input

#endif // OS_WINDOWS
