/*****************************************************************************
 * $LastChangedDate: 2008-06-22 19:01:10 -0400 (Sun, 22 Jun 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Macros/functions to assist and simplify writing glue code between C++ and Lua.
 * @remarks Indeed, macros.cc is a C++ file.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#include "base/module.hh"
using namespace base;
#include "world/module.hh"
#include "world/geo.hh"
using namespace world;
#include "lua_bind/module.hh"
#include "lua_bind/lua_lang.hh"
#include "lua_bind/bind.hh"

namespace lua_intern {

// Flag to suppress raising error in Lua if optional table members are absent.
bool Optional::msOptional = false;

/*****************************************************************************
 * Register a metatable to emulate a class in Lua.
 *****************************************************************************/
void
REGISTER_CLASS_METATABLE( lua_State* L,
                          const char* metaTableName,
                          const char* className,
                          const struct luaL_reg methodTableNew[],
                          const struct luaL_reg methodTable[] )
{
SAVE_STACK(L);

#if 1
	// Once only, make a global table for this C module.
	// It emulates a namespace inside Lua.
	lua_getglobal( L, LUABIND_MODULE_NAME );
	if ( lua_isnil( L, -1 ) )
	{
		lua_newtable( L );
		lua_setglobal( L, LUABIND_MODULE_NAME );
	}
	else
	{
		// Already exists.
		lua_pop( L, 1 );
	}

	// Then make a sub-table for this class in this C module's table.
	lua_getglobal( L, LUABIND_MODULE_NAME );
	lua_newtable( L );
	lua_setfield( L, -2, className );

	// Make a Lua metatable for this class.
	luaL_newmetatable( L, metaTableName );
	lua_pushstring( L, "__index" );
	lua_pushvalue( L, -2 );		// pushes the metatable
	lua_settable( L, -3 );		// metatable.__index = metatable
	luaL_register( L, NULL, methodTable );
	lua_getglobal( L, LUABIND_MODULE_NAME );
	if ( lua_type( L, -1 ) != LUA_TTABLE )
		THROW_LUA_EXCEPTION( "REGISTER_CLASS_METATABLE: lua_getglobal(module)" );
	lua_getfield( L, -1, className );
	if ( lua_type( L, -1 ) != LUA_TTABLE )
		THROW_LUA_EXCEPTION( "REGISTER_CLASS_METATABLE: lua_getfield(className)" );
	luaL_register( L, NULL, methodTableNew );

	// And register the methods with the table "sim.CLASS" itself.
	// This is similar to C++ "class methods".
	lua_getglobal( L, LUABIND_MODULE_NAME );
	lua_getfield( L, -1, className );
	luaL_register( L, NULL, methodTable );

#else // old way without using a namespace

	luaL_newmetatable( L, metaTableName );
	lua_pushstring( L, "__index" );
	lua_pushvalue( L, -2 );		// pushes the metatable
	lua_settable( L, -3 );		// metatable.__index = metatable
	luaL_register( L, NULL,      methodTable );
	luaL_register( L, className, methodTableNew );

#endif

RESTORE_STACK(L);
}

/*****************************************************************************
 * Pop a table containing a 4x4 matrix.
 *****************************************************************************/
void
POP_TABLE_INTO_MATRIX( lua_State* L, Matrix& matrix/*OUT*/ )
{
#if DEBUG
const int stackTop = lua_gettop(L);
#endif

	// Copy Lua table into C++ matrix.
	for ( int i = 15; i >= 0; --i )  // reversed, signed index
		matrix[i] = GET_MEMBER_AS_NUMBER( L, i+1 );  // Lua is 1-based

ASSERT( stackTop == lua_gettop(L) );

	// Pop table.
	lua_pop( L, 1 );
}

/*****************************************************************************
 * Push a table containing a 4x4 matrix.
 *****************************************************************************/
void
PUSH_TABLE_FROM_MATRIX( lua_State* L, const Matrix& matrix )
{
	lua_newtable( L );

#if DEBUG
const int stackTop = lua_gettop(L);
#endif

	for ( uint i = 0; i < 16; ++i )
		SET_MEMBER_AS_NUMBER( L, i+1, matrix[i] );  // Lua is 1-based

ASSERT( stackTop == lua_gettop(L) );
}

/*****************************************************************************
 * Pop RGBA from either 4 floats or a table with 4 elements.
 *****************************************************************************/
RGBA
POP_RGBA( lua_State* L )
{
	if ( EX( IF_TABLE( L, -1 ) ) )
	{
		fp rgba_f[4];
		rgba_f[RR] = GET_MEMBER_AS_NUMBER( L, int(RR) );
		rgba_f[GG] = GET_MEMBER_AS_NUMBER( L, int(GG) );
		rgba_f[BB] = GET_MEMBER_AS_NUMBER( L, int(BB) );
		rgba_f[AA] = GET_MEMBER_AS_NUMBER( L, int(AA) );
		lua_pop( L, 1 );  // pop RGBA table
		return RGBA( rgba_f );
	}
	else
	{
		// Pop in reverse order.
		fp a = POP_FP( L );
		fp b = POP_FP( L );
		fp g = POP_FP( L );
		fp r = POP_FP( L );
		return RGBA(r,g,b,a);
	}
}

/*****************************************************************************
 * Pop SphereVertex (heterogenous members).
 *****************************************************************************/
SphereVertex
POP_SPHERE_VERTEX( lua_State* L )
{
	const Vector3 v = POP_VECTOR3<Vector3>( L );
	return SphereVertex( Degree(v[0]), Degree(v[1]), Meter(v[2]) );
}

/*****************************************************************************
 * Push a spherical vertex as table with index starting at ZERO (0).
 *****************************************************************************/
void
PUSH_SPHERE_VERTEX( lua_State* L, const SphereVertex& sv )
{
	lua_newtable( L );
	SET_MEMBER_AS_NUMBER( L, LAT, sv.mLat.FP() );
	SET_MEMBER_AS_NUMBER( L, LON, sv.mLon.FP() );
	SET_MEMBER_AS_NUMBER( L, ALT, sv.mAlt.FP() );
}

} // namespace lua_intern
