/*****************************************************************************
 * $LastChangedDate: 2008-10-05 01:38:29 -0400 (Sun, 05 Oct 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Macros/functions to assist and simplify writing glue code between C++ and Lua.
 * @remarks module.txt documents how to export new C++ classes to Lua.
 * Avoid using names such as LUA_* which Lua itself uses.
 * Rather, write "lua::" to solve any naming conflicts.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef LUABIND_MACROS_HH
#define LUABIND_MACROS_HH 1

#include "math/module.hh"
#include "math/matrix.hh"
using namespace math;
#include "types/rgba.hh"
using namespace types;
#include "physics/types.hh"
using namespace physics;

using namespace lua;

namespace lua_intern {

/// @brief Special class to support optional Lua table members.
/// This allows C++ bindings to use try/catch to test for optional table members
/// without raising an error in Lua (usually, table members are expected/required).
class Optional
{
public:
	Optional( bool optional ) { msOptional = optional; }
	~Optional()               { msOptional = false; }
	CLASS_VAR bool msOptional;
};

const bool OPTIONAL = true;

/// Don't confuse this with Lua::ThrowLuaException() which restores stack.
#define THROW_LUA_EXCEPTION(TEXT)					\
	throw LuaException( TEXT );

/// For terseness.
#define CATCH_LUA_EXCEPTION							\
	catch ( const LuaException& e )					\
	{												\
		luaL_error( L, e.GetText().c_str() );       \
	}

// Macro to expand metatable's name as a C string.
#define LUABIND_METATABLE_STRING(NAME) LUABIND_MODULE_NAME "_" NAME

// luaconf.h defines a Lua number's type by "#define LUA_NUMBER double"
typedef LUA_NUMBER LuaNumber;

/// Save/restore stack.
/// Note: The Lua class has its own private methods SaveStack()/RestoreStack().
#define SAVE_STACK(L)	 int savedStack = lua_gettop( (L) )
#define RESTORE_STACK(L) lua_settop( (L), savedStack )

/// Unlike lua_isnumber/string(), these precisely identify a number or string (no coercion).
INLINE bool IF_NUMBER( lua_State* L, int idx ) { return lua_type( L, idx ) == LUA_TNUMBER; }
INLINE bool IF_STRING( lua_State* L, int idx ) { return lua_type( L, idx ) == LUA_TSTRING; }
INLINE bool IF_TABLE( lua_State* L, int idx )  { return lua_type( L, idx ) == LUA_TTABLE; }

/// Throw C++ exception (and raise error in Lua) if Lua type error.
INLINE void
THROW_IF_TYPE_ERROR( lua_State* L, int idx, int typeId )
{
	if ( UX( lua_type( L, idx ) != typeId ) )
	{
		if ( not Optional::msOptional ) 		  // don't raise Lua error if optional arg
			luaL_checktype( L, idx, typeId );     // this raises error in Lua
		THROW_LUA_EXCEPTION( "Lua type error" );  // this raises error in C++
	}
}

/// Pop a boolean ("true" or non-zero number) from Lua stack or raise error.
INLINE bool
POP_BOOL( lua_State* L )
{
	// Using lua_toboolean() alone isn't correct,
	// because the number 0 is true (indeed, true).
	// "lua_toboolean returns 1 for any Lua value different"
	// "from false and nil, otherwise it returns 0."
	bool res = false;
	const int luaType = lua_type( L, -1 );
	if ( luaType == LUA_TBOOLEAN )
	{
		res = BOOL_CAST( lua_toboolean( L, -1 ) );
	}
	else if ( luaType == LUA_TNUMBER )
	{
		res = (lua_tonumber( L, -1 ) != 0.0);
	}
	else if ( luaType == LUA_TNIL )
	{
		res = false;
	}
	else
	{
		THROW_LUA_EXCEPTION( "Lua type error (expected bool)" );
	}
	lua_pop( L, 1 );
	return res;
}

/// Pop a number (double) from Lua stack or raise error.
INLINE LuaNumber
POP_NUMBER( lua_State* L )
{
	THROW_IF_TYPE_ERROR( L, -1, LUA_TNUMBER );
	LuaNumber res = lua_tonumber( L, -1 );
	lua_pop( L, 1 );
	return res;
}
#define POP_FP POP_NUMBER

/// Pop a number (int) from Lua stack or raise error.
INLINE int
POP_INT( lua_State* L )
{
	return int(POP_NUMBER(L));
}

/// Pop a string from Lua stack or raise error.
INLINE const char*
POP_STRING( lua_State* L )
{
	THROW_IF_TYPE_ERROR( L, -1, LUA_TSTRING );
	const char* res = lua_tostring( L, -1 );
	if ( UX( res == NULL ) )
		THROW_LUA_EXCEPTION( "Lua type error" );
	lua_pop( L, 1 );
	return res;
}

/// Pop a Lua "lightuserdata" object (C pointer) from Lua stack or raise error.
template<typename T>
T*
POP_LIGHTUSERDATA( lua_State* L )
{
	THROW_IF_TYPE_ERROR( L, -1, LUA_TLIGHTUSERDATA );
	T* lightuserdata = const_cast<T*>( reinterpret_cast<const T*>( lua_topointer( L, -1 ) ) );
	lua_pop( L, 1 );
	return lightuserdata;
}

/// Directly pop Lua userdata and return its address.
/// This is low-level and doesn't know what is stored in the userdata.
INLINE void*
POP_USERDATA_ADDRESS( lua_State* L )
{
	THROW_IF_TYPE_ERROR( L, -1, LUA_TUSERDATA );
	void* addr = const_cast<void*>( lua_touserdata( L, -1 ) );
	lua_pop( L, 1 );
	return addr;
}

/// Pop Lua userdata, then deference the pointer it contains.
/// return *((T**)userdata)
template<typename T>
T*
POP_USERDATA_AS_PTR( lua_State* L )
{
	THROW_IF_TYPE_ERROR( L, -1, LUA_TUSERDATA );
	T** pPtr = reinterpret_cast<T**>( lua_touserdata( L, -1 ) );
	lua_pop( L, 1 );
	return *pPtr;  // dereference pointer-to-pointer
}

/// Pop user data, then return a copy of a ref-cnt variable that it contains.
/// Popping a base class pushed as a derivative should succeed
/// (shptr has the same memory layout regardless of the object's type it points to).
template<typename T>
shptr<T>
POP_USERDATA_AS_SHPTR( lua_State* L )
{
	THROW_IF_TYPE_ERROR( L, -1, LUA_TUSERDATA );
	shptr<T>* pShptr = reinterpret_cast<shptr<T>*>( lua_touserdata( L, -1 ) );
	CHECK_TYPESIG((*pShptr),TYPESIG_SHARED);
	lua_pop( L, 1 );
	return *pShptr;  // idiom: dereference pointer to shptr
}

/// Pop RGBA (Lua pushes 4 fp).
RGBA
POP_RGBA( lua_State* L );

/// Pop a table containing a 4x4 matrix.
void
POP_TABLE_INTO_MATRIX( lua_State* L, Matrix& matrix/*OUT*/ );

/// Push a table containing a 4x4 matrix.
void
PUSH_TABLE_FROM_MATRIX( lua_State* L, const Matrix& matrix );

/// Return table member by integer index.
INLINE LuaNumber
GET_MEMBER_AS_NUMBER( lua_State* L, const int idx )
{
	THROW_IF_TYPE_ERROR( L, -1, LUA_TTABLE );
	lua_pushnumber( L, idx );  // push idx as key
	lua_gettable( L, -2 );     // -2:table, -1:key
	return POP_NUMBER( L );    // keep table on stack
}

/// Pop a table member from Lua stack or raise error.
INLINE LuaNumber
GET_MEMBER_AS_NUMBER( lua_State* L, const char* memberName, const bool optional = false )
{
	Optional optionalObject( optional );
	THROW_IF_TYPE_ERROR( L, -1, LUA_TTABLE );
	lua_getfield( L, -1, memberName );
	return POP_NUMBER( L );
}

/// Pop a table member from Lua stack or raise error.
INLINE const char*
GET_MEMBER_AS_STRING( lua_State* L, const char* memberName, const bool optional = false )
{
	Optional optionalObject( optional );
	THROW_IF_TYPE_ERROR( L, -1, LUA_TTABLE );
	lua_getfield( L, -1, memberName );
	return POP_STRING( L );
}

/// Assign a table member (bool).
/// PRE: Table is on stack top.  POST: Table is still on stack top.
INLINE void
SET_MEMBER_AS_BOOL( lua_State* L, const char* key, const bool val )
{
	THROW_IF_TYPE_ERROR( L, -1, LUA_TTABLE );
	lua_pushboolean( L, val );
	lua_setfield( L, -2, key );  // and keep table on top
}

/// Write number to table member by a key string.
/// PRE: Table is on stack top.  POST: Table is still on stack top.
INLINE void
SET_MEMBER_AS_NUMBER( lua_State* L, const char* key, const LuaNumber val )
{
	THROW_IF_TYPE_ERROR( L, -1, LUA_TTABLE );
	lua_pushnumber( L, val );
	lua_setfield( L, -2, key );  // and keep table on top
}

/// Write number to table member by integer index.
/// PRE: Table is on stack top.  POST: Table is still on stack top.
INLINE void
SET_MEMBER_AS_NUMBER( lua_State* L, const int idx, const LuaNumber val )
{
	THROW_IF_TYPE_ERROR( L, -1, LUA_TTABLE );
	lua_pushnumber( L, idx );    // -2
	lua_pushnumber( L, val );    // -1
	lua_settable( L, -3 );       // -3:table
}

#define SET_MEMBER_AS_INT  SET_MEMBER_AS_NUMBER
#define SET_MEMBER_AS_FP   SET_MEMBER_AS_NUMBER

/// Pop a 3D vector as either x,y,z or {[0]=x,[1]=y,[2]=z}.
/// The Lua scripts define the class Vec3 which
/// is a table with index starting at ZERO (0).
template<typename VECTOR3>
VECTOR3
POP_VECTOR3( lua_State* L )
{
	// This gross template code supports GeoVertex(Meter,Meter,Meter) though.
	if ( EX( IF_TABLE( L, -1 ) ) )
	{
		typename VECTOR3::value_type x = typename VECTOR3::value_type( GET_MEMBER_AS_NUMBER( L, int(XX) ) );
		typename VECTOR3::value_type y = typename VECTOR3::value_type( GET_MEMBER_AS_NUMBER( L, int(YY) ) );
		typename VECTOR3::value_type z = typename VECTOR3::value_type( GET_MEMBER_AS_NUMBER( L, int(ZZ) ) );
		lua_pop( L, 1 );  // pop Vec3 table
		return VECTOR3( x, y, z );
	}
	else
	{
		// Pop in reverse.
		typename VECTOR3::value_type z = typename VECTOR3::value_type( POP_FP( L ) );
		typename VECTOR3::value_type y = typename VECTOR3::value_type( POP_FP( L ) );
		typename VECTOR3::value_type x = typename VECTOR3::value_type( POP_FP( L ) );
		return VECTOR3( x, y, z );
	}
}

/// Pop SphereVertex (heterogenous members).
SphereVertex
POP_SPHERE_VERTEX( lua_State* L );

/// Push a 3D vector as table {[0]=x,[1]=y,[2]=z} with index starting at ZERO (0).
template<typename VECTOR3>
void
PUSH_VECTOR3( lua_State* L, const VECTOR3& v )
{
	lua_newtable( L );
	SET_MEMBER_AS_NUMBER( L, int(XX), v[XX] );  // XX=0
	SET_MEMBER_AS_NUMBER( L, int(YY), v[YY] );  // int() because template requires exact arg types
	SET_MEMBER_AS_NUMBER( L, int(ZZ), v[ZZ] );
}

// Push a spherical vertex as table with index starting at ZERO (0).
void
PUSH_SPHERE_VERTEX( lua_State* L, const SphereVertex& sv );

/// This creates a method in Lua that produces a raw pointer to C/C++ object.
template<typename T>
int
NEW_USERDATA_PTR( lua_State* L, T* cppObj, const char* metaTableName )
{
	// Allocate memory for Lua userdata.
	void* userdata = lua_newuserdata( L, sizeof(T*) );

	// Write a pointer over Lua userdata.
	*reinterpret_cast<T**>(userdata) = cppObj;	

	luaL_getmetatable( L, metaTableName );
	lua_setmetatable( L, -2 );

	// Return amount of stack items back to Lua.
	return 1;
}

/// This creates a method in Lua that produces a reference-counted C++ object.
template<typename T>
int
NEW_USERDATA_SHPTR( lua_State* L, shptr<T> cppObj, const char* metaTableName )
{
	// Allocate memory for Lua userdata.
	void* userdata = lua_newuserdata( L, sizeof(shptr<T>) );

	// Now use C++ "placement" new to construct a reference-counted pointer variable.
	new (userdata) shptr<T>( cppObj.PTR() );

	luaL_getmetatable( L, metaTableName );
	lua_setmetatable( L, -2 );

	// Return amount of stack items back to Lua.
	return 1;
}

/// Destroy a smart pointer (not necessarily the object being pointed to).
/// This implements a Lua "__gc" method.
template<typename T>
int
DESTROY_USERDATA_SHPTR( lua_State* L, void(*CheckTypesig)(shptr<T>) )
{
	// Pop the userdata as a pointer to a reference-counted pointer (shptr).
	shptr<T>* pShptr = reinterpret_cast<shptr<T>*>( POP_USERDATA_ADDRESS( L ) );

	// Check integrity of C++ object that is pointed to.
	CHECK_TYPESIG( (*pShptr), TYPESIG_SHARED );
	CheckTypesig( *pShptr );

	// Explicitly call the shptr destructor in order to unreference the object.
	// This is the opposite of placement calling the constructor.
	// Calling a destructor doesn't free memory (shptr is at memory owned by Lua).
	pShptr->~shptr<T>();  // yuck

	// Return amount of stack items.
	return 0;
}

/// Register a metatable to emulate a class in Lua.
void
REGISTER_CLASS_METATABLE( lua_State* L,
                          const char* metaTableName,
                          const char* className,
                          const struct luaL_reg methodTableNew[],
                          const struct luaL_reg methodTable[] );

} // namespace lua_intern

using namespace lua_intern;

#endif // LUABIND_MACROS_HH
