/*****************************************************************************
 * $LastChangedDate: 2008-09-06 17:10:47 -0400 (Sat, 06 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Math functions for computer graphics.
 * @remarks	- To speed or improve FP execution:
 *            - use multiplication instead of division
 *            - power of 2 operations involve no loss of accuracy
 *            - math::PI is preferred over M_PI etc from math.h.
 *              math.h has confusing #defines such as M_2_PI
 *              which means 2/PI (not 2*PI).
 *          - NaN are caused by division and some standard C math functions.
 * REMINDERS: - Use MATH_EXCEPTION() instead of ASSERT().
 *            - C++ <cmath> is better than C <math.h> because C++ functions are overloaded.
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef MATH_MATH_HH
#define MATH_MATH_HH 1

#if _MSC_VER
#include <float.h>  // _finite()
#endif // _MSC_VER

namespace math {

const uint XX = 0;
const uint YY = 1;
const uint ZZ = 2;

const uint AXIS_X = 0;
const uint AXIS_Y = 1;
const uint AXIS_Z = 2;
const uint AXIS_INVALID = ~0;  // ~0 to cause segfault if used with array

// Reciprocals to use faster multiply.
const fp RECIP_256 = 1.0f / 256.0f;

} // namespace math

//-----------------------
#include "math/vertex.hh"
//-----------------------

namespace math {

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  D E B U G  ////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

// Same as ASSERT() but throws C++ exception.
// Clients of math now depend on C++ exception.
#define MATH_EXCEPTION( COND, TEXT ) {{ if ( UX( not (COND) ) ) throw Exception( TEXT ); }}

/// True if axis is XX,YY,ZZ.
#define ASSERT_AXIS( axis ) ASSERT( uint(axis) < 3 )

// Latitude should be in {0',..,180.0'}
// Longitude should be in {0',..,360.0'}
#if DEBUG
#define ASSERT_LAT_LON_DEG( LAT, LON )									\
	MATH_EXCEPTION( FP_GE<fp>(LAT,0.0f) and FP_LE<fp>(LAT,180.0f)		\
                and FP_GE<fp>(LON,0.0f) and FP_LE<fp>(LON,360.0f),		\
                     "(lat,lon) out-of-range" )
#else
#define ASSERT_LAT_LON_DEG( LAT, LON )  // ok as NOP for RELEASE
#endif

////////////////////////////////////////////////////////////////////////////////
///////////////////////////  C O M P A R I S O N  //////////////////////////////
////////////////////////////////////////////////////////////////////////////////

// IF_NAN() also returns true if infinity
// (divide by zero is technically infinity, not NaN, but IF_NAN() should catch divide errors).
// C99 isfinite() is true if neither NaN nor infinity.
// A NaN never equals anything -- not even another NaN.
// NOTE: !(n==n) is unreliable because the compiler could optimize as always false
// (GNU gcc 3.4 will do that).
#if _MSC_VER
#	define IF_NAN(N) (UX(!_finite(N)))
#else
#	define IF_NAN(N) (UX(!std::isfinite(N)))
#endif

/// Cope with floating-point inaccuracy.
/// Return true if two floats are approximately equal.
template<typename FP> bool FP_EQ( FP a, FP b ) { return ABS(a-b) < 0.001f; }
template<typename FP> bool FP_LE( FP a, FP b ) { return a-0.001f <= b or a+0.001f <= b; }
template<typename FP> bool FP_GE( FP a, FP b ) { return a-0.001f >= b or a+0.001f >= b; }

/// Compare floats: compare full integral part, compare fracDigits of fractional part.
/// Eg: FP_EQ( 1.234, 1.234, 4 )
/// Note: In practice, usually, fracDigits should be small, or miscompares will be frequent.
/// Tested by tests/math.cc.
template<typename FP>
bool
FP_EQ( const FP a, const FP b, const uint fracDigits )
{
	const FP diff = a - b;
	const FP scale = pow( double(10.0), double(fracDigits) );
	const uint64 tolerance = uint64(scale * 0.005f);
	return uint64(ABS(diff * scale)) <= tolerance;
}

/*****************************************************************************
 * @return True If two numbers have the same sign.
 *****************************************************************************/
template<typename T1,typename T2>
bool
IfSameSign( const T1 a, const T2 b )
{
#if 0
  //return (a * b) >= 0;  // wrong because if either is zero: -n * 0 = 0
#endif

	if ( a >= 0 and b >= 0 )  // both positive?
		return true;
	else if ( a < 0 and b < 0 )  // both negative
		return true;
	else
		return false;  // different signs
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  B A S I C   ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Select square root function according to type.
 *****************************************************************************/
INLINE float
SQRT( float f )
{
MATH_EXCEPTION( f >= 0.0f, "math error: sqrt(-n) undefined" );
//	return sqrtf( f );
	return std::sqrt( f );  // C++ <cmath>
}

INLINE double
SQRT( double f )  // use actual types (fpx may be double or long double)
{
MATH_EXCEPTION( f >= 0.0f, "math error: sqrt(-n) undefined" );
//	return sqrt( f );
	return std::sqrt( f );  // C++ <cmath>
}

INLINE long double
SQRT( long double f )
{
MATH_EXCEPTION( f >= 0.0f, "math error: sqrt(-n) undefined" );
//	return sqrtl( f );      // FreeBSD omits sqrtl()
	return std::sqrt( f );  // C++ <cmath>
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  T R I G   /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

// fpx is used in trig functions since they're used in matrix functions
// which need to maintain precision.

//  90' = PI / 2
// 180' = PI
// 360' = PI * 2
const fpx PI        = 3.14159265358979323846L;   // L for long double
const fpx PI_DOUBLE = 6.28318530717958647692L;   // pi * 2.0
const fpx PI_HALF   = 1.57079632679489661926L;   // pi / 2.0 [equivalent to 90']

// Turn a float type into a distinct type
// to ensure a degree/radian aren't confused.
/// @class math::Radian
/// @brief Distinct float type.
/// @class math::Degree
/// @brief Distinct float type.
#if SPEED
DECLARE_DISTINCT_TYPE( Radian, fp, FP )  // named conversion method FP() regardless
DECLARE_DISTINCT_TYPE( Degree, fp, FP )
#else
DECLARE_DISTINCT_TYPE( Radian, fpx, FP )
DECLARE_DISTINCT_TYPE( Degree, fpx, FP )
#endif

const Radian RADIAN_90  = math::PI_HALF;     // 90'  in radians (quarter circle)
const Radian RADIAN_180 = math::PI;          // 180' in radians (half circle)
const Radian RADIAN_360 = math::PI_DOUBLE;   // 360' in radians (full circle)

// Compare degrees/radians while ignoring sign.
#define IF_DEG_GT( DEG1, DEG2 ) (ABS(DEG1) > ABS(DEG2))
#define IF_DEG_LT( DEG1, DEG2 ) (ABS(DEG1) < ABS(DEG2))
#define IF_RAD_GT( RAD1, RAD2 ) (ABS(RAD1) > ABS(RAD2))
#define IF_RAD_LT( RAD1, RAD2 ) (ABS(RAD1) < ABS(RAD2))

/*****************************************************************************
 * Compute sine and cosine.
 * Example:
 * pair<fp,fp> si_co = SinCos<fp>( rad );
 * fp& s = si_co.first;
 * fp& c = si_co.second;
 *****************************************************************************/
template<typename FP>
pair<FP,FP>
SinCos( const Radian rad )
{
#if defined(__GNUC__) && CPU_X86
	pair<FP,FP> si_co;
	asm("fsincos" : "=t" (si_co.second), "=u" (si_co.first) : "0" (rad));
	return si_co;
#elif defined(__GNUC__) && _GLIBCXX_HAVE_SINCOS
	pair<double,double> si_co;  // must be specific in this case
	sincos( rad, &si_co.first, &si_co.second );  // GNU extension
	return make_pair<FP,FP>( si_co.first, si_co.second );
#else
	pair<FP,FP> si_co;
	si_co.first  = std::sin( rad );
	si_co.second = std::cos( rad );
	return si_co;
#endif
}

/*****************************************************************************
 * Convert a (360) degree into a radian.
 * 360' = 2 * PI
 *****************************************************************************/
INTERN const fpx degRadRatio = (1.0 / 360.0) * (math::PI_DOUBLE);  // fpx
#define DEG2RAD( DEG ) ((DEG) * (degRadRatio))

INLINE Radian
Deg2Rad( Degree deg )
{
	return DEG2RAD( deg );
	//return deg / 360.0 * (2.0 * math::PI);  // x87 FDIV is slower than FMUL
}

/*****************************************************************************
 * Convert a radian into a (360) degree.
 * 360' = 2 * PI
 *****************************************************************************/
INTERN const fpx radDegRatio = (1.0 / math::PI_DOUBLE) * 360.0;  // fpx
#define RAD2DEG( RAD ) ((RAD) * (radDegRatio))

INLINE Degree
Rad2Deg( Radian rad )
{
	return RAD2DEG( rad );
	//return rad / (2.0 * math::PI) * 360.0;  // x87 FDIV is slower than FMUL
}

/*****************************************************************************
 * 360' sin().
 *****************************************************************************/
INLINE Degree::value_type
Sin( Degree deg )
{
	return std::sin( Deg2Rad( deg ) );
}

/*****************************************************************************
 * 360' cos().
 *****************************************************************************/
INLINE Degree::value_type
Cos( Degree deg )
{
	return std::cos( Deg2Rad( deg ) );
}

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////  V E C T O R   //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Distance between vector and implied zero origin.
 *****************************************************************************/
template<typename FP>
FP
Distance( FP x, FP y, FP z )
{
	return SQRT( (x*x) + (y*y) + (z*z) );
}

INLINE fp
Distance( const Vector2& v )
{
	return SQRT( v[XX] * v[XX]
               + v[YY] * v[YY] );
}

INLINE fp
Distance( const Vector3& v )
{
	return SQRT( v[XX] * v[XX]
               + v[YY] * v[YY]
               + v[ZZ] * v[ZZ] );
}

/*****************************************************************************
 * Distance between two vectors (DistanceSquared() is faster than this).
 *****************************************************************************/
template<typename VECTOR3A,typename VECTOR3B>
fpx
Distance( const VECTOR3A& v1, const VECTOR3B& v2 )
{
	const fpx dx = v2[XX] - v1[XX];  // vector types may differ,
	const fpx dy = v2[YY] - v1[YY];  // use extended-precision
	const fpx dz = v2[ZZ] - v1[ZZ];
	return SQRT( dx * dx
               + dy * dy
               + dz * dz );
}

/*****************************************************************************
 * Calculate the distance^2 (squared) between two vectors.
 * DISTANCE RETURNED IS SQUARED (distance^2).
 * SQUARE ROOT OF DISTANCE IS SKIPPED FOR SPEED.
 * sqrt() and the underlying FPU instruction is very slow.
 *****************************************************************************/
INLINE Vector3::value_type
DistanceSquared( const Vector3& v1, const Vector3& v2 )
{
	Vector3::value_type xd = v2[XX] - v1[XX];
	Vector3::value_type yd = v2[YY] - v1[YY];
	Vector3::value_type zd = v2[ZZ] - v1[ZZ];
	return xd*xd + yd*yd + zd*zd;
}

/*****************************************************************************
 * @return A normalized vector (unit-length 1.0) of the same vector.
 * Not to be confused with calculating the normal vector of two vectors (CrossProduct).
 *****************************************************************************/
INLINE Vector3
Normalize( const Vector3& v )
{
	const Vector3::value_type d = Distance( v );
	const Vector3::value_type x = v[XX] / d;
	const Vector3::value_type y = v[YY] / d;
	const Vector3::value_type z = v[ZZ] / d;
MATH_EXCEPTION( (not IF_NAN(x)) and (not IF_NAN(y)) and (not IF_NAN(z)),
                "math error: Normalize(Vector3) NaN" );
    return Vector3( x, y, z );
}

template<typename VERTEX,typename FP>
VERTEX
Normalize( const FP x, const FP y, const FP z )
{
    const typename VERTEX::value_type d = Distance( x, y, z );
	const FP nx = x / d;
	const FP ny = y / d;
	const FP nz = z / d;
MATH_EXCEPTION( (not IF_NAN(nx)) and (not IF_NAN(ny)) and (not IF_NAN(nz)),
                "math error: Normalize(Vector3) NaN" );
    return VERTEX( nx, ny, nz );
}

/*****************************************************************************
 * Normalize a 3D point (unit 1 distance) then return one normalized coordinate.
 * The return value typically is passed to asin().
 *****************************************************************************/
INLINE Vector3::value_type
NormalizeCoord( uint axis, const Vector3& v )
{
ASSERT_AXIS(axis);
	Vector3::value_type d = Distance( v );
	Vector3::value_type n;
         if ( axis == XX ) n = v[XX] / d;
    else if ( axis == YY ) n = v[YY] / d;
    else if ( axis == ZZ ) n = v[ZZ] / d;
MATH_EXCEPTION( not IF_NAN(n), "math error: NormalizeCoord() NaN" );
	return n;
}

/*****************************************************************************
 * Increase/decrease a 3D vector by a scalar.
 *
 * The WRONG way, of course, is to simply add the scalar to each dimension.
 * For example, Vector(1,1) vs. Vector(2,2):
 * math.sqrt( (1.0*1.0) + (1.0*1.0) )
 * 1.4142135623730951
 * math.sqrt( (2.0*2.0) + (2.0*2.0) )
 * 2.8284271247461903  # oops != 2.4142
 *
 * The right way is to turn the scalar distance into a 3D vector
 * with the same direction as the input vector.
 * This is done by multiplying the scalar distance
 * by the normalized input vector, then adding.
 *****************************************************************************/
INLINE Vector3
AddDistance( const Vector3& v, const fp distance1 )
{
	const Vector3 norm3 = Normalize( v );
	const Vector3 distance3( distance1 * norm3[XX],
                             distance1 * norm3[YY],
                             distance1 * norm3[ZZ] );
	return Vector3( v[XX] + distance3[XX],
                    v[YY] + distance3[YY],
                    v[ZZ] + distance3[ZZ] );
}

/*****************************************************************************
 * Calculate a cross product of two vectors which yields a normal vector
 * that is perpendicular to the plane containing the two source vectors.
 * Note: The cross product won't be unit length.
 *****************************************************************************/
INLINE Vector3
CrossProduct( const Vector3& v1,
              const Vector3& v2,
              const Vector3  v0 = Vector3(0,0,0) ) // origin of v1 and v2
{
	Vector3::value_type xd1 = v1[XX] - v0[XX];
	Vector3::value_type yd1 = v1[YY] - v0[YY];
	Vector3::value_type zd1 = v1[ZZ] - v0[ZZ];
	Vector3::value_type xd2 = v2[XX] - v0[XX];
	Vector3::value_type yd2 = v2[YY] - v0[YY];
	Vector3::value_type zd2 = v2[ZZ] - v0[ZZ];
    return Vector3( yd1 * zd2 - yd2 * zd1,    // x: y*z
                    zd1 * xd2 - zd2 * xd1,    // y: z*x
                    xd1 * yd2 - xd2 * yd1 );  // z: x*y
}

/*****************************************************************************
 * Calculate the dot product of two vectors which yields a scalar.
 * If the angle between the two vectors is between 0 and 90 degrees,
 * the dot product will be positive, else negative.
 *****************************************************************************/
INLINE fp
DotProduct2( const Vector2& v1, const Vector2& v2 )
{
	return v1[XX] * v2[XX]
         + v1[YY] * v2[YY];
}

template<typename VECTOR3A,typename VECTOR3B>
fp
DotProduct3( const VECTOR3A& v1, const VECTOR3B& v2 )
{
	return v1[XX] * v2[XX]
         + v1[YY] * v2[YY]
         + v1[ZZ] * v2[ZZ];
}

// Extended-precision.
template<typename VECTOR3A,typename VECTOR3B>
fpx
DotProduct3_fpx( const VECTOR3A& v1, const VECTOR3B& v2 )
{
	return fpx(v1[XX]) * fpx(v2[XX])
         + fpx(v1[YY]) * fpx(v2[YY])
         + fpx(v1[ZZ]) * fpx(v2[ZZ]);
}

/*****************************************************************************
 * Calculate angle (in radians) between two vectors sharing the same origin.
 *****************************************************************************/
INLINE Radian
Angle2( const Vector2& u, const Vector2& v )  // Angle2/3() differ in DotProduct2/3()
{
	// Dot product formula indirectly provides the angle.
	// angle = arccos( DotProduct(u,v) / Distance(u) * Distance(v) )
	Vector2::value_type uv = Distance(u) * Distance(v);
	if ( ABS(uv) < 0.1f ) return 0.0f;  // prevent NaN and div by 0
	Vector2::value_type theta = DotProduct2(u,v) / uv;
	theta = Clamp( theta, -1.0f, 1.0f ); // acos() won't tolerate 1.00000001
errno = 0;
	Vector2::value_type rad = std::acos( theta );
MATH_EXCEPTION( errno == 0, "math error: Angle2() acos()" );  // if acos() failed
	return rad;
}

template<typename VECTOR>
Radian
Angle3( const VECTOR& u, const VECTOR& v )
{
	// Dot product formula indirectly provides the angle.
	// angle = arccos( DotProduct(u,v) / Distance(u) * Distance(v) )
	typename VECTOR::value_type uv = Distance(u) * Distance(v);
	if ( ABS(uv) < 0.1f ) return 0.0f;  // prevent NaN and div by 0
	typename VECTOR::value_type theta = DotProduct3(u,v) / uv;
	theta = Clamp( theta, -1.0f, 1.0f ); // acos() won't tolerate 1.00000001
errno = 0;
	typename VECTOR::value_type rad = std::acos( theta );
MATH_EXCEPTION( errno == 0, "math error: Angle3() acos()" );  // if acos() failed
	return rad;
}

/*****************************************************************************
 * Interpolate between two vectors.
 * @param   v1, v2
 * @param   fraction
 *          Fraction in range {0,..,1.0}.
 *          If fraction is 0.0, v1 is returned.
 *          If fraction is 1.0, v2 is returned.
 *          If fraction is 0.5, midpoint is returned.
 * @return Interpolated 3D point.
 *****************************************************************************/
INLINE Vector3
Interpolate( const Vector3& v1, const Vector3& v2, const fp fraction )
{
MATH_EXCEPTION( FP_GE(fraction, 0.0f) and FP_LE(fraction, 1.0f), "math error: Interpolate()" );
    return Vector3( v1 + ((v2-v1) * fraction) );
}

/*****************************************************************************
 * Midpoint.
 *****************************************************************************/
template<typename T>
T
Midpoint( T x1, T x2 )
{
//	return x1 + (x2 - x1) / 2;
	return (x1 + x2) / 2;  // faster
}

INLINE Vector3
Midpoint( const Vector3& v1, const Vector3& v2 )
{
	return Vector3( Midpoint(v1[XX], v2[XX]),
                    Midpoint(v1[YY], v2[YY]),
                    Midpoint(v1[ZZ], v2[ZZ]) );
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////  M I S C   M A T H   //////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * x^2
 *****************************************************************************/
template<typename T>
T
SQUARE( T x )
{
	return x * x;
}

/*****************************************************************************
 * Average.
 *****************************************************************************/
template<typename T>
T
AVG( T x, T y )
{
	return (x + y) / 2;
}

/*****************************************************************************
 * Force a number to be within a range.
 *****************************************************************************/
template<typename T>
T
Range( T val, T lo, T hi )
{
	if ( val < lo ) return lo;
	if ( val > hi ) return hi;
	return val;
}

/*****************************************************************************
 * @return True if two lines overlap.
 *****************************************************************************/
template<typename T>
T
IfInRange( T a0, T a1, T b0, T b1 )
{
	// Prove the reverse: test if out-of-range.
	     if ( a0 > b1 )
		return false;
	else if ( a1 < b0 )
		return false;
	else if ( b0 > a1 )
		return false;
	else if ( b1 < a0 )
		return false;
	else
		return true;  // within range
}

/*****************************************************************************
 * Compute a fraction for a number in a range.
 * Number is allowed to be outside the range (result will be clamped to 0 or 1).
 * Examples:
 * FractionInRange(100,100,200) = 0.0
 * FractionInRange(150,100,200) = 0.5
 * FractionInRange(200,100,200) = 1.0
 *****************************************************************************/
INLINE fp
FractionInRange( fp n, fp low, fp high )
{
// Negative range undefined.
MATH_EXCEPTION( (low < high) and (low >= 0 and high > 0),
                "math error: FractionInRange()" );

	if ( n <= low )
	{
		return 0.0f;
	}
	else if ( n >= high )
	{
		return 1.0f;
	}
	else  // inside range
	{
		return (n - low) / (high - low);
	}
}

/*****************************************************************************
 * Map a number from one range into another.
 *****************************************************************************/
INLINE fp
Remap( fp value, fp rangeOld, fp rangeNew )
{
	const fp n = value / rangeOld * rangeNew;
MATH_EXCEPTION( not IF_NAN(n), "math error: Remap()" );
	return n;
}

/*****************************************************************************
 * Truncate a float at an interval.
 * Eg, Truncate(24,10) returns 20.
 *****************************************************************************/
INLINE fp
Truncate( fp val, fp interval )
{
	float integral;  // float, not fp, as address is passed to modff()
	std::modf( val/interval, &integral );
	return integral * interval;
}

/*****************************************************************************
 * Modulo that handles negative values by subtracting from limit.
 * For example, ModuloAbs(deg,360.0f) can be used to accomodate a function
 * that requirs degrees to be in range {0,..,360}
 * ModuloAbs( -1.0,  360.0 ) = 359.0
 * ModuloAbs( -90.0, 360.0 ) = 270.0
 *****************************************************************************/
template<typename T>
T
ModuloAbs( const T n, const T limit )
{
	const T m = std::fmod( n, limit );
	if ( m >= 0 )
		return m;
	else  // m < 0
		return limit + m;  // effectively subtraction
}

/*****************************************************************************
 * Cycle {X,Y,Z}.
 *****************************************************************************/
INLINE uint CYCLE_AXIS( uint axis ) { return (axis + 1) % 3; }

} // namespace math

#endif // MATH_MATH_HH
