/*****************************************************************************
 * $LastChangedDate: 2008-09-06 17:10:47 -0400 (Sat, 06 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Matrix functions (OSG).
 * @remarks
 * Matrix functions are template functions because a few functions
 * in the OSG API operate on Matrixd (not Matrixf) so flexibility is needed.
 *//*
 * LEGAL:	COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef MATH_MATRIX_FUNCS_HH
#define MATH_MATRIX_FUNCS_HH 1

namespace math {

/*****************************************************************************
 * Convert an axis index to a 3D vector.
 * Useful for then passing 3D vector to OSG makeRotate() etc.
 * @param	axis
 *			XX,YY,ZZ
 * @param	dir
 *			-1,1
 *****************************************************************************/
INLINE osg::Vec3f
Axis2Vec3( const uint axis, const int dir )
{
ASSERT_AXIS( axis );
ASSERT( dir == 1 or dir == -1 );

	const fp one = 1.0f * dir;
	switch ( axis )
	{
		case XX: return osg::Vec3f( one,  0.0f, 0.0f ); break;
		case YY: return osg::Vec3f( 0.0f, one,  0.0f ); break;
		case ZZ: return osg::Vec3f( 0.0f, 0.0f, one  ); break;
	}
    ASSERT(false); return osg::Vec3f(0,0,0);   // oops
}

// Without dir arg.
INLINE osg::Vec3f
Axis2Vec3( const uint axis )
{
ASSERT_AXIS( axis );

	switch ( axis )
	{
		case XX: return osg::Vec3f( 1.0f, 0.0f, 0.0f ); break;
		case YY: return osg::Vec3f( 0.0f, 1.0f, 0.0f ); break;
		case ZZ: return osg::Vec3f( 0.0f, 0.0f, 1.0f ); break;
	}
    ASSERT(false); return osg::Vec3f(0,0,0);   // oops
}

/*****************************************************************************
 * Translate a Matrix by a vector offset.
 *****************************************************************************/
template<typename MATRIX>
void
MatrixTranslate( MATRIX& m /*IN/OUT*/, const Vector3& offset )
{
	MATRIX translation;
	translation.makeTranslate( offset );
	m = translation * m;
}

/*****************************************************************************
 * Translate a matrix along an axis in fixed space.
 * For translating the eye/viewpoint.
 *****************************************************************************/
template<typename MATRIX,typename FPM>
void
MatrixTranslateFixed( MATRIX& m /*IN/OUT*/, uint axis, FPM inc )
{
ASSERT_AXIS( axis );
	MATRIX translation;
	translation.makeTranslate( Axis2Vec3(axis,1) * inc );
	m = m * translation;
}

/*****************************************************************************
 * Translate a matrix along its own axis.
 * For translating a Dyna.
 *****************************************************************************/
template<typename MATRIX,typename FPM>
void
MatrixTranslateLocal( MATRIX& m /*IN/OUT*/, uint axis, FPM inc )
{
ASSERT_AXIS( axis );
	MATRIX translation;
	translation.makeTranslate( Axis2Vec3(axis,1) * inc );
	m = translation * m;
}

/*****************************************************************************
 * Rotate a matrix around a fixed axis.
 * This function is suited to rotating the viewpoint/eye.
 * The word "fixed" should be clear by looking at the math.
 * Eg, the value of the X coord won't be changed by a rotation around the X axis.
 *****************************************************************************/
template<typename MATRIX>
void
MatrixRotateFixed( MATRIX& m /*IN/OUT*/, uint axis, Radian rad )
{
ASSERT_AXIS( axis );
	MATRIX rotation;
	rotation.makeRotate( rad, Axis2Vec3(axis) );
	m = m * rotation;  // rotate around fixed axis
}

/*****************************************************************************
 * Rotate matrix around local axis.
 * Rotate a local coordinate system around its own axis.
 * This function is suited to rotating an independent object.
 * The rotation is relative to local coodinate system (not the fixed/eye system).
 * The trick is to load an identity-mapped matrix and rotate it, then transform
 * it thru the given matrix (which defines the local coordinate system)
 * as though it was the coords (1.0, 1.0, 1.0).  These coords of course define
 * the X,Y,Z axises.  The transformation is effectively a local rotation.
 *****************************************************************************/
template<typename MATRIX>
void
MatrixRotateLocal( MATRIX& m /*IN/OUT*/, uint axis, Radian rad )
{
ASSERT_AXIS( axis );
	MATRIX rotation;
	rotation.makeRotate( rad, Axis2Vec3(axis) );
	m = rotation * m;  // rotate around local axis
}

/*****************************************************************************
 * Reverse the mapping between two coordinate systems.
 * Math notes:
 * osg::Matrix::inverse() is more complex than a purely mathematical function.
 *****************************************************************************/
template<typename MATRIX>
MATRIX
MatrixTranspose( const MATRIX& src )
{
	return MATRIX::inverse( src );  // osg::Matrix::inverse()
}

/*****************************************************************************
 * Distance^2 between two matrixs.
 * Don't pass different kinds of matrixs (eg Eye origin is negative but Dyna is positive).
 *****************************************************************************/
template<typename MATRIX1, typename MATRIX2>
fpx
MatrixDistanceSquared( const MATRIX1& m1, const MATRIX2& m2 )
{
	Vector3 v1( m1[Ox], m1[Oy], m1[Oz] );
	Vector3 v2( m2[Ox], m2[Oy], m2[Oz] );
	return DistanceSquared( v1, v2 );
}

//------------------------------------------------------------------------------

void
MatrixScale( osg::MatrixTransform& transform /*IN/OUT*/, const Vector3& scale );

void
MatrixTranslate( osg::MatrixTransform& transform /*IN/OUT*/, const Vector3& offset );

void
MatrixRotateLocal( osg::MatrixTransform& transform /*IN/OUT*/, uint axis, Radian rad );

void
MatrixRotateFixed( osg::MatrixTransform& transform /*IN/OUT*/, uint axis, Radian rad );

ostream& operator<<( ostream& strm, const Matrix& m );

} // namespace math

#endif // MATH_MATRIX_FUNCS_HH
