/*****************************************************************************
 * $LastChangedDate: 2008-10-22 18:39:48 -0400 (Wed, 22 Oct 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Aircraft base class.
 *//*
 * LEGAL:	COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef OBJECT_AIRCRAFT_HH
#define OBJECT_AIRCRAFT_HH 1

#include "object/craft.hh"
#include "physics/module.hh"
#include "physics/physics_aircraft.hh"
using namespace physics;

namespace object {

////////////////////////////////////////////////////////////////////////////////
/// @brief Aircraft class.
///
/// This class is unusual:
/// - Aircraft has both a virtual and class method to get AircraftSpecs.
/// - Derivatives have a class method ClassGetSpecs() used for instantiating themselves.
///
/// mThrottle is a fp {0.0,...,1.0}.
/// mStep is a world coordinate.
/// Speed (translation increment) is determined by mThrottle * mStep.
///
/// Aircraft state enums:
/// The state enums (flying, landing, etc) are kept in the Aircraft class.
/// Both Aircraft and AircraftPhysics can affect the state.
///
class Aircraft : public Craft
{
PREVENT_COPYING(Aircraft)
typedef Craft Parent;
friend class shptr<Aircraft>; friend class shptr0<Aircraft>;
public:                         enum eState
                                {
                                    eState_PARKED,
                                    eState_TAKEOFF,
                                    eState_FLYING,
                                    eState_LANDING,
                                    eState_CRASHED,  // eCauseOfCrash complements this
                                };
public:                         enum eCauseOfCrash
                                {
                                    eCauseOfCrash_DEFAULT,
                                    eCauseOfCrash_DOWNWARD_TOO_FAST
                                };
// Only derivatives can access Aircraft ctor.
protected:						Aircraft( shptr<Graph> graph, const WorldVertex& pos, const AircraftSpecs& specs );
public:	   virtual				~Aircraft();
public:    CLASS_METHOD shptr<Graph> LoadModel( const AircraftSpecs& specs, const bool loadCopy );
public:    virtual void         Reset( void );
public:	   virtual const string	GetName( void ) { return string("Aircraft"); }

// Spatial:
public:    virtual void         SetPosition( const WorldVertex& position );
public:    virtual void			Translate( uint axis, fp inc );
public:    virtual void			Translate( const Vector3& v );
public:	   virtual void			Rotate( uint axis, Radian rad );

// Transformation:
public:    virtual void			SetMatrix( const Matrix& matrix, Dyna::eMatrixChange change );

// Control:
public:	   virtual void			PhysicalRotate( uint axis, fp controlFraction, Milliseconds controlFreq );
public:	   virtual void			SetThrottle( const fp throttle );
public:	   virtual fp			GetThrottle( void ) { return mThrottle; }
public:	   virtual void			EnableBrakes( const bool enable );
public:    virtual bool			IfBrakes( void );
public:    virtual bool			IfWheelBrakes( void );
public:    virtual bool			IfAirBrakes( void );
public:    virtual void			SetLandingGear( const bool down );
public:    virtual bool			IfLandingGear( void );  // true if extended down

// AircraftPhysics:
public:	   virtual AircraftPhysics& GetPhysics( void ) { CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);
                                                         return mPhysics; }
public:    void					EnableCoordinatedTurn( const bool enable );
public:    bool					IfCoordinatedTurnEnabled( void ) const;
public:    bool					IfCoordinatedTurnExecuting( void ) const;

// Animation:
protected: virtual void			Tick( const Milliseconds millisecElapsed );
protected: virtual void         AnimateStall( void );
public:    virtual void			SweepWings( const float fraction ) { }  // 1.0:swept-back

// States: collision-detection, crashing, contact force.
// Dyna has basic collision-detection.
// Aircraft determines what kind of collision: safe contact or fatal collision.
public:    virtual void			SetState( const eState state );
public:    virtual eState       GetState( void );
private:   void                 UpdateState( const Milliseconds millisecElapsed );  // subroutine of Tick()
public:    virtual bool			IfCrash( void );
public:	   virtual void			SetCauseOfCrash( const eCauseOfCrash causeOfCrash );
public:	   virtual eCauseOfCrash GetCauseOfCrash( void );
public:    virtual bool			IfCollisionRunway( void );

public:    virtual fp			GetHeight( void );
public:    virtual void			PlayEngineSound( const fp volume );
public:    virtual void			StopEngineSound( void );
public:    bool					IfFlying( void );
public:    bool					IfCanSetLandingGear( const bool down );
private:   void					MoreCrashDetermination( void );
public:    virtual const AircraftSpecs&	GetSpecs( void ) = 0;
public:    virtual bool			IfHasPropeller( void ) = 0;
public:    virtual bool			IfHasGuns( void ) { return true; }

// Data:
public:	   AircraftPhysics	mPhysics;					///< physics model (public to use AircraftPhysics interface)
private:   eState           mState;						///< general states { parked, landing, etc }
private:   Milliseconds     mStateUpdateTime;			///< timestamp when state was updated
private:   eCauseOfCrash	mCauseOfCrash;				///< the cause of the crash (fatal collision)
private:   bool				mCoordinatedTurnEnabled;	///< if coordinated turn is enabled
private:   bool				mCoordinatedTurnExecuting;	///< if program is performing a coordinated turn
private:   fp				mThrottle;					///< {0.0,..,1.0}
protected: bool				mLandingGearDown;
public:	   DECL_TYPESIG( TYPESIG_AIRCRAFT )
};

} // namespace object

#endif // OBJECT_AIRCRAFT_HH
