/*****************************************************************************
 * $LastChangedDate: 2008-09-14 21:44:12 -0400 (Sun, 14 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Dyna class.
 * @verbatim
 *
 * Translate() vs. SetMatrix():
 * ----------------------------
 * Calling Translate() or SetMatrix(m,TRANSLATION) is considered movement.
 * Calling SetMatrix(void) is considered placement.
 * Collision-detection is tested while moving only.
 *
 * Object defines *Collidable*() methods.
 * Dyna then defines *Collision*() methods.
 *
 * @endverbatim
 *//*
 * LEGAL:	COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define OBJECT_DYNA_CC 1
#include "base/module.hh"
using namespace base;
#include "math/module.hh"
#include "math/matrix.hh"
using namespace math;
#include "world/module.hh"
using namespace world;
#include "collision/module.hh"
using namespace collision;
#include "object/module.hh"
#include "object/defs.hh"
#include "object/dyna.hh"

namespace object {

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////  Dyna  //////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Dyna::Dyna( shptr<Graph> graph, const WorldVertex& pos )
:	Parent(graph,pos),
	mTickFunctor(new Dyna::TickFunctor(this)),
    mTranslationBegin(0,0,0),
    mTranslationEnd(0,0,0),
	mCollision(false),
	mCollisionObject(NULL)
{
	SET_TYPESIG(this,TYPESIG_DYNA);

	// Register a timer-tick handler to pulse animation.
	Timer::GetInstance().Register( mTickFunctor.PTR(), defs::DYNA_TICK_FREQ );

    // Most Dynas are collidable.
    SetCollidable( true );
}

Dyna::~Dyna()
{
	// Tell Timer to stop calling this Dyna which will be destroyed.
	Timer::GetInstance().Unregister( mTickFunctor.PTR() );

	INVALIDATE_TYPESIG(this,TYPESIG_DYNA);
}

/*****************************************************************************
 * Reset.
 *****************************************************************************/
void
Dyna::Reset( void )
{
CHECK_TYPESIG(this,TYPESIG_DYNA);

	Parent::Reset();

	// Forget previous translation.
    mTranslationBegin = WorldVertex(0,0,0);
    mTranslationEnd   = WorldVertex(0,0,0);

	// Leave mCollision unchanged (let derivative control it).
  //mCollision = false;
}

/*****************************************************************************
 * If SetMatrix() is called for movement (not placement, not rotation),
 * then this is a translation action, so mTranslationBegin/End must be updated
 * for collision-detection.
 *****************************************************************************/
void
Dyna::SetMatrix( const Matrix& matrix, Dyna::eMatrixChange change )
{
CHECK_TYPESIG(this,TYPESIG_DYNA);

	if ( change == Dyna::TRANSLATION )
	{
		mTranslationBegin = GetPosition();
		Parent::SetMatrix( matrix );  // changes position
		mTranslationEnd = GetPosition();
	}
	else
	{
		Parent::SetMatrix( matrix );
	}
}

/*****************************************************************************
 * Object defines *Collidable*() methods.
 * Dyna then defines *Collision*() methods.
 *****************************************************************************/
// If a collision has occurred or not.
void
Dyna::SetCollision( const bool collision )
{
CHECK_TYPESIG(this,TYPESIG_DYNA);

	mCollision = collision;
}

bool
Dyna::IfCollision( void )
{
CHECK_TYPESIG(this,TYPESIG_DYNA);

    return mCollision;
}

shptr0<Object>
Dyna::GetCollisionObject( void )
{
CHECK_TYPESIG(this,TYPESIG_DYNA);

    return mCollisionObject;
}

/*****************************************************************************
 * @return The first and last points of the current movement (step) of this Dyna.
 *****************************************************************************/
const pair<WorldVertex,WorldVertex>
Dyna::GetTranslationStep( void )
{
CHECK_TYPESIG(this,TYPESIG_DYNA);

    return make_pair<WorldVertex,WorldVertex>( mTranslationBegin, mTranslationEnd );
}

/*****************************************************************************
 * Animation.
 * @param	millisecElapsed
 *			Total milliseconds between now and when program was started.
 *****************************************************************************/
void
Dyna::Tick( const Milliseconds millisecElapsed )
{
CHECK_TYPESIG(this,TYPESIG_DYNA);

	// Detect collision.
	// Derived class is responsible for determining
	// if this is a safe or fatal collision (crashed).
	// Bypass if this was set non-collidable.
	if ( EX( IfCollidable() ) )
	{
		const pair<bool,shptr0<Object> > collision = GET_COLLISION().IfCollision( this );
		mCollision       = collision.first;
		mCollisionObject = collision.second;
	}
}

void
Dyna::TickFunctor::operator()( const Milliseconds millisecElapsed )
{
	// Reflect to Dyna::Tick().
	CHECK_TYPESIG(mDyna,TYPESIG_DYNA);
	mDyna->Tick( millisecElapsed );
}

} // namespace object
