/*****************************************************************************
 * $LastChangedDate: 2008-09-14 17:36:06 -0400 (Sun, 14 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Dyna class.
 *//*
 * LEGAL:	COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef OBJECT_DYNA_HH
#define OBJECT_DYNA_HH 1

#include "base/timer.hh"

namespace object {

////////////////////////////////////////////////////////////////////////////////
/// @brief Dyna class.
///
/// Optionally, a Dyna-derivative can do animation (or any periodic action)
/// by calling Dyna::RegisterTick() and overriding Dyna::Tick().
///
/// Translate() vs. SetMatrix():
/// ----------------------------
/// Calling Translate() or SetMatrix(m,TRANSLATION) is considered movement (stepping).
/// Calling SetMatrix(void) isn't.
/// The difference affects collision-detection.
///
/// Collision vs. crashed:
/// ----------------------
/// Collision means an intersection -- but a collision isn't necessarily fatal.
/// Two objects could be in "safe contact" (aircraft in contact with runway).
/// Dyna reports a collision, while physics, which knows the velocity,
/// determines whether a collision is fatal or not.
///
/// Dyna::GetHeight() can be overridden by derivatives
/// for more accurate collision-detection.
/// Collision-detection uses GetHeight() to form a vertical ray.
///
/// Collision-detection in Object hierarchy:
/// ----------------------------------------
/// Object defines the *Collidable*() methods.
/// Dyna defines the *Collision*() methods which test collision-detection.
/// Craft determines what kind of collision: safe contact or fatal collision.
///
class Dyna : public Object
{
PREVENT_COPYING(Dyna)
typedef Object Parent;
public:								Dyna( shptr<Graph> graph, const WorldVertex& pos );
public:		virtual					~Dyna();
public:     virtual void			Reset( void );
public:		virtual bool			Reproducible( void ) { return false; }
public:		virtual const string	GetName( void ) { return string("Dyna"); }

// Spatial:
public:     virtual fp              GetHeight( void ) { return GetRadius() * 2.0f; }

// Transformation:
public:		virtual void			SetMatrix( const Matrix& matrix, Object::eMatrixChange change = PLACEMENT );

// Collision-detection:
public:		virtual void			SetCollision( const bool collision );    // set whether collision has occurred
public:		virtual bool			IfCollision( void );                     // if collision has occurred
public:     virtual shptr0<Object>  GetCollisionObject( void );              // the other object this one collided into
public:     virtual const pair<WorldVertex,WorldVertex> GetTranslationStep( void );

// Animation:
protected:	virtual void			Tick( const Milliseconds millisecElapsed );

private:
	/// @brief Reflects tick event to Dyna::Tick().
	/// Indeed, TickFunctor doesn't have a shptr<Dyna>.
	/// It would create a cycle of reference-counted pointers.
	/// ~Dyna() unregisters Dyna::TickFunctor but ~Dyna() would
	/// never be called if TickFunctor had a reference to Dyna!
	///
	class TickFunctor : public Timer::Functor
	{
	PREVENT_COPYING(TickFunctor)
	public:
		TickFunctor( SafePtr<Dyna> dyna ) : mDyna(dyna) { }
		void operator()( const Milliseconds millisecElapsed );
		SafePtr<Dyna> mDyna;  // cannot be shptr!!
	};
	friend class TickFunctor;

// Data:
private:	shptr<TickFunctor>	mTickFunctor;       ///< reflects to Dyna::Tick()
protected:  WorldVertex         mTranslationBegin;	///< begin and end points of last translation
protected:  WorldVertex         mTranslationEnd;	///< used for collision-detection
protected:	bool				mCollision;			///< safe or fatal, updated every frame by Dyna::Tick()
protected:	shptr0<Object>		mCollisionObject;	///< the other Object this one collided into
public:		DECL_TYPESIG(TYPESIG_DYNA)
};

} // namespace object

#endif // OBJECT_DYNA_HH
