/*****************************************************************************
 * $LastChangedDate: 2008-10-21 14:43:19 -0400 (Tue, 21 Oct 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Base Object class.
 *//*
 * LEGAL:	COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef OBJECT_OBJECT_HH
#define OBJECT_OBJECT_HH 1

#define CHECK_OBJECT() {{ CHECK_TYPESIG(this,TYPESIG_OBJECT); }}

#include "graph/module.hh"

namespace object {

////////////////////////////////////////////////////////////////////////////////
/// @brief 3D Object class.
///
class Object : public Shared
{
PREVENT_COPYING(Object)

//------------------------------------------------------------------------------
// Interface:
// Construction, identity:
public: CLASS_METHOD shptr<Object>  MakeDummyObject( void );
public:								Object( shptr<Graph> graph, const WorldVertex& pos );
public:     virtual					~Object();
public:		virtual void			Reset( void ) { }
public:     virtual bool			Reproducible( void ) { return false; }
public:		virtual const string	GetName( void ) { return string("Object"); }
public:		bool					IfZombie( void ) { return mBits.mZombie; }
public:		void					Zombify( void );

// Drawing:
public:     virtual void			Draw( void );

// State-sorting:
public:     virtual StateBits		GetStateBits( void );

// Spatial:
public:     virtual void            SetPosition( const WorldVertex& position );
public:		virtual WorldVertex     GetPosition( void );
public:		virtual fp				GetRadius( void )
									{
									CHECK_OBJECT();
										// Was radius computed before?
										if ( EX( mRadius > 0.0f ) )  // initially -1 (invalid)
											return mRadius;
										else
											return (mRadius = ComputeRadius());
									}
private:	fp						ComputeRadius( void );

// Geometry:
public:     shptr<Graph>            GetGraph( void )
                                    { CHECK_OBJECT(); return mGraph; }
public:		void					Scale( const Vector3& v );
public:		void					Scale( const fp scale );
public:     RefPtr<Node>            GetRootNode( void )
                                    { CHECK_OBJECT(); return GetGraph()->GetRootNode(); }
// Transformation:
public:		enum eMatrixChange { PLACEMENT, ROTATION, TRANSLATION };  // see Dyna class
public:		virtual void			SetMatrix( const Matrix& matrix, Object::eMatrixChange change = PLACEMENT );
public:     virtual Matrix			GetMatrix( void )
									{ CHECK_OBJECT(); return mTransformNode->getMatrix(); }
public:		virtual void			SetQuaternion( const Quaternion& quat );
public:		virtual Quaternion		GetQuaternion( void );
public:		virtual void			Translate( uint axis, fp inc );
public:		virtual void			Translate( const Vector3& v );
public:		virtual void			Rotate( uint axis, Radian rad );

// Collidable/indestructible:
public:     virtual void            SetCollidable( const bool collidable );
public:     virtual bool            IfCollidable( void )
                                    { CHECK_OBJECT(); return mBits.mCollidable and not mBits.mZombie; }
// Info:
public:		CLASS_METHOD int		GetObjectCnt( void ) { return msObjectCnt; }
private:    CLASS_VAR int			msObjectCnt;

//------------------------------------------------------------------------------
// Data:

protected:  struct Bits
            {
				uint mCollidable	: 1;
				uint mZombie		: 1;
            };
protected:  shptr<Graph>					mGraph;
private:	RefPtr<osg::MatrixTransform>	mTransformNode;     ///< force derivatives to use SetMatrix() which can intercept incorrect positions
protected:  fp								mRadius;            ///< approximate
protected:  Bits							mBits;
public:     DECL_TYPESIG( TYPESIG_OBJECT )

};

} // namespace object

#endif // OBJECT_OBJECT_HH
