/*****************************************************************************
 * $LastChangedDate: 2008-10-04 00:18:42 -0400 (Sat, 04 Oct 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Aircraft specifications.
 *//*
 * LEGAL:	COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define PHYSICS_AIRCRAFT_SPECS_CC 1
#include "base/module.hh"
#include "base/conf.hh"
using namespace base;
#include "object/module.hh"
#include "object/rotation_list.hh"
using namespace object;
#include "world/module.hh"
using namespace world;
#include "physics/module.hh"
#include "physics/aircraft_specs.hh"

namespace physics {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////  AircraftSpecs  //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
/// @brief Fills a AircraftSpecs object from a .conf textual file.
///
/// Mass, max thrust, etc are defined in a data file (physics_*.conf).
/// These parameters are loaded into a AircraftSpecs for use by AircraftPhysics.
/// AircraftSpecs is data, AircraftPhysics is computational.
///
class AircraftSpecsConfTxt : public base::ConfTxt
{
public:
			AircraftSpecsConfTxt( AircraftSpecs& aircraftSpecs ) : mAircraftSpecs(aircraftSpecs) { }
    void	ProcessField( string& field, ifstream& conf );

private:
	AircraftSpecs& mAircraftSpecs;
};

/*****************************************************************************
 * ctor.
 *****************************************************************************/
AircraftSpecs::AircraftSpecs( void )
:	mMass(0),			// zero them for assertion checking
	mMaxThrustMag(0),	// magnitude (Newton1)
	mEngineType(eEngineType_PROPELLER),
	mDragAofALimit(0),
	mDragAofADegree(0),
	mLiftoffSpeed(0),
	mMaxAofA(0),
	mRangeOfLiftFactor(0),
	mRangeOfLiftFactorSpeed(0),
	mRangeOfLiftFactorAofA(0),
	mTurnDegreeLow(0),
	mTurnDegreeHigh(0),
	mLength(0),
	mHeight(0),
	mMaxSpeed(0),
	mStallSpeed(0),
	mRollRate(0),
	mPitchRate(0),
	mYawRate(0),
	mModelFile(),
	mModelScale(1.0f),
	mModelOffset(Meter(0),Meter(0),Meter(0)),
	mModelOffsetGear(0),
	mModelRotationList(),
	mModelLighting(true)
{
	// NOP
}

/*****************************************************************************
 * Read a Aircraft's specificiations/characteristics from a .conf file.
 *****************************************************************************/
void
AircraftSpecs::ReadSpecsFromConfFile( const string& confTxtFilename )
{
	try
	{
		AircraftSpecsConfTxt confTxt( *this );  // will load this AircraftSpecs
		confTxt.Read( string(base::GetDataDir()) + string(confTxtFilename) );
	}
	catch (...)
	{
		CERROR << "ERROR: cannot read " << confTxtFilename << endl;
		EXIT( 1 );
	}
}

void
AircraftSpecsConfTxt::ProcessField( string& field, ifstream& conf )
{
	bool error = false;

	// -------------------------- mass -----------------------------------------
	if ( field == "Mass" )
	{
		fp f; conf >> f;
		mAircraftSpecs.mMass = Kilogram( f );
	}
	// -------------------------- thrust -----------------------------------------
	else if ( field == "MaxThrust" )
	{
		fp f; conf >> f;
		mAircraftSpecs.mMaxThrustMag = Newton1(f);  // magnitude
	}
	else if ( field == "EngineType" )
	{
		// Convert string to enum.
		string s; conf >> s;
		if ( s == "propeller" )
			mAircraftSpecs.mEngineType = AircraftSpecs::eEngineType_PROPELLER;
		else if ( s == "jet" )
			mAircraftSpecs.mEngineType = AircraftSpecs::eEngineType_JET;
		else
			error = true;
	}
	// -------------------------- drag -----------------------------------------
	else if ( field == "DragAofALimit" )
	{
		fp f; conf >> f;
		mAircraftSpecs.mDragAofALimit = f;
	}
	else if ( field == "DragAofADegree" )
	{
		fp f; conf >> f;
		mAircraftSpecs.mDragAofADegree = f;
	}
	// -------------------------- lift -----------------------------------------
	else if ( field == "LiftoffSpeed" )
	{
		fp f; conf >> f;
		mAircraftSpecs.mLiftoffSpeed = SpeedKPH(f);
	}
	else if ( field == "MaxAofA" )
	{
		fp f; conf >> f;
		mAircraftSpecs.mMaxAofA = f;
	}
	else if ( field == "RangeOfLiftFactor" )
	{
		conf >> mAircraftSpecs.mRangeOfLiftFactor;
	}
	else if ( field == "RangeOfLiftFactorSpeed" )
	{
		conf >> mAircraftSpecs.mRangeOfLiftFactorSpeed;
	}
	else if ( field == "RangeOfLiftFactorAofA" )
	{
		conf >> mAircraftSpecs.mRangeOfLiftFactorAofA;
	}
	// -------------------------- turning -----------------------------------------
	else if ( field == "TurnDegreeLow" )
	{
		fp f; conf >> f;
		mAircraftSpecs.mTurnDegreeLow = f;
	}
	else if ( field == "TurnDegreeHigh" )
	{
		fp f; conf >> f;
		mAircraftSpecs.mTurnDegreeHigh = f;
	}
	// -------------------------- airframe -----------------------------------------
	else if ( field == "Length" )
	{
		fp f; conf >> f;
		mAircraftSpecs.mLength = Meter(f);
	}
	else if ( field == "Height" )
	{
		fp f; conf >> f;
		mAircraftSpecs.mHeight = Meter(f);
	}
	else if ( field == "MaxSpeed" )
	{
		fp f; conf >> f;
		mAircraftSpecs.mMaxSpeed = SpeedKPH(f);
	}
	else if ( field == "StallSpeed" )
	{
		fp f; conf >> f;
		mAircraftSpecs.mStallSpeed = SpeedKPH(f);
	}
	else if ( field == "RollRate" )
	{
		fp f; conf >> f;
		mAircraftSpecs.mRollRate = f;
	}
	else if ( field == "PitchRate" )
	{
		fp f; conf >> f;
		mAircraftSpecs.mPitchRate = f;
	}
	else if ( field == "YawRate" )
	{
		fp f; conf >> f;
		mAircraftSpecs.mYawRate = f;
	}
	// -------------------------- 3D model -------------------------------------
	else if ( field == "ModelFile" )
	{
		conf >> mAircraftSpecs.mModelFile;
	}
	else if ( field == "ModelScale" )
	{
		conf >> mAircraftSpecs.mModelScale;
	}
	else if ( field == "ModelOffset" )
	{
		fp f;
		conf >> f; mAircraftSpecs.mModelOffset[XX] = Meter(f);
		conf >> f; mAircraftSpecs.mModelOffset[YY] = Meter(f);
		conf >> f; mAircraftSpecs.mModelOffset[ZZ] = Meter(f);
	}
	else if ( field == "ModelOffsetGear" )
	{
		fp f; conf >> f; mAircraftSpecs.mModelOffsetGear = Meter(f);
	}
	else if ( field == "ModelRotationList" )
	{
		mAircraftSpecs.mModelRotationList.Parse( conf );
	}
	else if ( field == "ModelLighting" )
	{
		string flag; conf >> flag;
		if ( flag == "false" or flag == "0" )
			mAircraftSpecs.mModelLighting = false;
		else
			mAircraftSpecs.mModelLighting = true;
	}

    // Errors?
	if ( error )
	{
		ostringstream msg;
		msg << "ERROR[conf]: parameter for " << field << " is invalid." << endl;
		throw Exception( msg.str() );
	}
}

} // namespace physics
