/*****************************************************************************
 * $LastChangedDate: 2008-10-04 00:18:42 -0400 (Sat, 04 Oct 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Aircraft carrier.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define PROGRAM_CARRIER_CC 1
#include "base/module.hh"
using namespace base;
#include "glue/module.hh"
#include "glue/conf.hh"
#include "glue/light_nums.hh"
using namespace glue;
#include "graph/module.hh"
#include "graph/model_cache.hh"
using namespace graph;
#include "object/module.hh"
#include "object/aircraft.hh"
using namespace object;
#include "world/module.hh"
using namespace world;
#include "sky/module.hh"
using namespace sky;
#include "program/module.hh"
#include "program/carrier.hh"

namespace program {

INTERN const Meter CARRIER_RUNWAY_LENGTH( 220.0f );  // slightly more for collision-detection
INTERN const Meter CARRIER_RUNWAY_ALT( 23.43f );
INTERN const Meter CARRIER_RUNWAY_ALT_THRESHOLD( 2.5f );

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////  Carrier  /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

vector<shptr<Carrier> > Carrier::msCarriers;

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Carrier::Carrier( shptr<Graph> graph, const WorldVertex& pos, const uint carrierNum )
:	Parent(graph,pos)
{
	SET_TYPESIG(this,TYPESIG_CARRIER);

	// Orient the aircraft carrier.
	PROGRAM_CONF.mCarrierRotationList.Apply( *this );

	msCarriers.push_back( this );
}

Carrier::~Carrier()
{
	INVALIDATE_TYPESIG(this,TYPESIG_CARRIER);
}

/*****************************************************************************
 * Place aircraft on carrier.
 *****************************************************************************/
void
Carrier::PlaceAircraft( shptr<Aircraft> aircraft, const GeoVertex& relativeGeoPosition )
{
CHECK_TYPESIG(this,TYPESIG_CARRIER);

	// Compute position of aircraft.
	WorldVertex position = GetPosition();  // start at carrier's position
	position += Geo::GeoVertex2WorldVertex( relativeGeoPosition );  // add relative position
	aircraft->SetPosition( position );

	// Adjust altitude of aircraft so that wheels sit on runway.
	// mModelOffsetGear could be zero.
	const fp wheelOffset = Geo::Meters2Sim( aircraft->GetSpecs().mModelOffsetGear );
	aircraft->Translate( YY, wheelOffset );

	// Extend landing gear.
	aircraft->SetLandingGear( true );
}

/*****************************************************************************
 * Higher-level method to make an aircraft carrier.
 *****************************************************************************/
shptr<Carrier>  // CLASS_METHOD
Carrier::MakeCarrier( const WorldVertex& carrierPosition, const uint carrierNum )
{
	shptr<Graph> graph = ModelCache::GetInstance().LoadModel( PROGRAM_CONF.mCarrierModelName );
	shptr<Carrier> carrier = new Carrier( graph, carrierPosition, carrierNum );
	gWorld.Attach( carrier.PTR() );
	program::InsertLightIntoGraph( graph,
                                   glue::defs::LIGHT_NUM_MODEL_CARRIER,
                                   RGBA(0.70f, 0.70f, 0.70f, 1.00f),    // ambient
                                   RGBA(0.80f, 0.80f, 0.80f, 1.00f),    // diffuse
                                   RGBA(0.10f, 0.10f, 0.10f, 1.00f) );  // specular
	return carrier;
}

/*****************************************************************************
 * (class method).
 * Get carrier by index.
 *****************************************************************************/
shptr<Carrier>  // CLASS_METHOD
Carrier::GetCarrier( const uint carrierNum )
{
ASSERT( carrierNum < msCarriers.size() );
	return msCarriers[carrierNum];
}

/*****************************************************************************
 * @return True if a position is approximately on the carrier's flight deck (runway).
 * This is meant for collision-detection.
 *****************************************************************************/
bool
Carrier::IfRunwayAltitude( const WorldVertex& position )
{
CHECK_TYPESIG(this,TYPESIG_CARRIER);

	const SphereVertex posSph = Geo::WorldVertex2SphereVertex( position );
	const Meter altDelta = CARRIER_RUNWAY_ALT - posSph.mAlt;
	return ABS(altDelta.FP()) < CARRIER_RUNWAY_ALT_THRESHOLD.FP();  // within +-N meters
}

} // namespace program
