/*****************************************************************************
 * $LastChangedDate: 2008-10-04 18:55:38 -0400 (Sat, 04 Oct 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Top-level object (the program).
 *//*
 * LEGAL:   COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define PROGRAM_MODULE_CC 1
#include "base/module.hh"
using namespace base;
#include "math/module.hh"
using namespace math;
#include "view/conf.hh"
using namespace view;
#include "graph/module.hh"
#include "graph/model_cache.hh"
using namespace graph;
#include "object/module.hh"
#include "object/aircraft.hh"
using namespace object;
#include "world/module.hh"
#include "world/conf.hh"
using namespace world;
#include "fx/module.hh"
#include "fx/fx.hh"
using namespace fx;
#include "sound/module.hh"
using namespace sound;
#include "init/module.hh"
using namespace init;
#include "lua_bind/module.hh"
using namespace lua;
#include "program/module.hh"
#include "program/conf.hh"
#include "program/hud.hh"
#include "program/aircraft_models.hh"

//==============================================================================

/*****************************************************************************
 * main.
 *****************************************************************************/
int main( int argc, char** argv )
{
	try
	{
		// Start module and enter main loop.
		gProgram.Run( argc, argv );
	}
	catch ( const Exception& e )
	{
		CERROR << "EXCEPTION! " << endl << e.GetText() << endl;
		EXIT( 1 );
	}
	catch (...)
	{
		CERROR << "EXCEPTION!" << endl;
		EXIT( 1 );
	}
    EXIT( 0 );
}

//==============================================================================

namespace program {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  Program  ////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *
 * As always, beware of NULLs during initialization.
 *****************************************************************************/
Program::Program( void )
:	mPause(false),
	mWindow(NULL),
	mControl(NULL),
	mCommandQueue()
{
ASSERT_RUN_ONCE;
}

Program::~Program()
{
	// NOP
}

/*****************************************************************************
 * Run!
 *****************************************************************************/
void
Program::Run( int argc, char** argv )
{
	base::Streams::Init( argc, argv );

    // Splash.
	CLOG << "===========================================" << endl
         << "Palomino Flight Simulator "                  << endl
         << "(C) 2008 Jim E. Brooks - www.palomino3d.org" << endl
         << "===========================================" << endl
#if PROFILE == 1
		 << "[profNoOpt "
#elif PROFILE == 2
		 << "[profOpt "
#elif DEBUG
         << "[dbg "
#else
         << "[rel "
#endif
         << "OSG " << OPENSCENEGRAPH_MAJOR_VERSION << '.'
                   << OPENSCENEGRAPH_MINOR_VERSION << '.'
                   << OPENSCENEGRAPH_PATCH_VERSION << ']' << endl;
#if COMPILE_GUI
	CLOG << "Press F1 to start GUI." << endl;
#endif

	// Initialize and start modules.
	const init::InitArgs initArgs( argc, argv, defs::PROGRAM_NAME_SHORT );
	init::Init( initArgs );
	program::global::gConf.Init( initArgs.mArgc, initArgs.mArgv, defs::CONF_TXT_FILENAME );
	init::Start();

	// Assign pointers to Singletons (for speed).
	mWindow  = &Window::GetInstance();
	mControl = &Control::GetInstance();

	// Create World.
	gWorld.Make();

#if DEBUG
	SelfChecksAfterWorld();
#endif

	// Start Lua.
	GET_LUA().ExportCmdLine( argc, argv );
	GET_LUA().RunScript( "startup.lua" );

	// Optimize scene-graph.
  //GET_WORLD().OptimizeSceneGraph();  // let Lua do this

	// Register timer-tick handler.
	// Pausing must not stop this tick handler else the Lua keyboard handler won't run.
	Timer::GetInstance().Register( new TickFunctor,
                                   PROGRAM_CONF.mTimerTickFreq,
                                   Timer::CANNOT_DISABLE );
	// Splash.
	SplashHud::GetInstance().Start();

	// Enter the graphics main loop.
	mWindow->EnterMainLoop();

	// Bye!
	EXIT( 0 );
}

/*****************************************************************************
 * Reset program (Backspace pressed).
 *****************************************************************************/
void
Program::Reset( void )
{
	// Stop explosions.
	GET_FX().Reset();

	// Stop playing sounds.
	sound::Sound::GetInstance().StopAll();

#if DEBUG
	CDEBUG << "objects = " << Object::GetObjectCnt() << endl;
#endif
}

/*****************************************************************************
 * The pulse of the module.
 *****************************************************************************/
void
Program::TickFunctor::operator()( const Milliseconds millisecElapsed )
{
	gProgram.Tick( millisecElapsed );
}

void
Program::Tick( const Milliseconds millisecElapsed )
{
	// Run a Lua script every tick.
	GET_LUA().RunScript( "tick.lua" );
}

/*****************************************************************************
 * Pause simulation.
 *****************************************************************************/
void
Program::SetPause( bool pause_ )  // pause() is a syscall
{
	mPause = pause_;

	// To pause, disable most timer-tick handlers.
	// The tick that propagates to Lua (which handles keyboard) stays enabled regardless.
	Timer::GetInstance().Enable( not pause_ );
}

bool
Program::IfPaused( void )
{
	return mPause;
}

/*****************************************************************************
 * Command queue is a queue of strings which are processed by Lua scripts.
 *****************************************************************************/
void
Program::EnqueueCommand( const string& cmd )
{
	mCommandQueue.push_back( cmd );
}

/*****************************************************************************
 * @return Next command or empty string.
 *****************************************************************************/
string
Program::DequeueCommand( void )
{
	if ( mCommandQueue.empty() )
	{
		return string();
	}
	else
	{
		const string cmd = mCommandQueue.front();
		mCommandQueue.pop_front();
		return cmd;
	}
}

/*****************************************************************************
 * Debug
 *****************************************************************************/
#if DEBUG
void
Program::SelfChecksAfterWorld()
{
	// Test that World or something else doesn't keep a reference
	// to an Object after it is detached.
 {
	const int objectCnt = Object::GetObjectCnt();
   {
	shptr<Object> obj = Object::MakeDummyObject();
	ASSERT( objectCnt + 1 == Object::GetObjectCnt() );
	gWorld.Attach( obj );
	gWorld.Detach( obj );
   }
	ASSERT( objectCnt == Object::GetObjectCnt() );
 }

	// Test a complex Object (Aircraft) to ensure that
	// it has no references to itself (the ref-cnt-cycle bug).
	// Failing to free an Object loaded with a 3D model file
	// will waste lots of memory.
 {
	const int objectCnt = Object::GetObjectCnt();
   {
	shptr<Aircraft> aircraft = MakeAircraft( eAircraftType_P51, WorldVertex(0,0,0), ModelCache::LOAD_SHARED );
	ASSERT( objectCnt + 1 == Object::GetObjectCnt() );
	gWorld.Attach( aircraft.PTR() );
	gWorld.Detach( aircraft.PTR() );
   }
	ASSERT( objectCnt == Object::GetObjectCnt() );  // should be destroyed
 }
}
#endif // DEBUG

} // namespace program
