/*****************************************************************************
 * $LastChangedDate: 2008-09-14 01:03:50 -0400 (Sun, 14 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Module configuration.
 *//*
 * LEGAL:	COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define SKY_CONF_CC 1
#include "base/module.hh"
#include "base/conf.hh"
using namespace base;
#include "types/rgba.hh"
using namespace types;
#include "world/module.hh"
#include "world/conf.hh"
using namespace world;
#include "sky/module.hh"
#include "sky/defs.hh"
#include "sky/conf.hh"

namespace sky {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  Conf  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

INTERN bool sSkyColorWasSet = false;

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Conf::Conf( void )
:	mValid(false),
	// cmd-line args 
    // (none)
	// .conf file
	mSkyColor(sky::defs::DEFAULT_SKY_COLOR),
	mSkyFogDensity(sky::defs::DEFAULT_SKY_FOG_DENSITY),  // fpx
	mSkyDomeSliceCnt(sky::defs::DEFAULT_SKY_DOME_SLICE_CNT),
	mCloudsEnabled(true),
	mCloudConfFlat(),
	mCloudConfSphere(),
	mStarCnt(sky::defs::DEFAULT_STAR_CNT),
	mLightEnabled(true),  // Lua will configure lighting
	mFogEnabled(true),
	mFogColor(sky::defs::DEFAULT_FOG_COLOR),
    mFogMode(osg::Fog::EXP),
	mFogDensity(sky::defs::DEFAULT_FOG_DENSITY)  // fpx
{
}

/*****************************************************************************
 * Initialize/read configuration.
 *****************************************************************************/
void
Conf::Init( int argc, char** argv, const string& confTxtFilename )
{

BEGIN_ONCE {

	mValid = true;  // assign this first as *_CONF may expand to GetConf() which asserts it

	ParseCmdlineArgs( argc, argv );   // cmdline args parsed prior to reading .conf file
	ReadConfTxt( confTxtFilename );   // read from textual .conf file
	ParseCmdlineArgs2( argc, argv );  // cmdline args that override .conf file

} END_ONCE

}

/*****************************************************************************
 * Parse cmdline args.
 *****************************************************************************/
void
Conf::ParseCmdlineArgs( int argc, char** argv )
{
	// NOP
}

void
Conf::ParseCmdlineArgs2( int argc, char** argv )
{
	// Parse cmd-line args.
	for ( int i = 1; i < argc; ++i )
	{
		if ( STREQ( argv[i], "-high" ) )
		{
			sSkyColorWasSet = true;  // tell ConfTxt it was set
			SKY_CONF.mSkyColor = sky::defs::HIGH_SKY_COLOR;
		}
		else if ( STREQ( argv[i], "-no-light" ) )
		{
			// Disabling means to omit light sources.
			mLightEnabled = false;
			CLOG << "Disabled lighting." << endl;
		}
		else if ( STREQ( argv[i], "-fog" ) )
		{
			SKY_CONF.mFogEnabled = true;
			CLOG << "Enabled fog." << endl;
		}
		else if ( STREQ( argv[i], "-no-fog" ) )
		{
			SKY_CONF.mFogEnabled = false;
			CLOG << "Disabled fog." << endl;
		}
		else if ( STREQ( argv[i], "-no-clouds" )
               or STREQ( argv[i], "-fast" ) )
		{
			SKY_CONF.mCloudsEnabled = false;
			CLOG << "Disabled clouds." << endl;
		}
	}
}

/*****************************************************************************
 * Read from .conf file.
 *****************************************************************************/
void
Conf::ReadConfTxt( const string& confTxtFilename )
{
	try
	{
		ModuleConfTxt confTxt;
		confTxt.Read( string(base::GetDataDir()) + string(confTxtFilename) );
	}
	catch (...)
	{
		CERROR << "ERROR: cannot read " << confTxtFilename << endl;
		EXIT( 1 );
	}
}

void
Conf::ModuleConfTxt::ProcessField( string& field, ifstream& conf )
{
	bool error = false;

	if ( field == "SkyColor" and not sSkyColorWasSet )
	{
		fp r, g, b, a;
		conf >> r;
		conf >> g;
		conf >> b;
		conf >> a;
		SKY_CONF.mSkyColor.Set( r, g, b, a );
	}
	else if ( field == "SkyFogDensity" )
	{
		conf >> SKY_CONF.mSkyFogDensity;
	}
	else if ( field == "SkyDomeSlices" )
	{
		conf >> SKY_CONF.mSkyDomeSliceCnt;
	}
	else if ( field == "Stars" )
	{
		conf >> SKY_CONF.mStarCnt;
	}
	//---------------------  fog  ---------------------------------------------
	else if ( field == "FogColor" )
	{
		fp r, g, b, a;
		conf >> r;
		conf >> g;
		conf >> b;
		conf >> a;
		SKY_CONF.mFogColor.Set( r, g, b, a );
	}
	else if ( field == "FogMode" )
	{
		string s; conf >> s;  // string
	    if ( s == "linear" )
	    {
		    SKY_CONF.mFogMode = osg::Fog::LINEAR;
	    }
		else if ( s == "exp2" )
	    {
		    SKY_CONF.mFogMode = osg::Fog::EXP2;
	    }
	    else  // default
	    {
			error = (s != "exp");
	        SKY_CONF.mFogMode = osg::Fog::EXP;
	    }
	}
	else if ( field == "FogDensity" )
	{
		conf >> SKY_CONF.mFogDensity;
	}
	else
	{
		// Need to match prefix with correlating member.
		if ( field.find( "Flat" ) != string::npos )
			ProcessField( field, conf, SKY_CONF.mCloudConfFlat );
		else if ( field.find( "Sphere" ) != string::npos )
			ProcessField( field, conf, SKY_CONF.mCloudConfSphere );
	}
	
    // Errors?
	if ( error )
	{
		CERROR << "ERROR[conf]: parameter for " << field << " is invalid." << endl;
		EXIT( 1 );
	}
}

void
Conf::ModuleConfTxt::ProcessField( string& field, ifstream& conf, CloudConf& cloudConf )
{
	if ( field == "FlatClouds" or field == "SphereClouds" )
	{
		conf >> cloudConf.mCloudCnt;
	}
	else if ( field == "FlatCloudMinAlt" or field == "SphereCloudMinAlt" )
	{
		fp f; conf >> f; cloudConf.mCloudMinAlt = Meter(f);
	}
	else if ( field == "FlatCloudMaxAlt" or field == "SphereCloudMaxAlt" )
	{
		fp f; conf >> f; cloudConf.mCloudMaxAlt = Meter(f);
	}
	else if ( field == "FlatCloudMinWidth" or field == "SphereCloudMinWidth" )
	{
		fp f; conf >> f; cloudConf.mCloudMinWidth = Meter(f);
	}
	else if ( field == "FlatCloudMaxWidth" or field == "SphereCloudMaxWidth" )
	{
		fp f; conf >> f; cloudConf.mCloudMaxWidth = Meter(f);
	}
}

/*****************************************************************************
 * Print Conf configuration.
 *****************************************************************************/
ostream&
operator<<( ostream& strm, const Conf& conf )
{
	std::ios::fmtflags savedFlags = strm.flags();

	strm << "conf [sky]:"                                                        << endl
	     << "SkyColor                    = " << conf.mSkyColor                   << endl
	     << "FogColor                    = " << conf.mFogColor                   << endl
	     << "FogMode                     = " << conf.mFogMode                    << endl
	     << "FogDensity                  = " << conf.mFogDensity              ;//<< endl
	strm.flags(savedFlags);

	return strm;
}

} // namespace sky
