/*****************************************************************************
 * $LastChangedDate: 2008-09-28 22:26:59 -0400 (Sun, 28 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Window class.
 * @remarks - Window resize problems:
 *            OSG supportsResize is set to false since multiple problems
 *            occur if window is resize (some problems are external).
 *            The worst external problem is severe artifacts
 *            if motion-blur (accumulation buffer) is enabled (Nvidia 7600).
 *          - Trackball mode:
 *            The simulator runs in a reduced mode if trackball is enabled
 *            as it is intended for debugging or creating a scene.
 *//*
 * LEGAL:   COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define VIEW_WINDOW_CC 1
#include "base/module.hh"
#include "base/timer.hh"
using namespace base;
#include "glue/module.hh"
#include "glue/gfx.hh"
#include "glue/conf_bin_ext.hh"
using namespace glue;
#include "view/module.hh"
#include "view/window.hh"
#include "view/conf.hh"
#include "view/node_masks.hh"

#include <osg/StateAttribute>
#include <osgGA/StateSetManipulator>

namespace view {

const int DEFAULT_STATS_HANDLER_KEY = osgGA::GUIEventAdapter::KEY_F12;

////////////////////////////////////////////////////////////////////////////////
/// @brief Minimal keyboard handler for trackball mode.
///
/// Window installs a minimal keyboard handler since the Lua scripts
/// (which normally handle keyboard events) don't run in trackball mode.
///
class TrackballKeyboardHandler : public osgGA::GUIEventHandler
{
public:
    virtual bool handle( const osgGA::GUIEventAdapter& event, osgGA::GUIActionAdapter& )
	{
		if ( event.getEventType() & osgGA::GUIEventAdapter::KEYDOWN )
	    {
			switch ( event.getKey() )
			{
				case osgGA::GUIEventAdapter::KEY_F2:
				{
					GET_WINDOW().SaveViewpoint();
					return true;
				}
				break;

				case ' ':
				{
					GET_WINDOW().RestoreViewpoint();
					return true;
				}
				break;

				default: break;
			}
		}
		return false;
	}
};

////////////////////////////////////////////////////////////////////////////////
//////////////////////////   CameraDrawCallback  ///////////////////////////////
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
/// @brief For broadcasting pre-draw and post-draw events for a View.
///
/// OSG calls this before and after rendering an osg::Camera.
/// This is reflected to any listeners in the app registered by base::Event.
///
class CameraDrawCallback : public osg::Camera::DrawCallback
{
public:
	enum { PRE_DRAW = true, POST_DRAW = false };

	CameraDrawCallback( shptr<View> view, const bool preDraw )
	:	mView(view), mPreDraw(preDraw)
	{
	}

	virtual void operator()( const osg::Camera& ) const
	{
		// Reflect from OSG as broadcasted event.
		if ( mPreDraw )
			VIEW_EVENTS.mPreDrawOneView.Broadcast( mView );
		else
			VIEW_EVENTS.mPostDrawOneView.Broadcast( mView );
	}

private:
	shptr<View>		mView;
	const bool		mPreDraw;	///< pre-draw or post-draw
};

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////  Window  //////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

Window* Window::msInstance;  // there can be only one

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Window::Window( const WidthHeight<int> wh, const string& title )
:	mSceneGraph(new SceneGraph),
	mViewer(new osgViewer::Viewer),
    mViews(),
    mTitle(title),
	mFPS(),
    mTraits(NULL),
    mGraphicsContext(NULL),
    mGraphicsWindow(NULL),
	mOperation(new Operation),
	mStatsHandler(NULL),  // assume stats are disabled
	mTrackballManipulator(new osgGA::TrackballManipulator)
{
	// Catch user passing bad args with -wh.
	if ( wh.w <= 0 or wh.h <= 0 or wh.w > 10000 or wh.h > 10000 )
	{
		CERROR << "ERROR: invalid window width/height" << endl;
		EXIT( 1 );
	}

    // Create window.
	mTraits = new osg::GraphicsContext::Traits;
    mTraits->x = 100;
    mTraits->y = 100;
    mTraits->width  = wh.w;
    mTraits->height = wh.h;
    mTraits->windowDecoration = true;
    mTraits->doubleBuffer = true;
    mTraits->sharedContext = NULL;
	mTraits->windowName = title;
    mTraits->supportsResize = false;
    mGraphicsContext = osg::GraphicsContext::createGraphicsContext( mTraits.get() );
    mGraphicsWindow = dynamic_cast<osgViewer::GraphicsWindow*>( mGraphicsContext.get() );
    if ( mGraphicsWindow == NULL )
    {
        throw Exception( "ERROR: unable to create OSG GraphicsWindow." );
    }

	// [2007/12] Multiple threads causes instability
	// because this application modifies the scene graph
	// while OSG is still visiting it.
	// Use the single-threaded model.
    mViewer->setThreadingModel( osgViewer::Viewer::SingleThreaded );

	// Create the main view.
	AddView( new View( ENABLE,
                       Rect<int>( 0, 0, wh.w, wh.h ),
                       new ChaseViewpoint,
                       mViewer->getCamera() ),
             true );  // true : main view

	// Set the absolute root node of everything.
	RefPtr<Node> rootNode = mSceneGraph->GetSuperRootNode().get();
    mViewer->setSceneData( rootNode.get() );

	// If trackball, let OSG handle ESC, since -t bypasses Lua and its keyboard handler.
    if ( VIEW_CONF.mUseTrackball )
	{
		OsgSetKeyboardHandler( new TrackballKeyboardHandler );
	  //RestoreViewpoint();  // too soon, would be a NOP
	}
	else
    {
    	// To let app handle ESC key.
    	mViewer->setKeyEventSetsDone( 0 );
    }

	// Wirefame?
	if ( VIEW_CONF.mWireframe )
	{
	    RefPtr<osg::PolygonMode> polygonMode = new osg::PolygonMode;
		polygonMode->setMode( osg::PolygonMode::FRONT_AND_BACK, osg::PolygonMode::LINE );
		rootNode->getOrCreateStateSet()->setAttributeAndModes( polygonMode.get(), osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON );
	}

    // Enable statistics display when a key is pressed.
	if ( VIEW_CONF.mStatsEnabled )
	{
		mStatsHandler = new osgViewer::StatsHandler;  // ctor does NOT instantiate
		mViewer->addEventHandler( mStatsHandler.get() );
		OsgSetStatsKey( DEFAULT_STATS_HANDLER_KEY );
	}

	// To do low-level graphics operations.
	mGraphicsWindow->add( mOperation.get() );
}

Window::~Window()
{
	// NOP
}

/*****************************************************************************
 * Singleton method.
 * Parameters for creating a Window (size,title) come from the view module's conf file.
 *****************************************************************************/
Window*
Window::CreateInstance( void )
{
ASSERT_RUN_ONCE;
	return new Window( WidthHeight<int>( VIEW_CONF.mWindowWidth,
                                         VIEW_CONF.mWindowHeight ),
                       VIEW_CONF.mWindowTitle );
}

/*****************************************************************************
 * Enter the main graphics loop.
 *****************************************************************************/
void
Window::EnterMainLoop( void )
{
	// Bypass processing other views if trackball is enabled.
    if ( VIEW_CONF.mUseTrackball )
    {
		// SUBTLE ordering of OSG code:
		// RestoreViewpoint() alters the manipulator AFTER
		// it is attached to Viewer (else restoring will have no effect).
		mViewer->setCameraManipulator( mTrackballManipulator.get() );
		RestoreViewpoint();
        mViewer->run();
        EXIT( 0 );
    }

	// Mark beginning of time for base::Timer.
	// The higher-level base::Timer is pulsed by main loop using osg::Timer.
	osg::Timer timer;
	timer.setStartTick();

	while ( true )
	{
		// Pulse the Timer singleton.
		Milliseconds elapsedMilsec = Milliseconds( (unsigned long)timer.time_m() );
		base::Timer::GetInstance().Tick( elapsedMilsec );

		// Broadcast event.
		VIEW_EVENTS.mPreDrawAllViews.Broadcast( Void() );

		// For each View, re-compute its view matrix.
		for ( Views::iterator iter = mViews.begin(); iter != mViews.end(); ++iter )
		{
			shptr<View> view = *iter;
			CHECK_TYPESIG(view,TYPESIG_VIEW);
			if ( view->IfEnabled() )
			{
				// Recompute the view matrix (involves the Viewpoint functor).
				view->ComputeViewMatrix();
			}
		}

		// Render a frame.  This renders all views.
        mViewer->frame();

		// Compute FPS.
		mFPS.Tick( elapsedMilsec );

		// Broadcast event.
	  //VIEW_EVENTS.mPostDrawAllViews.Broadcast( Void() );  // done at Window::DoOperation()

		// Exit program?
		if ( mViewer->done() )
			break;
	}

	EXIT( 0 );
}

/*****************************************************************************
 * Attach osg::Camera to this Window.
 *****************************************************************************/
void
Window::OsgAddCamera( RefPtr<osg::Camera> camera, const eCamera cameraType )
{
	// Bypass processing other views if trackball is enabled.
    if ( not VIEW_CONF.mUseTrackball )
    {
    	// Add osg::Camera to osg::Viewer.
    	// But don't add master as a slave of itself (prevent cycle).
    	if ( not mViewer->containsCamera( camera.get() ) )
    	{
    		// HUD/2D camera shouldn't use 3D scene data.
    		const bool useMastersSceneData = (cameraType == eCamera_3D);
    		mViewer->addSlave( camera.get(), useMastersSceneData );
    	}
    }

	// Set where camera renders to.
	camera->setGraphicsContext( mGraphicsContext.get() );
}

/*****************************************************************************
 * Add View to this Window.
 *****************************************************************************/
void
Window::AddView( shptr<View> view, const bool mainView )
{
CHECK_TYPESIG(view,TYPESIG_VIEW);

	// Add View object to container.
	mViews.push_back( view );

	// Add underlying osg::Camera to osg::Viewer.
	RefPtr<osg::Camera> camera = view->OsgGetCamera();
	OsgAddCamera( camera );

    // Set viewport where camera renders into.
	const Rect<int> rect = view->GetRect();
    camera->setViewport( rect.x1,
                         rect.y1,
                         rect.x2 - rect.x1,    // width
                         rect.y2 - rect.y1 );  // height

	// Set the camera's cull mask (node mask).
	// This is for the ability to render 3D models in one view but not others.
	// A specific case is disabling the player's model in OTW view.
	camera->setCullMask( mainView ? defs::NODE_MASK_MAIN_CAMERA
                                  : defs::NODE_MASK_AUX_CAMERA );

	// Register a callback with OSG to broadcast events.
	camera->setPreDrawCallback(  new CameraDrawCallback( view, CameraDrawCallback::PRE_DRAW ) );
	camera->setPostDrawCallback( new CameraDrawCallback( view, CameraDrawCallback::POST_DRAW ) );

	// RELATIVE_RF isn't appropritate (but is for manipulators).
	if ( not VIEW_CONF.mUseTrackball )
		camera->setReferenceFrame( osg::Transform::ABSOLUTE_RF );

	// Try to suppress culling of aircraft (doesn't seem to help much).
	camera->setCullingMode( camera->getCullingMode() & ~osg::CullStack::SMALL_FEATURE_CULLING );
}

/*****************************************************************************
 * Save/restore viewpoint to binary configuration.
 *****************************************************************************/
void
Window::SaveViewpoint( void )
{
	if ( VIEW_CONF.mUseTrackball )
	{
		RefPtr<osgGA::MatrixManipulator> manipulator = mViewer->getCameraManipulator();
		const Matrix m        = manipulator->getInverseMatrix();
		const fp     distance = manipulator->getMinimumDistance();
		const bool res = WriteConfBin( glue::GetConfBin(), "trackballMatrix", m )
                     and WriteConfBin( glue::GetConfBin(), "trackballDistance", distance );
		if ( res )
		{
			CLOG << "Saved viewpoint (trackball)." << endl;
		}
		else
		{
			CERROR << "ERROR: Failed to save viewpoint (trackball)" << endl;
		}
	}
	else
	{
		CERROR << "[Window::SaveViewpoint] UNIMPLEMENTED case" << endl;
	}
}

void
Window::RestoreViewpoint( void )
{
	// Restore viewpoint from binary configuration.
	if ( VIEW_CONF.mUseTrackball )
	{
		Matrix  m;
		fp distance;
		const bool res = ReadConfBin( glue::GetConfBin(), "trackballMatrix", m )
	                 and ReadConfBin( glue::GetConfBin(), "trackballDistance", distance );
		if ( res )
		{
			CLOG << "Restored viewpoint (trackball)." << endl;
			mTrackballManipulator->setByInverseMatrix( m );
			mTrackballManipulator->setMinimumDistance( distance );
		}
	}
	else
	{
		CERROR << "[Window::RestoreViewpoint] UNIMPLEMENTED case" << endl;
	}
}

/*****************************************************************************
 * Get window size.
 *****************************************************************************/
WidthHeight<int>
Window::GetWindowSize( void )
{
	return WidthHeight<int>( GetWindowRect() );  // convert
}

/*****************************************************************************
 * Get ABSOLUTE window rectangle.
 * NOTE: x1,y1 will usually have non-zero values.
 *****************************************************************************/
Rect<int>
Window::GetWindowRect( void )
{
	// Get the current size of OSG GraphicsWindow (window may be resized externally).
	int x = 0, y = 0, w = 0, h = 0;
	mGraphicsWindow->getWindowRectangle( x, y, w, h );
	ASSERT( BOOL_CAST( w|h ) );
	return Rect<int>( x, y, x+w, y+h );  // convert w,h to x2,y2
}

/*****************************************************************************
 * Dump scene graph to a file.
 *****************************************************************************/
void
Window::SaveSceneGraphToFile( const string& filename )
{
	// Defer.
	GetSceneGraph()->DumpToFile( filename );
}

/*****************************************************************************
 * The "input" module calls this to install a keyboard handler.
 *****************************************************************************/
void
Window::OsgSetKeyboardHandler( RefPtr<osgGA::GUIEventHandler> keyboardHandler )
{
	mViewer->addEventHandler( keyboardHandler.get() );
}

/*****************************************************************************
 * Set which key toggles the statistics display.
 *****************************************************************************/
void
Window::OsgSetStatsKey( const int key )
{
	// Statistics might be entirely disabled.
	if ( VIEW_CONF.mStatsEnabled and mStatsHandler != NULL )
	{
		mStatsHandler->setKeyEventTogglesOnScreenStats( key );
	}
}

/*****************************************************************************
 * Called by OSG to let app do low-level graphics operations.
 *****************************************************************************/
void
Window::DoOperation( void )
{
	// Call post-draw listeners.
	VIEW_EVENTS.mPostDrawAllViews.Broadcast( Void() );

	// Draw outline around Views (unless disabled).
	Views::iterator viewIter = mViews.begin();
	ASSERT( viewIter != mViews.end() );
	++viewIter;  // skip main view
	for ( ; viewIter != mViews.end(); ++viewIter )
	{
		shptr<View> view = *viewIter;
		if ( view->IfEnabled() )
			DrawViewOutline( view, view->GetOutlineColor() );
	}
}

/*****************************************************************************
 * Draw outline around 2D viewpoint.
 *****************************************************************************/
void
Window::DrawViewOutline( shptr<View> view, const RGBA outlineColor )
{
CHECK_TYPESIG(view,TYPESIG_VIEW);

	GFX::DrawBox( view->GetRect(), outlineColor );
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////  Window::Operation  ////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Window::Operation::Operation( void )
:	osg::Operation("WindowOperation",true)
{
	// NOP
}

Window::Operation::~Operation()
{
	// NOP
}

/*****************************************************************************
 * Functor.
 *****************************************************************************/
void
Window::Operation::operator()( osg::Object* )
{
	// Reflect to Window.
	GET_WINDOW().DoOperation();
}

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////   Window::FPS  ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor.
 ****************************************************************************/
Window::FPS::FPS( void )
:	mFPS(0),
	mFrameCnt(0),
	mFrameStart(0)
{
	// NOP
}

/*****************************************************************************
 * To be called after rendering every View.
 *
 * This computes momentary FPS (FPS within the last second)
 * which reflects rapid changes in frame-rate.
 * Another way is average FPS but that hides frame-rate drop.
 *****************************************************************************/
void
Window::FPS::Tick( const Milliseconds elapsedMilsec )
{
	// Count this frame.
	++mFrameCnt;

	// Is this second finished?
	if ( elapsedMilsec - mFrameStart >= Milliseconds(1000) )
	{
		// mFrameCnt directly is frames/second.
		// Save value in mFPS which ComputeFPS() will return later.
		mFPS = mFrameCnt;

		// Reset for the next second.
		mFrameStart = elapsedMilsec;
		mFrameCnt = 0;
	}
}

} // namespace view
