/*****************************************************************************
 * $LastChangedDate: 2008-09-13 22:52:35 -0400 (Sat, 13 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Object to contain group nodes sorted by graphical state.
 *//*
 * LEGAL:   COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define WORLD_STATE_SORTER_CC 1
#if VERBOSE
#define STATE_SORTER_VERBOSE 1
#else
#define STATE_SORTER_VERBOSE 0
#endif
#include "base/module.hh"
using namespace base;
#include "glue/module.hh"
#include "glue/osg.hh"
using namespace glue;
#include "object/module.hh"
using namespace object;
#include "graph/module.hh"
using namespace graph;
#include "world/module.hh"
#include "world/defs.hh"
#include "world/state_sorter.hh"

#include <osg/BlendFunc>

namespace world {

StateBits StateSorter::msId = 0;

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////  StateSorter  ////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
StateSorter::StateSorter( void )
:	mSortedGroups()
{
	// NOP
}

StateSorter::~StateSorter()
{
	// NOP
}

/*****************************************************************************
 * @return Group node that sorts its sub-nodes by state.
 * When creating a new group, automatically attaches it under the passed parent node.
 * @param	parent
 *			If no group exists yet, it will be created and attached to this parent.
 * @param	stateBits
 *			Bitwise combination of STATE_BIT_* or ID from AllocateStateBitsId().
 * @param	stateSet
 *          Optional.  Used if no group exists yet.
 *			Allows client to set graphics state beyond STATE_BIT_* enums.
 *****************************************************************************/
RefPtr<GroupNode>
StateSorter::GetSortedGroup( RefPtr<GroupNode>     parent,
                             const StateBits       stateBits,
                             RefPtr<osg::StateSet> stateSet )
{
	// For speed, stateSet arg is NULL by default
	// (alternative is new osg::StateSet but is slow and usually wasted).

	// Find group-node in STL map using state-bits as key.
	SortedGroup::iterator iter = mSortedGroups.find( stateBits );
	if ( EX( iter != mSortedGroups.end() ) )
	{
		#if STATE_SORTER_VERBOSE
		CLOG << std::hex << "StateSorter: Found existing group node for StateBits = 0x" << stateBits << endl;
		#endif

		// Found group.
		return iter->second;
	}
	else
	{
		#if STATE_SORTER_VERBOSE
		CLOG << std::hex << "StateSorter: Creating group node for StateBits = 0x" << stateBits << endl;
		#endif

		// Not found, insert new group into map.
		RefPtr<GroupNode> group = new GroupNode;
		mSortedGroups.insert( make_pair<>(stateBits, group) );

		// Automatically attach new group under passed parent group.
		parent->addChild( group.get() );

		//......................................................................
		// Set stateset appropriately.
		if ( stateSet == NULL )
			stateSet = new osg::StateSet;
		group->setStateSet( stateSet.get() );  // assign arg

		if ( stateBits & STATE_BIT_NO_LIGHTING )
		{
			stateSet->setMode( GL_LIGHTING, osg::StateAttribute::OFF | osg::StateAttribute::OVERRIDE );
		}
		if ( stateBits & STATE_BIT_NO_FOG )
		{
			stateSet->setMode( GL_FOG, osg::StateAttribute::OFF );
		}
		if ( stateBits & STATE_BIT_BLEND )
		{
		    stateSet->setMode( GL_BLEND, osg::StateAttribute::ON );
		    stateSet->setAttributeAndModes( new osg::BlendFunc, osg::StateAttribute::ON );
		}
		if ( stateBits & STATE_BIT_TRANSPARENT )
		{
		    stateSet->setRenderingHint( osg::StateSet::TRANSPARENT_BIN );
		}
		if ( stateBits & STATE_BIT_NO_DEPTH_TEST )
		{
		    stateSet->setMode( GL_DEPTH_TEST, osg::StateAttribute::OFF );
		}
		//......................................................................

		return group;
	}
}

/*****************************************************************************
 * For state-sorting, allocate a unique value.  This generalizes StateSorter.
 *
 * An ID can represent any combination of graphical states (textures, colors, etc)
 * beyond the pre-defined STATE_BIT enums.
 *
 * Though, STATE_BIT_* enums and an ID can be bitwise-ORed.
 * This is useful because StateSorter::GetSortedGroup() will automatically
 * set graphical states per STATE_BIT_*.
 *****************************************************************************/
StateBits
StateSorter::AllocateStateBitsId( void )
{
	msId >>= STATE_BIT_SHIFT_ID;	// shift down to do arithmetic
	++msId;
	msId <<= STATE_BIT_SHIFT_ID;	// undo shift
#if STATE_SORTER_VERBOSE_
	CLOG << std::hex << "StateSorter: allocating id = 0x" << msId << endl;
#endif
	return msId;
}

} // namespace world
