/*****************************************************************************
 * $LastChangedDate: 2008-10-21 14:43:19 -0400 (Tue, 21 Oct 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	World class.
 *//*
 * LEGAL:   COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define WORLD_WORLD_CC 1
#define NEED_PLANET_CLASS 1
#include "base/module.hh"
#include "base/conf.hh"
using namespace base;
#include "glue/module.hh"
#include "glue/osg.hh"
using namespace glue;
#include "object/module.hh"
using namespace object;
#include "graph/module.hh"
using namespace graph;
#include "collision/module.hh"
using namespace collision;
#include "view/module.hh"
using namespace view;
#include "sky/module.hh"
#include "sky/conf.hh"
using namespace sky;
#include "world/module.hh"
#include "world/defs.hh"
#include "world/conf.hh"
#include "world/events.hh"
#include "world/planet.hh"

namespace world {

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////  World  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 * Make() is the actual ctor.
 *****************************************************************************/
World::World( void )
:	mRootGroup(new GroupNode),
	mStateSorter(new StateSorter)
{
	// NOP
}

World::~World()
{
	// NOP
}

/*****************************************************************************
 * Called by world::Init() [world/module.cc].
 *****************************************************************************/
void
World::Make( void )
{
BEGIN_ONCE {

	// Attach to scene.
	GET_SCENE_GRAPH()->AttachNode( mRootGroup.get() );

	// Set the clear/background color.
	if ( WORLD_CONF.mFlat )
	    GET_SCENE_GRAPH()->SetClearColor( RGBA(0.1f,0.1f,0.1f,1.0f) );
	else  // [2008/04] sky not fully implemented for OSSIM
		GET_SCENE_GRAPH()->SetClearColor( SKY_CONF.mSkyColor );

	// Create and attach Planet.  OSSIM is too dark with lighting.
	shptr<Object> planet = GET_PLANET().GetRootObject();
	planet->SetCollidable( true );  // planet is a collidable object
	Attach( planet, WORLD_CONF.mFlat ? STATE_BIT_DEFAULT : STATE_BIT_NO_LIGHTING );

} END_ONCE
}

/*****************************************************************************
 * Attach Object to World under appropriate sub-graph.
 *
 * The pair of stateBits,stateSet is somewhat of an idiom.
 * stateBits is a value that exists on the app-side to state-sort
 * the nodes of the scene-graph.
 * stateSet is a object that affects graphical state (applied by OSG).
 * Thru StateSorter, stateBits affects osg::StateSet.
 * The caller can pass a osg::StateSet arg modified in ways not possible by StateBits.
 *
 * @param	object
 * @param	stateBits
 *			Value that represents a graphical state (osg::StateSet).
 * @param	stateSet
 *			Actual graphical state (osg::StateSet).
 *          This is optional (if not passed, then new osg::StateSet is compiled).
 *****************************************************************************/
void
World::Attach( shptr<Object> object,
               const StateBits stateBits,
               RefPtr<osg::StateSet> stateSet )
{
CHECK_TYPESIG(object,TYPESIG_OBJECT);

    // Don't re-attach.
    RefPtr<Node> node = object->GetGraph()->GetRootNode();
    if ( EX( node->getNumParents() == 0 ) )
	{
		// Call subroutine.
		Attach( node, stateBits, stateSet );
	}
}

/*****************************************************************************
 * Attach Node to World under appropriate sub-graph.
 * [see Attach(Object)]
 *****************************************************************************/
void
World::Attach( RefPtr<Node> node,
               const StateBits stateBits,
               RefPtr<osg::StateSet> stateSet )
{
    // Ensure Node is unattached.
    if ( EX( node->getNumParents() == 0 ) )
	{
		// For optimal state-sorting.
		RefPtr<GroupNode> sortedGroup = \
			mStateSorter->GetSortedGroup( mRootGroup, stateBits, stateSet );
		sortedGroup->addChild( node.get() );

		// Optimize this node's subgraph.
		GET_SCENE_GRAPH()->OptimizeNode( node );
	}
}

/*****************************************************************************
 * Detach Object from World.
 * @post Detached Object will be marked non-collidable and probably destroyed.
 *****************************************************************************/
void
World::Detach( shptr<Object> object )
{
CHECK_TYPESIG(object,TYPESIG_OBJECT);

	// Call subroutine.
	Detach( object->GetRootNode() );

	// CRITICAL: All modules holding a reference to an Object
	// should listen for this event and release their reference,
	// to avoid consuming memory.
	WORLD_EVENTS.mDetachObject.Broadcast( object );

	// Zombify the Object to prevent exhausting memory because of large 3D models.
	object->Zombify();
}

/*****************************************************************************
 * Detach Node from World.
 * (mostly a subroutine)
 *****************************************************************************/
void
World::Detach( RefPtr<Node> node )
{
    // An node should be attached to 0 or 1 group nodes.
    const uint parentCnt = node->getNumParents();
    if ( parentCnt == 1 )
    {
    	// Has one parent which should be an osg::Group.
        RefPtr<osg::Group> parent = node->getParent( 0 );  // idx=0
    #if DEBUG
        ASSERT( parent->asGroup() != NULL );                // should be a Group or derivative
        ASSERT( parent->removeChild( node.get() ) );    // remove() should return true
    #else
        parent->removeChild( node.get() );
    #endif
    }
    else // parentCnt != 1
    {
        // Multiple parents shouldn't occur and isn't handled.
        // Objects may have shared nodes for 3D models but
        // those should be underneath unique group nodes (osg::MatrixTransform).
        ASSERT( parentCnt == 0 );
    }

    ASSERT( node->getNumParents() == 0 );  // should have zero parents
}

/*****************************************************************************
 * Optimize entire scene-graph.
 *****************************************************************************/
void
World::OptimizeSceneGraph( void )
{
	GET_SCENE_GRAPH()->OptimizeSceneGraph();
}

/*****************************************************************************
 * If the World is flat ;-)
 * This is queried since FlatPlanet is deficient and some configuration
 * is only applicable to FlatPlanet.
 *****************************************************************************/
bool
World::IfFlat( void )
{
	return GET_PLANET().IfFlat();
}

/*****************************************************************************
 * Get radius of World/Planet.
 *****************************************************************************/
fp
World::GetRadius( void )
{
	return GET_PLANET().RadiusGraphical();
}

/*****************************************************************************
 * Set compression of World.
 * @returns Previous flag.
 *****************************************************************************/
bool
World::SetCompression( const bool enable )
{
	return GET_PLANET().SetCompression( enable );
}

/*****************************************************************************
 * Get compression scale of World.
 *****************************************************************************/
fpx
World::GetCompressionScale( void )
{
	return GET_PLANET().GetCompressionScale();
}

/*****************************************************************************
 * Orient an Object to be tangent to a point on the world (in case world is spherical).
 *****************************************************************************/
void
World::RotateTangential( Object& object, const SphereVertex& sv )
{
	if ( not GET_PLANET().IfFlat() )
	{
		Matrix m = GetTangentMatrix( sv );
		m.setTrans( object.GetPosition() );  // keep Object's position
		MatrixRotateLocal( m, XX, Deg2Rad(90.0f) );  // otherwise Object points up (normal to surface)
		object.SetMatrix( m );
	}
	// else flat so NOP
}

/*****************************************************************************
 * Compute a matrix that is tangent to a point on the world.
 *****************************************************************************/
const Matrix&
World::GetTangentMatrix( const SphereVertex& sv )
{
	return GET_PLANET().GetTangentMatrix( sv );
}

/*****************************************************************************
 * Compute a normalized "up vector" extending from a point on the world.
 * A "down vector" can be computed by multipling by -1.0f.
 *****************************************************************************/
WorldVertex
World::ComputeUpVector( const SphereVertex& sv )
{
	return GET_PLANET().ComputeUpVector( sv );
}

/*****************************************************************************
 * Compute an "up vector" where an Object is at.
 *****************************************************************************/
WorldVertex
World::ComputeUpVector( Object& object )
{
CHECK_TYPESIG(&object,TYPESIG_OBJECT);

	return ComputeUpVector( Geo::WorldVertex2SphereVertex( object.GetPosition() ) );
}

/*****************************************************************************
 * @return Volume of world in 3D space.
 *****************************************************************************/
BoxVolume
World::GetVolume( void )
{
	return GET_PLANET().GetVolume();
}

} // namespace world
